/* 
    AnimSequence.cpp

	Store/restore animation of models.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include <stdafx.h>
#include <assert.h>
#include "AnimSequence.h"

AnimSaver::AnimSaver(FencerModel *m1, FencerModel *m2, CJudge *judge, char *fname)
{
	m_m1 = m1; m_m2 = m2; m_judge = judge;
	m_f = fopen(fname, "wb");
	if (!m_f) 
	{
		::MessageBox(0, "Can't open file for saving animation", "Error", MB_OK);	    
		return;
	}
	m_timer.reset();
	// na zacatek napisu kvailty a pocty trojuhelniku obou modelu:
	assert(m_m1->m_quality == m_m2->m_quality);
	fwrite(&m_m1->m_quality, sizeof(int), 1, m_f);
	fwrite(&m_m2->m_quality, sizeof(int), 1, m_f);
	int tri1 = m_m1->m_model->getNumTriangles();
	fwrite(&tri1, sizeof(int), 1, m_f);
	int tri2 = m_m2->m_model->getNumTriangles();
	fwrite(&tri2, sizeof(int), 1, m_f);
}

// zapise aktualni cas (v ms) a stav modelu do souboru:
void AnimSaver::saveState(int tag)
{
	if (!m_f) return;
	float time = m_timer.getTime();
	fwrite(&time, sizeof(float), 1, m_f);
	fwrite(m_m1->getModelState(), sizeof(ModelState), 1, m_f);
	fwrite(m_m2->getModelState(), sizeof(ModelState), 1, m_f);
	fwrite(&tag, sizeof(int), 1, m_f);
	fwrite(m_judge->getText(), cMaxJudgeMessage, 1, m_f);
	fwrite(m_judge->getScore(), cMaxJudgeMessage, 1, m_f);
	for (int n=0; n<m_m1->m_model->getNumTriangles(); n++)
	{
		byte a = m_m1->m_model->getTriangleHit(n);
		fwrite(&a, sizeof(byte), 1, m_f);
	}
	for (n=0; n<m_m2->m_model->getNumTriangles(); n++)
	{
		byte a = m_m2->m_model->getTriangleHit(n);
		fwrite(&a, sizeof(byte), 1, m_f);
	}
}

/*
typedef struct SAnimRecord {
	float time;
	ModelState s1, s2;
} AnimRecord;

class AnimLoader {
protected:
	int m_records;
	AnimRecord *seq; // vlastni sekvence

public:
	AnimLoader(RemoteModel *m1, RemoteModel *m2, char *fname);
	~AnimLoader();

	void loadState(float time);
};
*/

AnimLoader::AnimLoader(const char *fname)
{
	m_m1 = 0; m_m2 = 0; m_quality = 0;
	m_seq = 0; m_records = 0; m_actRecord = 0; // kdyby to nahodou nevyslo...
	FILE *f = fopen(fname, "rb");
	if (!f) 
	{
		::MessageBox(0, "Can't open file with animated sequence", "Error", MB_OK);		
		return;
	}
	fseek(f, 0, SEEK_END);
	int len = ftell(f);
	len -= 4*sizeof(int); // tech 4*sizeof(int) je za uvodni 4 inty:
	fseek(f, 0, SEEK_SET);

	int qua[2];
	fread(&qua[0], sizeof(int), 1, f);
	fread(&qua[1], sizeof(int), 1, f);
	if (qua[0] != qua[1])
	{
		MessageBox(0, "Invalid sequence file (different 'quality')", "Error", MB_ICONSTOP);
		return;
	}
	m_quality = qua[0];

	fread(&m_tri[0], sizeof(int), 1, f);
	fread(&m_tri[1], sizeof(int), 1, f);
	TRACE("In file: qua0 = %i, qua1 = %i, tri0 = %i, tri1 = %i\n", qua[0], qua[1], m_tri[0], m_tri[1]);	

	int recSize = sizeof(float)+2*sizeof(ModelState)+sizeof(int) + 2*cMaxJudgeMessage + 
		m_tri[0]*sizeof(bool) + m_tri[1]*sizeof(bool);	
	//assert(len % recSize == 0); // melo by byt delitelny
	if (len % recSize != 0)
	{
		MessageBox(0, "Invalid sequence file (len % recSize != 0)", "Error", MB_ICONSTOP);
		return;
	}
	
	m_records = len/recSize;
	m_seq = new AnimRecord[m_records];
	for (int n=0; n<m_records; n++)
	{
		fread(&m_seq[n].time, sizeof(float), 1, f);
		fread(&m_seq[n].s1, sizeof(ModelState), 1, f);
		fread(&m_seq[n].s2, sizeof(ModelState), 1, f);
		fread(&m_seq[n].tag, sizeof(int), 1, f);
		fread(&m_seq[n].judgeText, cMaxJudgeMessage, 1, f);
		fread(&m_seq[n].judgeScore, cMaxJudgeMessage, 1, f);
		m_seq[n].m_tri[0] = new byte[m_tri[0]];
		m_seq[n].m_tri[1] = new byte[m_tri[1]];
		fread(m_seq[n].m_tri[0], sizeof(byte)*m_tri[0], 1, f);
		fread(m_seq[n].m_tri[1], sizeof(byte)*m_tri[1], 1, f);
	}
	assert(!feof(f));	
	fclose(f);
	m_actRecord = 0;
}

AnimLoader::~AnimLoader()
{
	delete[] m_seq;
}

int AnimLoader::loadState(float time)
{
	time *=2;
	assert(!"Dodelat");
	return 0;
}

int AnimLoader::loadRecord(int rec)
{
	if (rec < 0 || rec >= m_records) return -1;

	m_m1->setModelState((char *)&m_seq[rec].s1);
	m_m1->changedModelState();
	m_m2->setModelState((char *)&m_seq[rec].s2);
	m_m2->changedModelState();

	for (int n=0; n<m_m1->m_model->getNumTriangles(); n++)
		m_m1->m_model->setTriangleHit(n, m_seq[rec].m_tri[0][n]);
	for (n=0; n<m_m2->m_model->getNumTriangles(); n++)
		m_m2->m_model->setTriangleHit(n, m_seq[rec].m_tri[1][n]);

	return m_seq[rec].tag;
}

