/* 
    AnimSequence.h

	Store/restore animation of models.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "LocalRemoteModel.h"
#include "Communication.h"
#include "Win32Timer.h"
#include "Judge.h"

#ifndef ANIMSEQUENCE_H
#define ANIMSEQUENCE_H

class AnimSaver {
private:
	FILE *m_f;
	FencerModel *m_m1, *m_m2; // modely na sejvovani
	CJudge *m_judge; // rozhodciho hlasky na sejvovani
	Win32Timer m_timer;

public:
	AnimSaver(FencerModel *m1, FencerModel *m2, CJudge *judge, char *fname);
	~AnimSaver() { if (m_f) fclose(m_f); }

	// zapise aktualni cas a stav modelu do souboru, tag je libovolna doplnkova informace:
	void saveState(int tag = 0);
};

typedef struct SAnimRecord {
	float time;
	ModelState s1, s2;
	int tag;
	char judgeScore[cMaxJudgeMessage];
	char judgeText[cMaxJudgeMessage];
	byte *m_tri[2]; // pole zasazenych trojuhelniku pro oba modely
} AnimRecord;

class AnimLoader {
protected:
	int m_records; // celkovy pocet zaznamu
	int m_actRecord; // na kterem zaznamu jsem
	AnimRecord *m_seq; // vlastni sekvence
	RemoteModel *m_m1, *m_m2;
	int m_quality; // ulozena v souboru
	int m_tri[2]; // pocty trojuhelniku ulozenych modelu

public:
	AnimLoader(const char *fname);
	~AnimLoader();
	void setRemoteModels(RemoteModel *m1, RemoteModel *m2) { m_m1=m1; m_m2=m2; }

	// nahraje stav v nejvyssim case <= time, vraci tag:
	int loadState(float time);
	// nastavi stav v rec-tem zaznamu, vraci tag:
	int loadRecord(int rec);
	// vrati text/score (i = 0/1) v rec-tem zaznamu:
	const char *getJudgeMsg(int i, int rec)
	{
		if (rec < 0 || rec >= m_records) return 0;
		if (i==0) return m_seq[rec].judgeText;
		else if (i==1) return m_seq[rec].judgeScore;
		else { assert(false); return 0; }
	}
	// vraci pocet zaznamu:
	int getNumRecords() { return m_records; }
	// vrati time v rec-tem zaznamu:
	float getTime(int rec) 
	{ 
		if (rec < 0 || rec >= m_records) return -1;
		return m_seq[rec].time; 
	}
	int getQuality() { return m_quality; }
	int getNumTri(int i) { return m_tri[i]; }
};

#endif