// CameraView.cpp : implementation file
//

#include "stdafx.h"
#include "gl1.h"
#include "CameraView.h"
#include "ChildView.h"
#include <assert.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCameraView

CCameraView::CCameraView(CWnd *view, CRect rect, int bits, int camId)
{
	m_view = view;
	m_active = false;
	m_fullscreen = false;
	m_bits = bits;
	m_camId = camId;
	m_yrot = 270;
	m_zrot = 0;
	m_zpos = m_ypos = 0;
	m_viewMode = cExamine;
	m_transparent[0] = m_transparent[1] = false;
	m_watchJoint = 0; // root
	// vyrobim okno:	
	LPCTSTR className = AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW | CS_OWNDC);
	char nazev[64];
	sprintf(nazev, "Camera %i", camId);
	CreateEx(WS_EX_APPWINDOW, className, nazev, WS_OVERLAPPEDWINDOW|WS_VISIBLE|WS_CAPTION, rect, 0, 0);	
	m_myMenu.LoadMenu(IDR_MENU1);
	SetMenu(&m_myMenu);
	//TRACE("m_pokusny = %p\n", m_pokusny);
	InitRendering(bits);
}

CCameraView::~CCameraView()
{
	if (m_fullscreen)										// Are We In Fullscreen Mode?
	{
		::ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
		::ShowCursor(TRUE);								// Show Mouse Pointer
	}

}

bool CCameraView::InitRendering(int bits)
{
  GLuint		PixelFormat;			// Holds The Results After Searching For A Match
  HDC hDC;
  static	PIXELFORMATDESCRIPTOR pfd=				// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),				// Size Of This Pixel Format Descriptor
		1,											// Version Number
		PFD_DRAW_TO_WINDOW |						// Format Must Support Window
		PFD_SUPPORT_OPENGL |						// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,							// Must Support Double Buffering
		PFD_TYPE_RGBA,								// Request An RGBA Format
		bits,										// Select Our Color Depth
		0, 0, 0, 0, 0, 0,							// Color Bits Ignored
		0,											// No Alpha Buffer
		0,											// Shift Bit Ignored
		0,											// No Accumulation Buffer
		0, 0, 0, 0,									// Accumulation Bits Ignored
		32,											// 16Bit Z-Buffer (Depth Buffer)  
		0,											// No Stencil Buffer
		0,											// No Auxiliary Buffer
		PFD_MAIN_PLANE,								// Main Drawing Layer
		0,											// Reserved
		0, 0, 0										// Layer Masks Ignored
	};

	hDC=::GetDC(this->m_hWnd);
	if (!hDC)							// Did We Get A Device Context?
	{
//		KillGLWindow();								// Reset The Display
		::MessageBox(NULL,"Can't Create A GL Device Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	PixelFormat=::ChoosePixelFormat(hDC,&pfd);
	if (!PixelFormat)	// Did Windows Find A Matching Pixel Format?
	{
		//KillGLWindow();								// Reset The Display
		::MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!::SetPixelFormat(hDC,PixelFormat,&pfd))		// Are We Able To Set The Pixel Format?
	{
//		int err = GetLastError();
		//KillGLWindow();								// Reset The Display
		::MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);		
		return FALSE;								// Return FALSE
	}

	hRC=::wglCreateContext(hDC);
	if (!hRC)				// Are We Able To Get A Rendering Context?
	{
		//KillGLWindow();								// Reset The Display
		::MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!::wglMakeCurrent(hDC,hRC))					// Try To Activate The Rendering Context
	{
		//KillGLWindow();								// Reset The Display
		::MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	::ShowWindow(this->m_hWnd,SW_SHOW);						// Show The Window
	::SetForegroundWindow(this->m_hWnd);						// Slightly Higher Priority
	::SetFocus(this->m_hWnd);									// Sets Keyboard Focus To The Window

	m_active = true;
	ReSizeGLScene();					// Set Up Our Perspective GL Screen

	if (!InitGL())									// Initialize Our Newly Created GL Window
	{
		FreeRendering();							
		::MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	::ReleaseDC(this->m_hWnd, hDC);	
	return TRUE;
}

void CCameraView::FreeRendering(void)
{
 m_active = false;
 if (hRC)											// Do We Have A Rendering Context?
	{
		glDeleteLists(m_base, 96);							// Delete All 96 Characters
		if (!::wglMakeCurrent(NULL,NULL))					// Are We Able To Release The DC And RC Contexts?
		{
			::MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}

		if (!::wglDeleteContext(hRC))						// Are We Able To Delete The RC?
		{
			::MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		hRC=NULL;										// Set RC To NULL		
	}
}

BOOL CCameraView::PreCreateWindow(CREATESTRUCT& cs) 
{
    if (!CWnd::PreCreateWindow(cs))
		return FALSE;

	cs.dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	cs.style &= ~WS_BORDER;    	

	cs.lpszClass = AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_OWNDC|CS_DBLCLKS, 
		::LoadCursor(NULL, IDC_CROSS), NULL, NULL);

	return TRUE;
}

/*
GLvoid CCameraView::ReSizeGLScene()					// Resize And Initialize The GL Window
{
    RECT t;
	GetClientRect(&t);
	int width = t.right - t.left;
	int height = t.bottom - t.top;


	if (height==0)										// Prevent A Divide By Zero By
	{
		height=1;										// Making Height Equal One
	}

	glViewport(0,0,width,height);						// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,10.0f,1000.0f);

	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix
}
*/

GLvoid CCameraView::ReSizeGLScene()					// Resize And Initialize The GL Window
{
	if (!m_active) return;
	setMyRC();

//	TRACE("CCameraView::ReSizeGLScene\n");

    RECT t;
	GetClientRect(&t);
	m_width = t.right - t.left;
	m_height = t.bottom - t.top;

	if (m_height==0)										// Prevent A Divide By Zero By
	{
		m_height=1;										// Making Height Equal One
	}

	glViewport(0,0,m_width,m_height);						// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,(GLfloat)m_width/(GLfloat)m_height,1.0f,1000.0f);
	//gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.1f,100.0f);

	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix
}


int CCameraView::InitGL(GLvoid)							// All Setup For OpenGL Goes Here
{
	setMyRC();
	((CChildView *)m_view)->reloadTextures(m_camId);
	glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
    glClearColor(0.0f, 1.0f, 0.0f, 0.0f);				// zelene pozadi
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
	//glDepthFunc(GL_LEQUAL);	-- to blbne s modelem!!!							// The Type Of Depth Testing To Do	
	glDepthFunc(GL_LESS);								// The Type Of Depth Testing To Do	
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations
	glEnable(GL_LIGHTING);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);		  

	// fontik:
	HFONT	font;										// Windows Font ID
	m_base = glGenLists(96);								// Storage For 96 Characters
	font = CreateFont(	-32,							// Height Of Font
						0,								// Width Of Font
						0,								// Angle Of Escapement
						0,								// Orientation Angle
						FW_BOLD,						// Font Weight
						FALSE,							// Italic
						FALSE,							// Underline
						FALSE,							// Strikeout
						ANSI_CHARSET,					// Character Set Identifier
						OUT_TT_PRECIS,					// Output Precision
						CLIP_DEFAULT_PRECIS,			// Clipping Precision
						ANTIALIASED_QUALITY,			// Output Quality
						FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
						"Arial");					// Font Name
	if (font == NULL)
	{
		TRACE("Nepodarilo se inicializovat font\n");
	}
	CDC *dc = GetDC();
	SelectObject(dc->m_hDC, font);					// Selects The Font We Want
	wglUseFontBitmaps(dc->m_hDC, 32, 96, m_base);	// Builds 96 Characters Starting At Character 32
	ReleaseDC(dc);
	return TRUE;										// Initialization Went OK
}

int CCameraView::DrawGLScene(GLvoid)
{
	if (!m_active) return FALSE;
	CChildView *m_childView = (CChildView *)m_view;
	// priprava kresleni:
	setMyRC();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);				// Clear The Screen And The Depth Buffer
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();												// Reset The Modelview Matrix

	// lighting setup
	//GLfloat LightAmbient[]= { 0.5f, 0.5f, 0.5f, 1.0f }; 				// Ambient Light Values 
	GLfloat LightDiffuse[]= { 1.0f, 1.0f, 1.0f, 1.0f };				 // Diffuse Light Values 
	GLfloat LightPosition[]= { 0.0f, 0.0f, 0.0f, 1.0f };				 // Light Position 

	//glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);				// Setup The Ambient Light           
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);				// Setup The Diffuse Light
	glLightfv(GL_LIGHT1, GL_POSITION,LightPosition);			// Position The Light
	glEnable(GL_LIGHT1);							// Enable Light One

	//	gluLookAt( 75, 75, 75, 0, 0, 0, 0, 1, 0 );						// (3) Eye Postion (3) Center Point (3) Y-Axis Up Vector

	if (m_viewMode == cExamine || 
		(m_childView->getMyState()!=cSimulate && m_childView->getMyState()!=cReplay &&
		 m_childView->getMyState()!=cViewer)) // neni-li stav simulace nebo replay, pak je povolen jenom rezim Examine
	{
		//glTranslatef(0.0, -30.0, -200.0);
		glTranslatef(0.0, 0.0, -250.0);
		//TRACE("yrot = %f, zrot = %f\n", m_yrot, m_zrot);
		//	float rad = ((float)m_zrot*M_PI)/180;
		//	// rotace okolo orotovane osy y (je orotovana kolem osy z o zrot stupnu):
		//	glRotatef(m_yrot, sin(rad), cos(rad), 0.0);
		glTranslatef(0, -m_ypos, m_zpos);
		glRotatef(m_yrot, 0.0, 1.0, 0.0);
		glRotatef(m_zrot, 0.0, 0.0, 1.0);
	}
	else if (m_viewMode == cAvatar0)
	{
		//assert(m_childView->getMyState() == 1);
		float pos[3], posOp[3];
		m_childView->getModel(0)->getJointPos(pos, m_childView->getModel(0)->m_model->m_headJoint);
		m_childView->getModel(1)->getJointPos(posOp, m_watchJoint);
		//pos[2] += 20; // kooscek posoopnu (za hlavu)
		glTranslatef(0, -m_ypos, m_zpos);		
		glRotatef(m_yrot, 0.0, 1.0, 0.0);
		glRotatef(m_zrot, 1.0, 0.0, 0.0);
		//TRACE("m_ypos = %f, m_zpos = %f, m_yrot = %f, m_zrot = %f\n", m_ypos, m_zpos, m_yrot, m_zrot);
		gluLookAt(pos[0], pos[1], pos[2], posOp[0], posOp[1], posOp[2], 0.0, 1.0, 0.0);
	}
	else if (m_viewMode == cAvatar1)
	{
		float pos[3], posOp[3];
		m_childView->getModel(1)->getJointPos(pos, m_childView->getModel(1)->m_model->m_headJoint);
		m_childView->getModel(0)->getJointPos(posOp, m_watchJoint);
		//pos[2] -= 20; // kooscek posoopnu (za hlavu)
		glTranslatef(0, -m_ypos, m_zpos); // transfo jdou pozpatku, tady uz byla provedena txce LookAt
		glRotatef(m_yrot, 0.0, 1.0, 0.0);
		glRotatef(m_zrot, 1.0, 0.0, 0.0);
		gluLookAt(pos[0], pos[1], pos[2], posOp[0], posOp[1], posOp[2], 0.0, 1.0, 0.0);
	}
	else assert(false);

	// vlastni vykreslovani modelu:
	((CChildView *)m_view)->mainDraw(this, m_camId, m_transparent);
	
	// vypis texticku:
	if (m_text || m_score)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();	
		gluOrtho2D(0.0, (GLfloat) m_width, (GLfloat) m_height, 0.0);
		glMatrixMode(GL_MODELVIEW);						
		glLoadIdentity();					// Reset The View		
		glDisable(GL_LIGHTING);
		glColor3f(0.0, 0.0, 1.0);			
		CDC *dc = GetDC();
		if (m_text) 
		{	
			int txt_width = (dc->GetTextExtent(m_text)).cx;
			int txt_height = (dc->GetTextExtent(m_text)).cy;	
			glPrint(m_width/2 - txt_width/2, m_height/3 - txt_height/2, m_text);
		}
		if (m_score) glPrint(10, 30, m_score);

		// vratit zpatky:	
		ReleaseDC(dc);			
		glEnable(GL_LIGHTING);		
		ReSizeGLScene();
	}

	// prepnuti bufferu:
	CDC *cDC = GetDC();
	::SwapBuffers(cDC -> m_hDC);
	ReleaseDC(cDC);
	
	return TRUE;	
/*
static double rtri = 0.0;

	GLfloat v[4][3] = {{0.0, 1.0, 0.0},{-1.0, -1.0, 1.0},{1.0, -1.0, 1.0},
	{0.0, -1.0, -1.0}};
//	GLfloat c[4][3] = {{1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{0.0, 0.0, 1.0},{0.0, 0.0, 0.0}};

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
	glLoadIdentity();									// Reset The Current Modelview Matrix

	glTranslatef(-0.0f,0.0f,-5.0f);					// Move Left 1.5 Units And Into The Screen 6.0
	glRotatef(rtri, 1.0, 1.0, 0.0);

//    glEnable(GL_LIGHTING);

    float a[3] = {1.0, 0.0, 0.0};
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, a);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, a);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, a);

	glBegin(GL_TRIANGLES);						// Drawing Using Triangles	

   	    //glColor3fv( c[0]);
		glVertex3fv( v[0]);			// Top Of Triangle (Front)
		//glColor3fv( c[1]);
		glVertex3fv( v[1]);			// Left Of Triangle (Front)
		//glColor3fv( c[2]);
		glVertex3fv( v[2]);			// Right Of Triangle (Front)


	glEnd();							// Finished Drawing The Triangle

    rtri += 0.4;

	// prepnuti bufferu:
	CDC *cDC = GetDC();
	int ret = ::SwapBuffers(cDC -> m_hDC);
	assert(ret);
	ReleaseDC(cDC);

	return TRUE;										// Everything Went OK
*/
}


void CCameraView::setMyRC()
{
	CDC *cDC2 = GetDC();
	int ret = wglMakeCurrent(cDC2->m_hDC,hRC);
	assert(ret);
	ReleaseDC(cDC2);
}


BEGIN_MESSAGE_MAP(CCameraView, CWnd)
	//{{AFX_MSG_MAP(CCameraView)
	ON_WM_SIZE()
	ON_WM_CLOSE()
	ON_WM_DESTROY()
	ON_COMMAND(ID_MAIN_CLOSE, OnMainClose)
	ON_COMMAND(ID_MAIN_FULLSCREEN, OnMainFullscreen)
	ON_WM_MOUSEMOVE()
	ON_COMMAND(ID_VIEW_EXAMINE, OnViewExamine)
	ON_COMMAND(ID_VIEW_AVATAR0, OnViewAvatar0)
	ON_COMMAND(ID_VIEW_AVATAR1, OnViewAvatar1)
	ON_COMMAND(ID_VIEW_TRANSPARENTAVATAR0, OnViewTransparentavatar0)
	ON_COMMAND(ID_VIEW_TRANSPARENTAVATAR1, OnViewTransparentavatar1)
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_COMMAND(ID_LOOKAT_ROOT, OnLookatRoot)
	ON_COMMAND(ID_LOOKAT_ELBOW, OnLookatElbow)
	ON_COMMAND(ID_LOOKAT_HAND, OnLookatHand)
	ON_COMMAND(ID_LOOKAT_HEAD, OnLookatHead)
	ON_COMMAND(ID_LOOKAT_SHOULDER, OnLookatShoulder)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CCameraView message handlers

void CCameraView::OnSize(UINT nType, int cx, int cy) 
{
	CWnd::OnSize(nType, cx, cy);
		
	if (hRC) ReSizeGLScene();
}

void CCameraView::OnClose() 
{
	//m_active = false;
	CWnd::OnClose();
}

void CCameraView::OnDestroy() 
{
	FreeRendering();
	CWnd::OnDestroy();
	((CChildView *)m_view)->destroyCamera(m_camId);
}

void CCameraView::OnMainClose() 
{
	// TODO: Add your command handler code here
	DestroyWindow();	
}

void CCameraView::toggleFullscreen()
{
	FreeRendering();
	if (m_fullscreen)  
	{			
		DEVMODE dmScreenSettings;								// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));	// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= 640;				// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= 480;				// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= m_bits;					// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (::ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			// Failed
			::MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card.","GL",MB_OK|MB_ICONEXCLAMATION);
			m_fullscreen=false;		// Windowed Mode Selected.  Fullscreen = FALSE						
		}
		/**/
	} else /* !fullscreen */ {
		::ChangeDisplaySettings(NULL,0); // Switch Back To The Desktop
	}

/* -- neni k nicemu
	if (*fullscreen)  {
		m_wndView.ModifyStyle(WS_CLIPSIBLINGS|WS_CLIPCHILDREN,0);
		m_wndView.ModifyStyleEx(WS_EX_WINDOWEDGE,0);
	} else {
        m_wndView.ModifyStyle(0,WS_CLIPSIBLINGS|WS_CLIPCHILDREN);
		m_wndView.ModifyStyleEx(0,WS_EX_WINDOWEDGE);
	}
*/

	if (m_fullscreen)  {
		ModifyStyle(WS_OVERLAPPEDWINDOW, WS_POPUP);
		//m_wndStatusBar.ShowWindow(SW_HIDE);
        //m_wndToolBar.ShowWindow(SW_HIDE);
	} else {
        ModifyStyle(WS_POPUP, WS_OVERLAPPEDWINDOW);
		//m_wndStatusBar.ShowWindow(SW_SHOW);
        //m_wndToolBar.ShowWindow(SW_SHOW);
	}
	// spravne jeste adjust
	SetWindowPos(&wndTop, 0, 0, 640, 480, SWP_SHOWWINDOW|SWP_DRAWFRAME);
	InitRendering(m_bits);
	DrawMenuBar();
}

void CCameraView::OnMainFullscreen() 
{
	m_fullscreen = !m_fullscreen;
	toggleFullscreen();
	if (m_fullscreen)
	{
		m_myMenu.CheckMenuItem(ID_MAIN_FULLSCREEN, MF_CHECKED);
	}
	else
	{
		m_myMenu.CheckMenuItem(ID_MAIN_FULLSCREEN, MF_UNCHECKED);
	}
}

void CCameraView::reloadTextures()
{
	setMyRC();
	((CChildView *)m_view)->reloadTextures(m_camId);
}

float normalizeDeg(float ang)
{
	while (ang<0) ang += 360;
	while (ang>=360) ang -= 360;
	return ang;
}

void CCameraView::OnMouseMove(UINT nFlags, CPoint point) 
{	
	if (nFlags & MK_CONTROL)
	{ // rotace objektu:
		CPoint posun = point - m_lastPoint;
		// zmenit kurzor musim vzdycky, protoze pri pohybu mysi se automaticky nastavuje na kurzor okeni tridy
		CWinApp *myApp = ::AfxGetApp();
		::SetCursor(myApp->LoadStandardCursor(IDC_SIZEALL));			
		// vynuluju ten mensi posun, nebo oba, jsou-li stejne:
		if (fabs(posun.x) <= fabs(posun.y)) posun.x = 0;
		if (fabs(posun.y) <= fabs(posun.x)) posun.y = 0;

		m_yrot = normalizeDeg(m_yrot + posun.x);
		m_zrot = normalizeDeg(m_zrot + posun.y);		
	}
	if (nFlags & MK_SHIFT)
	{ // posun objektu:
		CPoint posun = point - m_lastPoint;
		// zmenit kurzor musim vzdycky, protoze pri pohybu mysi se automaticky nastavuje na kurzor okeni tridy
		CWinApp *myApp = ::AfxGetApp();
		::SetCursor(myApp->LoadStandardCursor(IDC_SIZEALL));			
		// vynuluju ten mensi posun, nebo oba, jsou-li stejne:
		if (fabs(posun.x) <= fabs(posun.y)) posun.x = 0;
		if (fabs(posun.y) <= fabs(posun.x)) posun.y = 0;

		m_ypos = m_ypos + posun.y;
		m_zpos = m_zpos + posun.x;		
	}

	// minula pozice mysi:
	m_lastPoint = point;
	CWnd::OnMouseMove(nFlags, point);
}

void CCameraView::OnViewExamine() 
{	
	m_viewMode = cExamine;
	checkViewMode();
	m_yrot = 270; m_zrot = 0; // aby byl fakt videt
	m_ypos = 0; m_zpos = 0;
}

void CCameraView::OnViewAvatar0() 
{	
	m_viewMode = cAvatar0;
	checkViewMode();
	// by default se zpruhledni:
	m_transparent[0] = true;
	GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR0, MF_CHECKED);	
	//// resetuju rotaci, aby byl fakt videt:
	//m_yrot = 0; m_zrot = 0;
	// defaultni pekne nastaveni:
	m_ypos = -11.000000; m_zpos = -88.000000;
	m_yrot = 1.000000; m_zrot = 350.000000;
}

void CCameraView::OnViewAvatar1() 
{
	m_viewMode = cAvatar1;
	checkViewMode();
	// by default se zpruhledni:	
	m_transparent[1] = true;
	GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR1, MF_CHECKED);	
	//// resetuju rotaci, aby byl fakt videt:
	//m_yrot = 0; m_zrot = 0;
	// defaultni pekne nastaveni:
	m_ypos = -11.000000; m_zpos = -88.000000;
	m_yrot = 1.000000; m_zrot = 350.000000;
}

void CCameraView::checkViewMode()
{
	CMenu *menu = GetMenu();
	if (m_viewMode == cExamine) 
	{
		menu -> CheckMenuItem(ID_VIEW_EXAMINE, MF_CHECKED);
		menu -> CheckMenuItem(ID_VIEW_AVATAR0, MF_UNCHECKED);
		menu -> CheckMenuItem(ID_VIEW_AVATAR1, MF_UNCHECKED);
		GrayAllLookAt();
	}
	else if (m_viewMode == cAvatar0)
	{
		menu -> CheckMenuItem(ID_VIEW_EXAMINE, MF_UNCHECKED);
		menu -> CheckMenuItem(ID_VIEW_AVATAR0, MF_CHECKED);			
		menu -> CheckMenuItem(ID_VIEW_AVATAR1, MF_UNCHECKED);
		EnableAllLookAt();
	}
	else if (m_viewMode == cAvatar1)
	{
		menu -> CheckMenuItem(ID_VIEW_EXAMINE, MF_UNCHECKED);
		menu -> CheckMenuItem(ID_VIEW_AVATAR0, MF_UNCHECKED);
		menu -> CheckMenuItem(ID_VIEW_AVATAR1, MF_CHECKED);			
		EnableAllLookAt();
	}
}

void CCameraView::OnViewTransparentavatar0() 
{
	m_transparent[0] = !m_transparent[0];

	if (m_transparent[0]) GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR0, MF_CHECKED);
	else GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR0, MF_UNCHECKED);
}

void CCameraView::OnViewTransparentavatar1() 
{
	m_transparent[1] = !m_transparent[1];

	if (m_transparent[1]) GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR1, MF_CHECKED);
	else GetMenu()->CheckMenuItem(ID_VIEW_TRANSPARENTAVATAR1, MF_UNCHECKED);
}

// forwardovani zprav hlavnimu oknu:
void CCameraView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	((CChildView *)m_view)->OnKeyDown(nChar, nRepCnt, nFlags);
	CWnd::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CCameraView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{	
	((CChildView *)m_view)->OnKeyUp(nChar, nRepCnt, nFlags);
	CWnd::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CCameraView::OnLButtonDown(UINT nFlags, CPoint point) 
{	
	((CChildView *)m_view)->OnLButtonDown(nFlags, point);
	CWnd::OnLButtonDown(nFlags, point);
}

void CCameraView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	((CChildView *)m_view)->OnLButtonUp(nFlags, point);
	CWnd::OnLButtonUp(nFlags, point);
}

void CCameraView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	((CChildView *)m_view)->OnRButtonDown(nFlags, point);
	CWnd::OnRButtonDown(nFlags, point);
}

void CCameraView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	((CChildView *)m_view)->OnRButtonUp(nFlags, point);
	CWnd::OnRButtonUp(nFlags, point);
}

void CCameraView::OnLookatRoot() 
{		
	m_watchJoint = 0;
	UncheckAllLookAt();
	GetMenu()->CheckMenuItem(ID_LOOKAT_ROOT, MF_CHECKED);
}

void CCameraView::OnLookatElbow() 
{	
	int mi = 0; if (m_viewMode == cAvatar0) mi = 1;
	m_watchJoint = ((CChildView *)m_view)->getModel(mi)->m_model->m_ikj[ELBOW];
	UncheckAllLookAt();
	GetMenu()->CheckMenuItem(ID_LOOKAT_ELBOW, MF_CHECKED);
}

void CCameraView::OnLookatHand() 
{	
	int mi = 0; if (m_viewMode == cAvatar0) mi = 1;
	m_watchJoint = ((CChildView *)m_view)->getModel(mi)->m_model->m_ikj[HAND];
	UncheckAllLookAt();
	GetMenu()->CheckMenuItem(ID_LOOKAT_HAND, MF_CHECKED);
}

void CCameraView::OnLookatHead() 
{
	int mi = 0; if (m_viewMode == cAvatar0) mi = 1;
	m_watchJoint = ((CChildView *)m_view)->getModel(mi)->m_model->m_headJoint;
	UncheckAllLookAt();
	GetMenu()->CheckMenuItem(ID_LOOKAT_HEAD, MF_CHECKED);
}

void CCameraView::OnLookatShoulder() 
{
	int mi = 0; if (m_viewMode == cAvatar0) mi = 1;
	m_watchJoint = ((CChildView *)m_view)->getModel(mi)->m_model->m_ikj[SHOULDER];
	UncheckAllLookAt();
	GetMenu()->CheckMenuItem(ID_LOOKAT_SHOULDER, MF_CHECKED);
}

void CCameraView::UncheckAllLookAt()
{
	CMenu *menu = GetMenu();
	menu->CheckMenuItem(ID_LOOKAT_ROOT, MF_UNCHECKED);
	menu->CheckMenuItem(ID_LOOKAT_HEAD, MF_UNCHECKED);
	menu->CheckMenuItem(ID_LOOKAT_SHOULDER, MF_UNCHECKED);
	menu->CheckMenuItem(ID_LOOKAT_ELBOW, MF_UNCHECKED);
	menu->CheckMenuItem(ID_LOOKAT_HAND, MF_UNCHECKED);
}

void CCameraView::GrayAllLookAt()
{
	GetMenu()->EnableMenuItem(ID_LOOKAT_ROOT, MF_GRAYED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_HEAD, MF_GRAYED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_SHOULDER, MF_GRAYED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_ELBOW, MF_GRAYED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_HAND, MF_GRAYED);
}

void CCameraView::EnableAllLookAt()
{
	GetMenu()->EnableMenuItem(ID_LOOKAT_ROOT, MF_ENABLED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_HEAD, MF_ENABLED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_SHOULDER, MF_ENABLED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_ELBOW, MF_ENABLED);
	GetMenu()->EnableMenuItem(ID_LOOKAT_HAND, MF_ENABLED);
}

void CCameraView::glPrint(int x, int y, const char *fmt, ...)
{
	char		text[1024];								// Holds Our String
	va_list		ap;										// Pointer To List Of Arguments

	if (fmt == NULL)									// If There's No Text
		return;											// Do Nothing

	va_start(ap, fmt);									// Parses The String For Variables
	    vsprintf(text, fmt, ap);						// And Converts Symbols To Actual Numbers
	va_end(ap);											// Results Are Stored In Text

	glRasterPos2f(x, y);
	glPushAttrib(GL_LIST_BIT);							// Pushes The Display List Bits
	glListBase(m_base - 32);								// Sets The Base Character to 32
	glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);	// Draws The Display List Text
	glPopAttrib();										// Pops The Display List Bits
}
