/* 
    CameraView.h

	Window representing a camera view.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#if !defined(AFX_CAMERAVIEW_H__7EA404CB_7D48_4E9C_AABC_2274F9D3B776__INCLUDED_)
#define AFX_CAMERAVIEW_H__7EA404CB_7D48_4E9C_AABC_2274F9D3B776__INCLUDED_

#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <gl\glu.h>			// Header File For The GLu32 Library
#include <gl\glaux.h>		// Header File For The Glaux Library

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
// CameraView.h : header file
//

/////////////////////////////////////////////////////////////////////////////
// CCameraView window

// mody pohledu:
const cExamine = 1, cAvatar0 = 2, cAvatar1 = 3;

class CCameraView : public CWnd
{
protected:
	void setMyRC();
	CWnd *m_view; // ve skutecnosti CChildView
	CMenu m_myMenu; // moje menu
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);

	HGLRC		hRC;		// Permanent Rendering Context	
	bool	m_active; // bude true po inicializaci OpenGL az do zavreni okna
	bool	m_fullscreen; // jsem fullscreen?
	int		m_bits; // bits per pixel
	int		m_camId;

	int		m_width, m_height; // aktualni velikost klientske casti okna
	
	int		m_viewMode; // mod pohledu: cExamine, nebo cAvatar0, nebo cAvatar1
	float	m_zrot, m_yrot; // rotace zobrazovaneho modelu ve stupnich
	float	m_zpos, m_ypos; // posun zobrazovaneho modelu
	bool	m_rotating; // jestli hybu s modelem
	CPoint	m_lastPoint;

	bool	m_transparent[2]; // pruhledni avatari?
	int		m_watchJoint; // kloub, ktery sleduje avatar
	
	CString	m_text, m_score; // texty, ktery vypisuji

// Construction
public:	
	CCameraView(CWnd *view, CRect rect, int bits, int camId);

// Attributes
protected:
	GLint m_base; // baze fontu

// Operations
public:
	bool InitRendering(int bits);
	void FreeRendering(void);
	GLvoid ReSizeGLScene();						// Resize And Initialize The GL Window
	int InitGL(GLvoid);							// All Setup For OpenGL Goes Here
	int DrawGLScene(GLvoid);					// Here's Where We Do All The Drawin

	void setText(const CString &text) { m_text = text; }
	void setScore(const CString &score) { m_score = score; }

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CCameraView)
	//}}AFX_VIRTUAL

// Implementation
public:
	void UncheckAllLookAt();
	void reloadTextures();
	virtual ~CCameraView();

	// Generated message map functions
protected:
	void glPrint(int x, int y, const char *fmt, ...);
	void toggleFullscreen();
	//{{AFX_MSG(CCameraView)
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnClose();
	afx_msg void OnDestroy();
	afx_msg void OnMainClose();
	afx_msg void OnMainFullscreen();
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnViewExamine();
	afx_msg void OnViewAvatar0();
	afx_msg void OnViewAvatar1();
	afx_msg void OnViewTransparentavatar0();
	afx_msg void OnViewTransparentavatar1();
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnRButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnLookatRoot();
	afx_msg void OnLookatElbow();
	afx_msg void OnLookatHand();
	afx_msg void OnLookatHead();
	afx_msg void OnLookatShoulder();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
private:
	void EnableAllLookAt();
	void GrayAllLookAt();
	void checkViewMode();
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_CAMERAVIEW_H__7EA404CB_7D48_4E9C_AABC_2274F9D3B776__INCLUDED_)
