/* 
    ChildView.cpp

	Implementation of the CChildView class: input and display

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "gl1.h"
#include "ChildView.h"
#include "MilkshapeModel.h"
#include "Ikan.h"
#include "MainFrm.h"
#include "ColDet.h"
#include "ColResp.h"
#include "Physics.h"
#include "Matrix.h"
#include "AnimSequence.h"
#include "DirectInput.h"
#include "InputDevice.h"
#include "paths.h"
#include <assert.h>
#include "ModeSimulation.h"
#include "ModeReplay.h"
#include "ModeViewer.h"
#include "ModeAnimTest.h"
#include "ModeWait.h"
#include "ButtonConfigDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#pragma warning( disable : 4244 4305 )  

/////////////////////////////////////////////////////////////////////////////
// CChildView

#include "geometry.h"

CChildView::CChildView(CFrameWnd *owner, CWinApp *thisapp)
{
	m_inputManager = 0;
	m_quality = 0;
	m_dbgDialog.setParent(this);
	// sit se inicializuje az na prani:
	m_cs = 0; m_ss = 0; m_mode = 0;
	m_dbgDialogActive = false;
	int bits;
	// zjistime aktualni nastaveni barvne hloubky:
	DEVMODE dm;
	EnumDisplaySettings(0, ENUM_CURRENT_SETTINGS, &dm);
    bits = dm.dmBitsPerPel;

	// inicializace kamer:
	for (int n=0; n<cMaxCameras; n++) 
		m_camera[n] = 0;
	
	CRect winRect(0, 0, 640, 480);
	m_camera[0] = new CCameraView(this, winRect, bits, 0);	
	
	m_owner = owner;
	m_thisapp = thisapp;

	CRect rect(0, 50, 400, 450); // kus posunout kvuli liste...
	m_mid = new CMouseInputDevice(rect);

//	int kDOP = 18;	

	geometry_test();
//	matrix_test();
	pos_x = 20;	
	hRC = NULL;
}

void CChildView::afterWindowCreation()
{
	// na vytvareni dialogu bacha, musi se uvest parent, nebo si to domysli posledne vytvorene okno!
	// taky je potreba pockat az bude ChildView hotovej (ne v konstruktoru)
	if (!m_dbgDialog.Create(IDD_DIALOG1, m_owner)) TRACE0("Dialog creation failed.\n");			
	m_dbgDialogActive = true;
	resizeDebugDialog();

	//m_dbgDialog.m_edit1.SetWindowText("Hu1");
	//m_dbgDialog.m_edit2.SetWindowText("Hu2");
	m_dbgDialog.m_slider.SetRange(0, 100);
	m_dbgDialog.m_slider.SetPos( 20 );

/* pro swivel angle:
	m_dbgDialog.m_slider2.SetRange(0, 360);
	m_dbgDialog.m_slider2.SetPos( 0 );
/*
	m_dbgDialog.m_slider2.SetRange(-1, 30);
	m_dbgDialog.m_slider2.SetPos( -1 );
/**/
	m_dbgDialog.m_editHost.SetWindowText("192.168.0.1");
	m_dbgDialog.m_editPort.SetWindowText("1331");	

	m_dbgDialog.rebuild = true;

	// joysticky:	
	m_di = new CDirectInput(m_owner->m_hWnd, m_thisapp->m_hInstance, m_owner);
	m_owner->DrawMenuBar(); // mozna pribyly joysticky

	// zatim zadam jako locModely 0, spravne hodnoty se vyplni pmci InputManager.setLocModels
	m_inputManager = new InputManager(m_di, m_mid, 0, 0);
	((CMainFrame *)m_owner)->setInput(ID_INPUT0_MOUSE);
	((CMainFrame *)m_owner)->setInput(ID_INPUT1_MOUSE);

	// server socket je zatim NULL
	//m_mode = new CModeSimulation(m_inputManager, m_owner, &m_dbgDialog, 0);  
	//m_mode = new CModeReplay(m_owner, &m_dbgDialog);	
	//m_mode = new CModeViewer(m_owner, &m_dbgDialog, 0);	
	//m_mode = new CModeAnimTest(m_inputManager, m_owner, &m_dbgDialog);

	m_mode = 0;
	initMode(cSimulate);
	//initMode(cReplay);

//	initColRespSave();		
//	initReplay();
//	initAnimTest();
//	initCurvatureTest(); // pripominka jednoho z recenzentu

	m_netStatus = 0;		
	m_camera[0]->SetFocus(); // aby mi sly klavesy
}

void CChildView::initMode(int mod)
{
	delete m_mode;
	m_inputManager->setLocModels(0, 0); // kdyz to nahodou konstruktor nenastavi (replay a viewer)	
	//TRACE("CChildView::initMode    quality = %i\n", quality);
	if (mod == cSimulate) 
	{		
		m_dbgDialog.m_buttonReset.EnableWindow(true);
		m_mode = new CModeSimulation(m_inputManager, m_owner, &m_dbgDialog, &m_ss, m_quality);
	}	
	else if (mod == cReplay)
	{   // v CModeReplay automaticke nastaveni kvality (podle .sav souboru)
		m_dbgDialog.m_buttonReset.EnableWindow(false);
		m_mode = new CModeReplay(m_owner, &m_dbgDialog);
	}
	else if (mod == cViewer)
	{
		m_dbgDialog.m_buttonReset.EnableWindow(false);
		m_mode = new CModeViewer(m_inputManager, m_owner, &m_dbgDialog, &m_cs, m_quality);	
	}
	else if (mod == cAnimTest)
	{
		m_dbgDialog.m_buttonReset.EnableWindow(true);
		m_mode = new CModeAnimTest(m_inputManager, m_owner, &m_dbgDialog, m_quality);
	}
	else if (mod == cWait)
	{
		m_dbgDialog.m_buttonReset.EnableWindow(false);
		m_mode = new CModeWait();
	}
	else assert(false);

	if (m_quality==0) m_dbgDialog.m_subdivInfo.SetWindowText("No subdivision");
	else if (m_quality==1) m_dbgDialog.m_subdivInfo.SetWindowText("Subdivided");
	else assert(false);

	for (int n=0; n<cMaxCameras; n++)
		if (m_camera[n]) m_camera[n]->reloadTextures();
}


CChildView::~CChildView()
{
	delete m_mode;
	delete m_inputManager;

	delete m_ss;
	delete m_cs;
//	destroyCurrentState();	
	delete m_di;
	delete m_mid;

	delete m_camera[0];
	// vsechny ostatni kamery by mely byt zruseny pomoci fce destroyCamera:
	for (int n=0; n<cMaxCameras; n++)
		assert( m_camera[n] == NULL );	
}

BEGIN_MESSAGE_MAP(CChildView,CWnd )
	//{{AFX_MSG_MAP(CChildView)
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_COMMAND(ID_NETWORK_LISTENSERVER, OnNetworkListenserver)
	ON_COMMAND(ID_NETWORK_CONNECTCLIENT, OnNetworkConnectclient)
	ON_WM_SIZE()
	ON_COMMAND(ID_MAIN_NEWCAMERA, OnMainNewcamera)	
	ON_COMMAND(ID_MAIN_STARTJUDGE, OnMainStartjudge)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CChildView message handlers

BOOL CChildView::PreCreateWindow(CREATESTRUCT& cs) 
{
    if (!CWnd::PreCreateWindow(cs))
		return FALSE;

	cs.dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	cs.style &= ~WS_BORDER;    	

	cs.lpszClass = AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_OWNDC|CS_DBLCLKS, 
		::LoadCursor(NULL, IDC_CROSS), NULL, NULL);

	return TRUE;
}

void CChildView::reloadTextures(int camId)
{
	if (m_mode) m_mode->reloadTextures(camId);
}

void CChildView::mainLoop()
{
//	TRACE("CChildView::mainLoop - threadId = %i\n", GetCurrentThreadId());

	m_timer.reset();

	// podle aktualniho stavu provedu simulaci - bez kresleni:
	m_mode->play();

	// odtud se mi zpatky zavola metoda CChildView.mainDraw
	for (int n=0; n<cMaxCameras; n++)
		if (m_camera[n]) m_camera[n]->DrawGLScene();

	//FPS:
	double elapsed = m_timer.getTime();
	char pom[64];
	sprintf(pom, "%.2f", 1000/elapsed);	 
	m_dbgDialog.m_editFps.SetWindowText(pom);	
	//TRACE("current FPS = %f\n", 1000/elapsed);
}

void CChildView::OnMouseMove(UINT nFlags, CPoint point) 
{
//	MouseMove(nFlags, point);

	CWnd ::OnMouseMove(nFlags, point);
}

BOOL CChildView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{
//	MouseMove(nFlags, pt);

	if (zDelta>0) pos_x+=5;
	else pos_x-=5;
	char pom[16];
	sprintf(pom,"%.3f",pos_x);
	m_dbgDialog.m_edit3.SetWindowText(pom);
	//TRACE("%i",zDelta );
	
	return CWnd ::OnMouseWheel(nFlags, zDelta, pt);
}

void CChildView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->mouseEvent(WM_LBUTTONDOWN);
	/*
	if (m_myState == 1 || m_myState == 3)
	{
		m_locModel[0]->m_mid->feedButton(WM_LBUTTONDOWN);
		if (m_myState!=3) SetCapture(); // aby nam neutekla zprava o pusteni tlacitka
	}
	*/
	
	CWnd ::OnLButtonDown(nFlags, point);
}

void CChildView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->mouseEvent(WM_LBUTTONUP);
	/*
	if (m_myState == 1 || m_myState == 3)
	{
		m_locModel[0]->m_mid->feedButton(WM_LBUTTONUP);
		if (!(nFlags&MK_RBUTTON)) ReleaseCapture();
	}
	*/
	
	CWnd ::OnLButtonUp(nFlags, point);
}

void CChildView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->mouseEvent(WM_RBUTTONDOWN);
	/*
	if (m_myState == 1 || m_myState == 3)
	{
		m_locModel[0]->m_mid->feedButton(WM_RBUTTONDOWN);
		if (m_myState!=3) SetCapture(); // aby nam neutekla zprava o pusteni tlacitka
	}
	*/
	
	CWnd ::OnRButtonDown(nFlags, point);
}

void CChildView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->mouseEvent(WM_RBUTTONUP);
	/*
	if (m_myState == 1 || m_myState == 3)
	{
		m_locModel[0]->m_mid->feedButton(WM_RBUTTONUP);
		if (!(nFlags&MK_LBUTTON)) ReleaseCapture();
	}
	*/
	
	CWnd ::OnRButtonUp(nFlags, point);
}

void CChildView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->keyboardEvent(true, nChar);
	//if (m_myState == 1 || m_myState == 3) m_locModel[0]->m_kid -> KeyboardEvent(true, nChar, nRepCnt, nFlags);
	
	CWnd ::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CChildView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (m_inputManager) // kvuli kliknuti pred vyrobenim inputManagera
		m_inputManager->keyboardEvent(false, nChar);
	//if (m_myState == 1 || m_myState == 3) m_locModel[0]->m_kid -> KeyboardEvent(false, nChar, nRepCnt, nFlags);
	
	CWnd ::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CChildView::OnNetworkListenserver() 
{
	// 1) se prepnu do rezimu viewer a nedovolim to zmenit zpatky:	
	CMainFrame *mfr = (CMainFrame *)m_owner;
	initMode(cSimulate);	
	// mfr->DisableAll(); -- nefunguje: blby MFC...
	mfr->UnCheckAll();
	mfr->GetMenu()->CheckMenuItem(ID_FILE_GO, MF_CHECKED);

	Disconnect();
	// zakazu menit subdiv:
	m_dbgDialog.m_buttonReset.EnableWindow(false);
	// 2) inicializuju servrovsky socket
	m_ss = new CServerSocket(m_owner, m_quality);

	CString sport;
	m_dbgDialog.m_editPort.GetWindowText(sport);
	int port = atoi(sport);
	m_ss->startServer(port);
	TRACE("Listening started\n");
	m_netStatus = cServer;
	m_owner->GetMenu()->CheckMenuItem(ID_NETWORK_LISTENSERVER, MF_CHECKED);	
	//((CMainFrame *)m_owner)->setStatusText("Server");
}

void CChildView::OnNetworkConnectclient() 
{	
	Disconnect();
	// zatim jen unchecknu vse, mod viewer nastavim teprv az prijde potvrzeni o spojeni:
	((CMainFrame *)m_owner)->UnCheckAll();
	// cekaci rezim (taky zakazat prepnuti !):
	initMode(cWait);		
	
	// 2) inicializuju klientsky soket:
	// modely v m_mode se nesmi zmenit, jinak to bude CommSocket posilat buhvikam
	// proto nesmim z rezimu viewer odejit bez diskonnektnuti
	m_cs = new CCommSocket(this);		
	CString sport,shost;
	m_dbgDialog.m_editPort.GetWindowText(sport);
	int port = atoi(sport);
	m_dbgDialog.m_editHost.GetWindowText(shost);
	m_cs->startClient((LPCTSTR)shost, port);	

	// zbytek inicializace se provede az prijde prvni zprava od servru; fce onClientInit
}

void CChildView::onClientInit(int quality)
{
	// nastavim spravne prepinac kvalitky:
	m_quality = quality;
	TRACE("onClientInit hlasi kvalitu %i\n", quality);
	m_dbgDialog.m_checkedSubdiv.SetCheck(m_quality);
	// 1) se prepnu do rezimu viewer a nedovolim to zmenit zpatky:			
	initMode(cViewer);		
	m_cs->setModels(m_mode->getModel(0)->m_model, m_mode->getModel(1)->m_model);

	((CMainFrame *)m_owner)->GetMenu()->CheckMenuItem(ID_MAIN_VIEWERCLIENT, MF_CHECKED);

	m_netStatus = cClient;
	// nastavim, kam posilat zpravy od rozhodciho (musi byt CModeViewer!)
	m_cs->setViewer(m_mode);
	// vysilam oba avatary
	m_inputManager->setCommSockets(m_cs, m_cs);
	m_owner->GetMenu()->CheckMenuItem(ID_NETWORK_CONNECTCLIENT, MF_CHECKED);
	//((CMainFrame *)m_owner)->setStatusText("Client");
}

// volana z CCommSocketu kdyz se nepovedlo prikonekteni:
void CChildView::onClientFailed()
{
	delete m_cs; m_cs = 0;
	initMode(cSimulate);
	((CMainFrame *)m_owner)->GetMenu()->CheckMenuItem(ID_FILE_GO, MF_CHECKED);
}

// odpojim server/klient
void CChildView::Disconnect()
{
	if (m_cs)
	{
		// dulezite - zrusit:
		m_inputManager->setCommSockets(0, 0);
		delete m_cs;
		m_cs = 0;
	}
	if (m_ss)
	{
		delete m_ss;
		m_ss = 0;
	}
	m_owner->GetMenu()->CheckMenuItem(ID_NETWORK_LISTENSERVER, MF_UNCHECKED);
	m_owner->GetMenu()->CheckMenuItem(ID_NETWORK_CONNECTCLIENT, MF_UNCHECKED);
	m_dbgDialog.m_buttonReset.EnableWindow(true);
}

// kresleni podle aktualniho stavu:
void CChildView::mainDraw(CCameraView *camera, int camId, const bool transpAvatar[2])
{
	m_mode->draw(camera, camId, transpAvatar);
}

void CChildView::resizeDebugDialog()
{
	if (!m_dbgDialogActive) return;
	CRect rect;
	m_owner->GetClientRect(rect);
	m_dbgDialog.MoveWindow(rect);
}

void CChildView::OnSize(UINT nType, int cx, int cy) 
{
	CWnd ::OnSize(nType, cx, cy);
	
	resizeDebugDialog();
}

void CChildView::OnMainNewcamera() 
{
	int nc = -1;
	for (int n=0; n<cMaxCameras; n++)
	{
		if (!m_camera[n])
		{
			nc = n;
			break;
		}
	}
	if (nc==-1)
	{
		TRACE("Neni uz misto na nove kamery\n");
		return;
	}
	CRect winRect(0+nc*30, 0+nc*30, 640+nc*30, 480+nc*30);
	int bits;
	// zjistime aktualni nastaveni barvne hloubky:
	DEVMODE dm;
	EnumDisplaySettings(0, ENUM_CURRENT_SETTINGS, &dm);
    bits = dm.dmBitsPerPel;
	m_camera[n] = new CCameraView(this, winRect, bits, nc);
}

void CChildView::destroyCamera(int camId)
{
	assert(camId>=0 && camId<cMaxCameras);
	delete m_camera[camId];
	m_camera[camId] = 0;
}

void CChildView::setInputDevice(int player, int code, bool remote)
{
	assert(player == 0 || player == 1);
	if (!remote)
		m_inputManager->setInputDevice(player, code);
	else
	{		
		m_inputManager->setInputDevice(player, -3); // -3 znamena remotni
		CString txt;	
		assert(m_ss->getClientString(code, txt));
		TRACE("Selected client %i (%s) for player %i\n", code, (const char *)txt, player);

		// nastaveni modelu prijimajich InputEventy:
		assert(m_ss); // musi byt server, kdyz ma pripojene klienty
		m_ss->setLocModel(code, player, (LocalModel *)m_mode->getModel(player));
	}
}

void CChildView::subdivCheck(bool checked) 
{		
	m_quality = (checked==true);

	int oldmode = m_mode->getMyState();
	TRACE("OnSubdivChech: %i, oldmode = %i\n",checked,oldmode);
	assert(oldmode == cSimulate || oldmode == cAnimTest || oldmode == cViewer || 
		oldmode == cReplay);
	initMode(oldmode);
}

void CChildView::OnMainStartjudge() 
{
	if (m_mode->getMyState() != cSimulate)
	{
		MessageBox("Judge can be started only in simulation mode", "Error", MB_OK);
		return;
	}
	((CModeSimulation *)m_mode)->retreat(); // aby se vratili
	((CModeSimulation *)m_mode)->startJudge();
}	

void CChildView::OnButtonSelect(int joy)
{
	CButtonConfigDialog *cbd = new CButtonConfigDialog(this, m_di, joy);	
	cbd->DoModal();
	delete cbd;
}
