/* 
    ColDet.h

	Collision detection for deformable models using DOP trees

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "Model.h"

#ifndef COLDET_H
#define COLDET_H

#define MAX_DOP2 13 // maximalni pocet os (13 -> 26-DOPy)

typedef struct SDOPTree {
	float min[MAX_DOP2], max[MAX_DOP2]; // intervaly na jednotlivych primkach
	int imin, imax; // interval trojuhelniku (m_pTri), ktere obaluje
	struct SDOPTree *left, *right;
} DOPTree;

class CDModel {
private:
	Model *m_model;
	int *m_pTri; // permutace cisel trojuhelniku	
	DOPTree *root;
	int m_k2; // k/2, kde k je to v "k-DOP"
	float m_axis[MAX_DOP2][3]; // osy DOPu
	bool m_weapon;
public:	
	int m_numTri;	
public:
	// priznak 'weapon' znamena, ze se bude uvazovat pouze posledni mesh
	CDModel(Model *m, int k = 6, bool weapon = false); // k-DOPs
	virtual ~CDModel();

	// stavi DOP strom; ocekava vyplnene pole m_model->m_pActVers
	void build();
	// vlastni test kolize
	bool collide(CDModel *m2);
	// test kolize nastavujici cisla kolidujicich trojuhelniku
	bool colTri(CDModel *m2);
	// zmena velikosti DOP pri deformaci modelu
	void refit();
	// smaze strom
	void delTree(DOPTree *v);
	// vykresli AABB v hloubce h (pro experimentalni ucely)
	void draw(int h);
	// nastavi obalovy kvadr korene:
	void getBBox(float min[3], float max[2]);

protected:
	DOPTree *intBuild(int zac, int kon);
	// hleda minimalni DOP
	void getMinMax(int zac, int kon, float min[MAX_DOP2], float max[MAX_DOP2]);
	int prohazej(int zac, int kon, int osa, float delta);
	bool negTri(int t, int osa, float delta);
	void mydraw(DOPTree *v, int h, int ah);
	// DOP-DOP test
	bool boxbox(DOPTree *t1, DOPTree *t2);
	bool tritri(int t1, int t2);
	CDModel *m_model2;	
	//bool *m_tr1, *m_tr2; // pole na zapis kolidujicich trku
	bool mycollide(DOPTree *t1, DOPTree *t2);
	bool mycollide2(DOPTree *t1, DOPTree *t2); // navic zapisuje do poli
	void myrefit(DOPTree *v);

	int ttops, bbops;
};

// test triangle-triangle (z RAPIDu)
int 
tri_contact (float *P1, float *P2, float *P3,
		    float *Q1, float *Q2, float *Q3);

#endif