/* 
    ColResp.cpp

	Collision response evaluation

    Author: Ladislav Kavan, 2002, 2003
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
//#include "myvec.h"
#include "ColResp.h"
#include "Model.h"
#include "Vector.h"
#include <assert.h>
#include "myvec.h"
#include "ikan.h"
#include "geometry.h"
#include "paths.h"

/*
#undef TRACE
void AFX_CDECL disableTrace(LPCTSTR lpszFormat, ...) {}
#define TRACE disableTrace
*/

float max3(float dist1, float dist2, float dist3)
{
	if (dist1>=dist2 && dist1>=dist3) return dist1;
	else if (dist2>=dist1 && dist2>=dist3) return dist2;
	else if (dist3>=dist1 && dist3>=dist2) return dist3;
	else { assert(false); return 0; }
}

float min3(float dist1, float dist2, float dist3)
{
	if (dist1<=dist2 && dist1<=dist3) return dist1;
	else if (dist2<=dist1 && dist2<=dist3) return dist2;
	else if (dist3<=dist1 && dist3<=dist2) return dist3;
	else { assert(false); return 0; }
}

int min3arg(float dist1, float dist2, float dist3)
{
	if (dist1<=dist2 && dist1<=dist3) return 1;
	else if (dist2<=dist1 && dist2<=dist3) return 2;
	else if (dist3<=dist1 && dist3<=dist2) return 3;
	else { assert(false); return 0; }
}

ColResp::ColResp(FencerModel *fmodelA, FencerModel *fmodelB, IKAN *ikanA, IKAN *ikanB)
{
	m_fmodel[0] = fmodelA;
	m_fmodel[1] = fmodelB;
	m_model[0] = fmodelA->m_model;
	m_model[1] = fmodelB->m_model;
	m_weapon[0] = m_model[0]->getWeapon();
	m_weapon[1] = m_model[1]->getWeapon();
	m_cdw[0] = fmodelA->m_cdw;
	m_cdw[1] = fmodelB->m_cdw;
	m_ikan[0] = ikanA;
	m_ikan[1] = ikanB;	
	m_lastTime = -1;
	if (!loadMassProp(0, DATA_PATH + "mecmass.txt"))
		::MessageBox( NULL, "Couldn't load weapon mass properties data/mecmass.txt", "Error", MB_OK | MB_ICONERROR );
	if (!loadMassProp(1, DATA_PATH + "mecmass.txt"))
		::MessageBox( NULL, "Couldn't load weapon mass properties data/mecmass.txt", "Error", MB_OK | MB_ICONERROR );
	//m_mass[0]*=8;
	//m_invIbody[0].mulScal(0.125, m_invIbody[0]);
	for (int n=0; n<2; n++)
	{
		m_minLinvelMagn[n] = 0;
		m_minAngvelMagn[n] = 0;
		for (int m=0; m<3; m++) 
		{
			m_linvel[n][m]=0;
			m_angvel[n][m]=0;			
		}	
	}
	resetDisplacements();
	m_unControlTime = -1;
	m_epsilon = 0.2;
	m_R = 1.0; m_Rn = 0.5; m_Rb = 1.25;
	//m_ZR = 1.0; m_ZRn = 0.5; m_ZRb = 1.25;
	// polomery tel:
	m_Rbody[0] = m_Rbody[1] = 10.0;
	m_RnBody[0] = m_RnBody[1] = 8.0;
	
	m_maxSeparationTrans = 0.5;
	m_odraz = false;
	m_appliedImpulse = false;
	cpvector(m_lastImpulse, zero3);
}
	
ColResp::~ColResp()
{}

// 'go' se vola kazdy animacni krok;
// vraci true pokud doslo ke kolizi zbran/zbran
bool ColResp::GO(AnimSaver *as)
{	
	bool collided = false;
	m_as = as;

//	bool hnuto[2]; // jestli bylo hnuto s SHres (a tudiz bude treba prepocitavat)
	Matrix SHres[2];
	float EPres[2][3];
	float linvel[2][3], angvel[2][3]; // aktualni rychlosti v okamziku kolize
	float newLinvelMagn[2], newAngvelMagn[2]; // velikost aktualni prumerne rychlosti
	Matrix rootNew[2];
	float rootVel[2][3]; // linearni rychlost pohybu rootu
	float timeLeft; // kolik sekund jeste zbyva odsimulovat (kvuli vicenasobnym kolizim)
	Matrix actSHmin[2], actSHnew[2]; // minSH a newSH v prubehu simualce useku m_deltaTime
	Matrix SHodr[2]; // SHord = SHres + pohyb za cas timeLeft

	m_deltaTime = (m_timer.getTime() - m_lastTime)/1000;
	if (m_lastTime==-1) m_deltaTime = -1;	
//	TRACE("m_deltaTime = %f\n", m_deltaTime);
//	m_lastTime = actTime;	//ne ted, ale az na konci - viz. 6.12.02
	for (int h=0; h<2; h++)
	{
		m_SHnew[h].set(m_model[h]->getSHF());
		cpvector(EPres[h], m_ikan[h]->elbowPos);	
		rootNew[h].set(m_model[h]->m_pJoints[0].m_final.getMatrix());
	}
	float tCol;
	// aktualni prumernou rychlost musim spocitat z celeho kousku, ne jen z toho pred kontaktem (ten muze byt libovolne maly!)
	for (h=0; h<2; h++)
	{
		// aktualni prumerne rychlosti:
		getLinVelocity(h, m_SHmin[h], m_SHnew[h], m_deltaTime, linvel[h]);
		getAngVelocity(m_SHmin[h], m_SHnew[h], m_deltaTime, angvel[h]);
		// vektory rychlosti rozdelim na velikost a smer:
		newLinvelMagn[h] = norm(linvel[h]);
		newAngvelMagn[h] = norm(angvel[h]);
		unitize(linvel[h]);
		unitize(angvel[h]);
		// a jeste rychlost pohybu rootu:			
		float tmp[3];
		m_rootMin[h].getTranslation(tmp);
		rootNew[h].getTranslation(rootVel[h]);
		Vector::sub3(rootVel[h], tmp);
		Vector::mul3(rootVel[h], 1/m_deltaTime);
		//TRACE("rootVel: %f %f %f\n", rootVel[0], rootVel[1], rootVel[2]);			
	}		
	int typKolize = -1;
	if (m_deltaTime == -1) tCol = -1; // pri prvnim spusteni nesmi kolidovat:
	else tCol = findCollision(m_SHmin, m_SHnew, &typKolize);
	//if (m_unControlTime != -1) tCol = -1; // zrusim kolizi behem kolize - pro ladeni
	//m_timeEstimate = m_deltaTime; // odhaduji, ze pristi snimek bude trvat totez co minuly
	m_appliedImpulse = false;	
	if (tCol != -1)
	{ // mame hlasenou kolizi
		collided = true;
		// spocitam velikosti aktualni okamzite rychlosti (linterpolace, viz. 5.12.02)
		float actLinvelMagn[2], actAngvelMagn[2];
		for (h=0; h<2; h++)
		{
			float t = tCol;
			//t=1.0; // rusi interpolaci rychlosti, pouziji se nove prumerne rychlosti
			actLinvelMagn[h] = (1-t)*m_minLinvelMagn[h] + t*newLinvelMagn[h];
			actAngvelMagn[h] = (1-t)*m_minAngvelMagn[h] + t*newAngvelMagn[h];
			// a dam jednotkovym vektorum linvel, angvel spravnou velikost:
			Vector::mul3(linvel[h], actLinvelMagn[h]);
			Vector::mul3(angvel[h], actAngvelMagn[h]);
		}

		// jestli budeme modelovat odraz je v prom. m_odraz
		actSHmin[0].set(m_SHmin[0].getMatrix());
		actSHmin[1].set(m_SHmin[1].getMatrix());
		actSHnew[0].set(m_SHnew[0].getMatrix());
		actSHnew[1].set(m_SHnew[1].getMatrix());
		timeLeft = m_deltaTime;
		int cnt=1;
		// odsimuluji vsechny kolize v case timeLeft:
		do
		{
			//minDist = getMinDist(W0W1, actSHmin, actSHnew); // TODO: zrusit, je to jen pro ladeni
			rollBack(tCol, actSHmin, actSHnew, SHres, EPres);			
			timeLeft = timeLeft*(1-tCol); // ukousnu cas nez doslo ke kolizi			
			TRACE("\nkolize %i, typ = %i: zbyva odsimulovat cas %f z %f\n", cnt++, typKolize, timeLeft, m_deltaTime);
//			TRACE("linvel[0] = %f %f %f, angvel[0] = %f %f %f, linvel[1] = %f %f %f, angvel[1] = %f %f %f\n", 
//				linvel[0][0], linvel[0][1], linvel[0][2], angvel[0][0], angvel[0][1], angvel[0][2],
//				linvel[1][0], linvel[1][1], linvel[1][2], angvel[1][0], angvel[1][1], angvel[1][2]);


			// zapsat zmeny do stavu modelu:
			for (h=0; h<2; h++)
			{
				Matrix or, inv; float pos[3];
				or.set(SHres[h].getMatrix());
				m_fmodel[h]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(inv);
				or.preMultiply(inv); // podle konvence do stavu modelu ma prijit matice vzhledem k root.final
				or.getTranslation(pos);
				or.setTranslation(zero3);
				m_fmodel[h]->updateModelState(or, pos, EPres[h]);
			}
			as->saveState(1); // a zasejvovat
			/**/
			

			if (applyImpulse(SHres, typKolize, linvel, angvel, m_linvel, m_angvel))
			{ // kolize potvrzena:														
				m_appliedImpulse = true;
				// do SHodr dam SHres plus pohyb s prave spocitanymi rychlostmi za cas timeLeft:
				for (int h=0; h<2; h++)
				{
					SHodr[h].set(SHres[h].getMatrix());
					applyVelocity(h, m_linvel[h], m_angvel[h], SHodr[h], timeLeft);
				}
				//TRACE("minDist po impulsu = %f\n", getMinDist(W0W1, SHres, SHodr));
				
				// jestlipak reakce nevede k dalsi kolizi?
				tCol = findCollision(SHres, SHodr, &typKolize);
				if (tCol == -1)
				{ // odraz nevede bezprostredne k dalsi kolizi, budeme koncit a pak simulovat odleteni mece:
					m_odraz = true;					
				}
				else
				{ // budeme pokracovat v cyklu:					
					for (int h=0; h<2; h++)
					{
						// rychlosti po kolizi jsou ted vstupnimi rychlostmi:
						cpvector(linvel[h], m_linvel[h]);
						cpvector(angvel[h], m_angvel[h]);
						// do actSHmin dam matici k niz jsme dorollBackovali:
						actSHmin[h].set(SHres[h].getMatrix());
						// do actSHnew zas matici ke ktere jsem se ted chtel dostat (ale neproslo to):
						actSHnew[h].set(SHodr[h].getMatrix());
					}
					m_odraz = false; // jen kvuli warningu					
				}
			}
			else // applyImpulse vraci resting/retrating contact:
			{ 		
				//if (cnt==2) assert(false&&"jen tak");
				/*
				assert(minDist >= 2*m_Rn); // velka kolize & resting contact = chyba !
				// necham puvodni pohyb, i kdyz byl (malo) kolidujici:
				SHres[0].set(actSHnew[0].getMatrix());
				SHres[1].set(actSHnew[1].getMatrix());
				*/
				if (!separatePerPartes(SHres, EPres, linvel, angvel, timeLeft)) // i kdyz pro male kolize je to kanon na vrabce
				{
					// nepovedlo se separovat, vsechno zpet:
					SHres[0].set(m_SHmin[0].getMatrix());
					SHres[1].set(m_SHmin[1].getMatrix());
				}
				m_odraz = false;				
				break; // klouze, nebudou dalsi kolize
			}		
		}
		while (tCol != -1);
		// simulaci za cas m_deltaTime mam vyrizenou; co bude dal?:
		if (m_odraz)
		{ // mame modelovat odraz s rychlostmi m_linvel a m_angvel od pozice SHodr:
			// nastavim displacementy:
			m_unControlTime = 1.0; // compDeflection to chce mit nastavene
			float linvelRel[2][3];
			for (h=0; h<2; h++)
			{
				// SHodr si nakopiruju do SHres (kvuli invariantu):
				SHres[h].set(SHodr[h].getMatrix());			
				// odectu rychlost rootu od vysledne lin. rychlosti zbrane:
				cpvector(linvelRel[h], m_linvel[h]);
				Vector::sub3(linvelRel[h], rootVel[h]);
				//compDeflection(h, SHodr[h], linvelRel, m_angvel[h], m_noDef[h], m_maxDef[h]);
			}				
			compDeflection(SHodr, linvelRel, m_angvel, m_noDef, m_maxDef, m_EPnoDef, m_EPmaxDef);
			cpvector(EPres[0], m_EPnoDef[0]);
			cpvector(EPres[1], m_EPnoDef[1]);
			m_sinceCol.reset();				
		}						
	}
	else
	{ // nebyla kolize, jen okopirovat m_SHnew do SHres kvuli invariantu:
		SHres[0].set(m_SHnew[0].getMatrix());
		SHres[1].set(m_SHnew[1].getMatrix());
		//hnuto = false; // SHres je pouze m_SHnew
		m_odraz = false; // samozrejme
	}
	// invariant: v SHres mam vyslednou polohu SHF (at uz se s ni hnulo nebo ne)
	// zavolam separaci (vraci jestli pohla s SHres):
	if (!separate(SHres))
	{
		// nepovedlo se separovat, vsechno zpet:
		SHres[0].set(m_SHmin[0].getMatrix());
		SHres[1].set(m_SHmin[1].getMatrix());
		// pripadny odraz necham, proc ne?
	}	

/*
	// otestuju pomoci normy matic, jestli se nejaky zmenily oproti 
	float nor[2];
	for (h=0; h<2; h++)
	{
		nor[h] = m_SHmin[h].l2norm(SHres[h]);
		if (m_SHmin[h].l2norm(SHres[h]) < GEOM_EPSILON) hnuto[h] = false;
		else hnuto[h] = true;
	}
	TRACE("norma[0] = %f, norma[1] = %f\n", nor[0], nor[1]);
*/

/*	{ // testik vzdalenosti zbrane od sermire:
		Matrix inv, relSHres;
		m_fmodel[0]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(inv);
		relSHres.set(SHres[0].getMatrix());
		relSHres.preMultiply(inv);
		float dist = bodyWeaponDist(0, 0, relSHres, relSHres, false);
		TRACE("Vzdalenost modelu 0 od jeho mece: %f\n", dist);
	}
/**/

//	if (hnuto[0] || hnuto[1])
//	{ // bylo hnuto se SHF (SHres), musim zavolat IK:
		updateArm(SHres, EPres);
//	}
	// ulozim minulost:
	m_lastTime = m_timer.getTime();	// zjednodusujici predpoklad ze vse v teto funkci vyse trva zanedbatelny cas (viz. 6.12.02)
	for (h=0; h<2; h++)
	{
		// velikosti minulych prumernych rychlosti (s pripadnou aktualni kolizi):
		getLinVelocity(h, m_SHmin[h], SHres[h], m_deltaTime, linvel[h]);
		getAngVelocity(m_SHmin[h], SHres[h], m_deltaTime, angvel[h]);		
		m_minLinvelMagn[h] = norm(linvel[h]);
		m_minAngvelMagn[h] = norm(angvel[h]);
		// minule SHF, Elpos a rootpos:
		m_SHmin[h].set(SHres[h].getMatrix()); // viz. invariant
		cpvector(m_oldElPos[h], EPres[h]);
		m_rootMin[h].set(rootNew[h].getMatrix());		
	}
	// jeste zapsat zmeny do stavu modelu:
	for (h=0; h<2; h++)
	{
		Matrix or, inv; float pos[3];
		or.set(SHres[h].getMatrix());
		m_fmodel[h]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(inv);
		or.preMultiply(inv); // podle konvence do stavu modelu ma prijit matice vzhledem k root.final
		or.getTranslation(pos);
		or.setTranslation(zero3);
		m_fmodel[h]->updateModelState(or, pos, EPres[h]);
	}
	// prepocitani vrcholu se uskutecni ve volajici funkci
	return collided;
}

// vezme dulezite vrcholy (A,D,E,F) z m_weapon[h]->m_pVertices, transformuje 'cinterp'
// a vysledek zapise do m_weapon[h]->m_pActVers
void ColResp::transformVertices(int h, const Matrix cinterp)
{
	int tot = m_weapon[h]->m_numVertices;
	// transformuju vrcholy A a D
	Vector Apos, Dpos;

	Apos.set(m_weapon[h]->m_pVertices[tot-vA].m_location); // z Acka vezmu
	Apos.transform( cinterp ); // transformuju
	if (m_weapon[h]->m_leftHanded) Apos.mul(1, 1, -1);	
	Vector::set3( m_weapon[h]->m_pActVers[tot-vA].m_location, Apos.getVector() ); // do Acka zapisu
		
	Dpos.set(m_weapon[h]->m_pVertices[tot-vD].m_location); // z Dcka vezmu
	Dpos.transform( cinterp ); // transformuju
	if (m_weapon[h]->m_leftHanded) Dpos.mul(1, 1, -1);			
	Vector::set3( m_weapon[h]->m_pActVers[tot-vD].m_location, Dpos.getVector() ); // do Dcka zapisu
	
	Vector Epos, Fpos; // jeste zastitu (11.2.2003)

	Epos.set(m_weapon[h]->m_pVertices[tot-vE].m_location);
	Epos.transform( cinterp ); // transformuju
	if (m_weapon[h]->m_leftHanded) Epos.mul(1, 1, -1);			
	Vector::set3( m_weapon[h]->m_pActVers[tot-vE].m_location, Epos.getVector() );

	Fpos.set(m_weapon[h]->m_pVertices[tot-vF].m_location);
	Fpos.transform( cinterp ); // transformuju
	if (m_weapon[h]->m_leftHanded) Fpos.mul(1, 1, -1);			
	Vector::set3( m_weapon[h]->m_pActVers[tot-vF].m_location, Fpos.getVector() ); // do Dcka zapisu
}

// nastavi m_weapon[h]->m_pActVers[A, D, E, F] na hodnoty zbrane interpolovane mezi polohami SHmin a SHnew
// interpoluje se s parametrem t bud normalne, nebo proti smeru 'obracene'
// je-li interp neNULL, tak do nej zapise vyslednou interpolovanou matici v IKANskych konvencich:
void ColResp::interpWeapon(int h, float t, const Matrix SHmin, const Matrix SHnew, 
						   ikMatrix interp, bool obracene)
{
	assert(0<=t && t<=1);	
	
	ikMatrix iSHmin, iSHnew, iinterp;
	Matrix cinterp, cSHmin, cSHnew;
	// napocitam si interpolovane matice
	
	// 1) posunu matice do hmotneho stredu (dulezite, protoze uhlova rychlost taky pusobi vzhledem k HmS)
	cSHmin.set(SHmin.getMatrix());
	cSHnew.set(SHnew.getMatrix());
	SHF2SHC(h, cSHmin);
	SHF2SHC(h, cSHnew);
	// 2) interpoluju mezi posunutyma maticema:
	Matrix2ikMatrix(cSHmin, iSHmin);
	Matrix2ikMatrix(cSHnew, iSHnew);
	if (obracene) linterp2ikMatrix(iinterp, iSHmin, iSHnew, t);
	else linterpikMatrix(iinterp, iSHmin, iSHnew, t);
	ikMatrix2Matrix(iinterp, cinterp);
	// 3) vratim zpatky ze HmS do SHF:
	SHC2SHF(h, cinterp);
	if (interp) Matrix2ikMatrix(cinterp, interp);	 

	transformVertices(h, cinterp);
}

// 'us' urcuje dve usecky, u1 a u2. Do 'v1[i]' a 'v2[i]' se zapisi aktualni
// koncove body usecky ui
void ColResp::setSegVertices(int us, float v1[2][3], float v2[2][3])
{
	int tot[2] = {m_weapon[0]->m_numVertices,
		   	      m_weapon[1]->m_numVertices};
	if (us == W0W1) // cepel0 - cepel1
	{		
		cpvector(v1[0], m_weapon[0]->m_pActVers[tot[0]-vA].m_location);
		cpvector(v2[0], m_weapon[0]->m_pActVers[tot[0]-vD].m_location);
		cpvector(v1[1], m_weapon[1]->m_pActVers[tot[1]-vA].m_location);
		cpvector(v2[1], m_weapon[1]->m_pActVers[tot[1]-vD].m_location);
	}
	else if (us == W0Z1) // cepel0 - zastita1
	{
		cpvector(v1[0], m_weapon[0]->m_pActVers[tot[0]-vA].m_location);
		cpvector(v2[0], m_weapon[0]->m_pActVers[tot[0]-vD].m_location);
		cpvector(v1[1], m_weapon[1]->m_pActVers[tot[1]-vE].m_location);
		cpvector(v2[1], m_weapon[1]->m_pActVers[tot[1]-vF].m_location);
	}
	else if (us == Z0W1) // zastita0 - cepel1
	{
		cpvector(v1[0], m_weapon[0]->m_pActVers[tot[0]-vE].m_location);
		cpvector(v2[0], m_weapon[0]->m_pActVers[tot[0]-vF].m_location);
		cpvector(v1[1], m_weapon[1]->m_pActVers[tot[1]-vA].m_location);
		cpvector(v2[1], m_weapon[1]->m_pActVers[tot[1]-vD].m_location);
	}
	else assert(false);
}

// vraci minimalni vzdalenost vsech 3 dvojic usecek interpolovanych mezi SHmin a SHnew 
// v bode t kolem HmS; do interp nastavuje (jsou-li neNULL) vysledne swordholder.finaly
// 'us' rika ktere usecky (W0W1, ...)
// side effect: v modelu (m_pActVers) nastavuje vrcholy A, D, E, F na nove hodnoty -- to je jinak zbytecne
// vraci vzdalenost usecek v case t a v 'typ' ktera dvojice usecek koliduje
float ColResp::interpSegDist(float t, const Matrix SHmin[2], const Matrix SHnew[2], int *typ, ikMatrix interp0, ikMatrix interp1)
{
	float dummy[2][3];
	assert(0<=t && t<=1);	

	interpWeapon(0, t, SHmin[0], SHnew[0], interp0);
	interpWeapon(1, t, SHmin[1], SHnew[1], interp1);
	
	float v1[2][3], v2[2][3];
	setSegVertices(W0W1, v1, v2);
	float dist1 = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy[0], dummy[1]);
	setSegVertices(W0Z1, v1, v2);
	float dist2 = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy[0], dummy[1]);
	setSegVertices(Z0W1, v1, v2);
	float dist3 = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy[0], dummy[1]);
	if (typ) *typ = min3arg(dist1, dist2, dist3);
	return min3(dist1, dist2, dist3);
	// kdyby to bylo potreba, mohla by se vracet tez informace o typu
}

// vraci minimalni vzdalenost mezi interpolovanymi useckami (pro ladici ucely):
float ColResp::getMinDist(const Matrix SHmin[2], const Matrix SHnew[2])
{
	ikMatrix intSH[2]; Matrix mySH[2];



	float minDist = 1000000;
// hruba sila, protoze to chci mit presne:
	float t, dist=0, step = 0.01;
	for (t=0; t<1; t+=step)
	{
		dist = interpSegDist(t, SHmin, SHnew, 0, intSH[0], intSH[1]);
		if (dist<minDist) minDist = dist;				

/*
				// jeste zapsat zmeny do stavu modelu:
	for (int h=0; h<2; h++)
	{
		ikMatrix2Matrix(intSH[h], mySH[h]);
		Matrix or, inv; float pos[3];
		or.set(mySH[h].getMatrix());
		m_fmodel[h]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(inv);
		or.preMultiply(inv); // podle konvence do stavu modelu ma prijit matice vzhledem k root.final
		or.getTranslation(pos);
		or.setTranslation(zero3);
		m_fmodel[h]->updateModelState(or, pos, m_oldElPos[h]);
	}
	m_as->saveState(3); // a zasejvovat
/**/
	}
	return minDist;
}

// vrati maximalni okamzitou rychlost na primce dane vektorem u a pary rmin, rmax (19.12.02):
float maxVel(const float linvel[3], const float angvel[3], const float u[3], float rmin, float rmax)
{
	float ANGxU[3], K1, v2;
	crossproduct(ANGxU, angvel, u);
	K1 = DOT(ANGxU, ANGxU);
	if (K1 < GEOM_EPSILON)
	{
		// rotacni slozka rychlosti je prilis mala, vratime linearni:
		return norm(linvel);
	}
	// pocitame maximum na usecce (rmin, rmax):
	v2 = DOT(linvel, linvel);
	



	// oprava 10.2.03: je potreba, aby horni odhad platil pro kazdou polohu zbrane/usecky!
	float UBVmin = sqrt(rmin*rmin*K1) + sqrt(v2);
	float UBVmax = sqrt(rmax*rmax*K1) + sqrt(v2);
	float UB;

	if (UBVmin > UBVmax) UB = UBVmin;
	else UB = UBVmax;
	//TRACE("UBVmin = %f, UBVmax = %f, UB = %f\n", UBVmin, UBVmax, UB);
	return UB;

/* puvodni verze (podle 19.12.02), ktera delala jenom odhad vzhledem k okamzite poloze:

	float K2 = 2*DOT(ANGxU, linvel);
	// kvadrat rychlosti v bode r je nyni = v2 + r*K2 + K1*r*r
	float VvRmin = v2 + rmin*K2 + K1*rmin*rmin; // kvadrat rychlosti v bode rmin
	float VvRmax = v2 + rmax*K2 + K1*rmax*rmax; // kvadrat rychlosti v bode rmax
	float Vmax; // ta vetsi rychlost:
	if (VvRmin > VvRmax) Vmax = VvRmin;
	else Vmax = VvRmax;
	// r tz. rmin<r<rmax s lokalnim extremem:
	float rder = -K2/(2*K1);
	// TODO: toto zrusit, rder je lokalni minimum !!!
	if (rder>rmin && rder<rmax)
	{ // musim uvazit lokalni extrem:		
		float VvRder = v2 + rder*K2 + K1*rder*rder; // kvadrat rychlosti v bode rder
		//TRACE("rder = %f, VvRder = %f\n", rder, VvRder);
		if (VvRder > Vmax) 
		{
			assert(false && "maxima se nabyva v lokalnim minimu?!");
			Vmax = VvRder;
		}
	}
	//TRACE("VvRmin = %f, VvRmax = %f, Vvysledna = %f\n", sqrt(VvRmin), sqrt(VvRmax), sqrt(Vmax));
	TRACE("Horni odhad = %f, stary (lokalni) odhad = %f\n", UB, sqrt(Vmax));
	assert(UB >= sqrt(Vmax));
	return sqrt(Vmax); // odmocnit z kvadratu!
*/
}

// vrati maximalni okamzitou rychlost zbrane pohybujici se s danymi rychlostmi v poloze SHF:
// pozor, aby to bylo korektni, je treba aby vzdalenost krajnich bodu zastity od HmS byla
// mensi nez vzdalenost hlavice od HmS (protoze body na zastite se neuvazuji)
float ColResp::weaponMaxVel(int h, const float linvel[], const float angvel[], const Matrix SHF)
{
	// spocitat vektor u ve smeru usecky a pary usecky rmin, rmax:
	float rmin, rmax, u[3];
	int tot = m_weapon[h]->m_numVertices;		   	      
	float Apos[3], Dpos[3], Adis[3], Ddis[3];

	// polohy bodu A, D v souradnicich modelu:
	cpvector(Apos, m_weapon[h]->m_pVertices[tot-vA].m_location);
	cpvector(Dpos, m_weapon[h]->m_pVertices[tot-vD].m_location);
	// vzdalenost A od HmS (A je spicka):
	cpvector(Adis, Apos);
	Vector::sub3(Adis, m_com[h]);
	rmax = norm(Adis);
	// vzdalenost D od HmS (D je hlavice):
	cpvector(Ddis, Dpos);
	Vector::sub3(Ddis, m_com[h]);
	rmin = -norm(Ddis);
	// vektor u: jednotkovy ve smeru usecky (od rmin do rmax)
	cpvector(u, Apos);
	Vector::sub3(u, Dpos);
	unitize(u);
	// Pozn.: vzhledem ke konvenci zbrane by ted u melo byt (0, 0, -1)
	// u je potreba natocit podle aktualni polohy zbrane (rmin, rmax se otocenim nezmeni):
	Vector uvect(u);
	uvect.transform3(SHF);
	uvect.get(u);

	// vratim maximalni okamzitou rychlost zbrane:	
	return maxVel(linvel, angvel, u, rmin, rmax);
}

// spocita maximalni velikost casu deltaT za ktery se valcoidy priblizi maximalne o R-Rn (19.12.02):
float ColResp::maxStep(const float linvel[2][3], const float angvel[2][3], const Matrix SHF[2])
{
	float mVel[2]; // maximalni okamzite rychlosti zbrani
	for (int hrac=0; hrac<2; hrac++)
		mVel[hrac] = weaponMaxVel(hrac, linvel[hrac], angvel[hrac], SHF[hrac]);

	if (mVel[0] + mVel[1] < GEOM_EPSILON) return 1.0; // max. okamz. rychlosti jsou tak male, ze nemusim nic trasovat
	float deltaT = (m_R - m_Rn) / (mVel[0] + mVel[1]);
	return deltaT;
}

// hleda okamzik kontaktu mezi SHmin a SHnew; pokud neexistuje vraci -1:
// do minDist nastavuje minimalni nalezenou vzdalenost (pro ladeni)
// vstupni predpoklad: v m_SHmin neni kolize
// v 'typ' vraci typ kolize (W0W1, ...)
float ColResp::findCollision(const Matrix SHmin[2], const Matrix SHnew[2], int *typ)
{
	float t, tb = -1, tc = -1, dist=0, step;
// spoctu si lin. a uhlove rychlosti za predpokladu ze pohyb z SHmin do SHnew trval 1 sekundu:
	float linvel[2][3], angvel[2][3];
	for (int h=0; h<2; h++)
	{
		getLinVelocity(h, SHmin[h], SHnew[h], 1.0, linvel[h]);
		getAngVelocity(SHmin[h], SHnew[h], 1.0, angvel[h]);
	}
	step = maxStep(linvel, angvel, SHmin);
//	TRACE("findCollision: step = %f -> %f kroku\n", step, 1/step);

	for (t=0; t<1; t+=step)
	{
		dist = interpSegDist(t, SHmin, SHnew, typ);
		if (dist<=2*m_R) 
		{
			tb = t-step;
			tc = t;
			break;
		}
	}	
	if (tb==-1) 
	{ // nenasli jsme kolizi:		
		/* testik s presnejsim krokem:
		step = step/10;		
		for (t=0; t<1; t+=step)
		{
			dist = interpSegDist(t, SHmin, SHnew);
			if (dist<=2*m_R) 
			{
				tb = t-step;
				tc = t;
				break;
			}
		}
		if (tb != -1)
		{
			TRACE("puvodni step hlasil nekolizi, presnejsi dava kolizi s dist = %f\n", dist);
			assert(dist >= m_Rn);
			assert(false && "toto neni chyba, jen upozorneni");
		}
		/**/
		return -1;		
	}
	if (tb<0) 
	{
		TRACE("podezrely: kolize v 0, dist v 0 je %f\n", interpSegDist(0, SHmin, SHnew));
	}
	TRACE("collision: tb = %f, dist in tb= %f, tc = %f, dist in tc= %f\n", tb, 
		interpSegDist(tb, SHmin, SHnew), tc, interpSegDist(tc, SHmin, SHnew));
// binarne dohledam presny okamzik kolize
	assert(interpSegDist(tb, SHmin, SHnew)>2*m_R);
	assert(interpSegDist(tc, SHmin, SHnew)<=2*m_R);

	float b0=tb, b1=tc;
	int iter = 30;	
	while (iter-- > 0)
	{
		float bs = (b0+b1)/2;		
		dist = interpSegDist(bs, SHmin, SHnew, typ);

		if (dist > 2*m_R) b0 = bs; // nedoslo ke kolizi
		else b1 = bs;
		if (fabs(dist-2*m_R)<GEOM_EPSILON) 
		{
			//TRACE("collision in %f, dist = %f\n", b0, interpSegDist(b0));
			return b0; // b0 dulezite
		}
	} // invariant: b0 je vzdy bez kolize	
	TRACE("podezrely, nepodarilo se presne urcit kontakt; dist = %f\n", dist);
	return b0;
}

// tc je okamzik kontaktu; nastavuje vysledny SHF (SHres_ a ElPos: (pozor na kumulaci chyby)
// jako side effect taky nastavuje m_pActVers[A] a D na nove hodnoty
void ColResp::rollBack(float tc, const Matrix SHmin[2], const Matrix SHnew[2], Matrix SHres[2], float ElPosRes[2][3])
{
	ikMatrix intSH[2]; Matrix mySH[2];
	float dist = interpSegDist(tc, SHmin, SHnew, 0, intSH[0], intSH[1]); // (*)
	//TRACE("dist = %f\n", dist);	
	assert(dist>2*m_R); // kontakt je bez kolize!
	// do SHres zapisu interpolovany SHF, abych se vyhnul kumulaci chyby IK (17.9.02)
	ikMatrix2Matrix(intSH[0], SHres[0]);ikMatrix2Matrix(intSH[1], SHres[1]);

	// interpoluju jeste pozici lokte (Pozor! pouzivam m_ikan[h]->elbowPos !)
	for (int h=0; h<2; h++)
		vecinterp(ElPosRes[h], m_oldElPos[h], m_ikan[h]->elbowPos, tc);	
}

// vola IK ktera propaguje zmeny v SHF a pozice lokte (elPos) do cele paze
// a taky prepocita vrcholy A, D v modelu:
void ColResp::updateArm(const Matrix SHF[2], const float elPos[2][3])
{
	for (int h=0; h<2; h++)
	{	
		cpvector(m_ikan[h]->elbowPos, elPos[h]);
		// pustim IK:		
		Matrix EEor(SHF[h]);		
		float nul[3] = {0,0,0}; EEor.setTranslation(nul);
		SHF[h].getTranslation(nul);
		Vector EEpos(nul);		
		m_ikan[h]->setGoal(EEpos, EEor, m_ikan[h]->elbowPos);
		m_fmodel[h]->setFinalMatrices();				
	}
	// nastavim minule vrcholy A, D (u obou modelu)
	// (compActVers nasledujici po ColResp.go to pak nacpe do B, C)
	interpSegDist(0, m_SHmin, m_SHmin); // (volam jen kvuli side-effectu)
}

// nacte ze souboru hmotnost, teziste a tensor setrvacnosti
bool ColResp::loadMassProp(int h, const char *fname)
{
	FILE *f = fopen(fname, "rt");
	if (!f) return false;
	fscanf(f, "Mass: %f\n", &m_mass[h]);
	fscanf(f, "CoM: %f %f %f\n", &m_com[h][0], &m_com[h][1], &m_com[h][2]);
	float Im[16];
	for (int n=0; n<16; n++) Im[n] = 0;
	Im[15] = 1;
	fscanf(f, "Ixx: %f\n", &Im[0]);
	fscanf(f, "Iyy: %f\n", &Im[1 + 1*4]);
	fscanf(f, "Izz: %f\n", &Im[2 + 2*4]);
	fscanf(f, "Ixy: %f\n", &Im[1]);
	fscanf(f, "Izx: %f\n", &Im[2]);
	if (feof(f)) return false;
	fscanf(f, "Izy: %f\n", &Im[2+4]);
	// symetrie:
	Im[1*4] = Im[1];
	Im[2*4] = Im[2];
	Im[2*4+1] = Im[2+1*4];	
	m_invIbody[h].set(Im);
	bool inv;
	inv = m_invIbody[h].invert();
	assert(inv); // tensor setrvacnosti nesmi byt singularni (to by znamenalo netrojrozmerne teleso)
	fclose(f);
	return true;
}

// posune translacni cast matice z SHF do stredu hmotnosti (pouziva m_com):
void ColResp::SHF2SHC(int h, Matrix &shf)
{	
	float comr[3] = {m_com[h][0], m_com[h][1], m_com[h][2]};	
	shf.RotateVect(comr);
	shf.addTranslation(comr);
}

// posune translacni cast matice ze stredu hmotnosti zbrane do pocatku, tj. SHF (pouziva m_com):
void ColResp::SHC2SHF(int h, Matrix &shf)
{
	float comr[3] = {m_com[h][0], m_com[h][1], m_com[h][2]};
	shf.RotateVect(comr);
	Vector::mul3(comr, -1);
	shf.addTranslation(comr);
}

// nastavi linearni rychlost zbrane v [cm/s]; pouziva m_com:
void ColResp::getLinVelocity(int h, const Matrix &mSHmin, const Matrix &mSHnew, 
							 float m_deltaTime, float vel[3])
{	
	Matrix posSHmin(mSHmin), posSHnew(mSHnew);
	SHF2SHC(h, posSHmin); SHF2SHC(h, posSHnew); // prevedu na frejm teziste	
	posSHnew.getTranslation(vel);
	posSHmin.inverseTranslateVect(vel);
	Vector::mul3(vel, 1/m_deltaTime);
}

// nastavi uhlovou rychlost v [rad/s]:
void ColResp::getAngVelocity(const Matrix &mSHmin, const Matrix &mSHnew, float m_deltaTime, 
							 float omega[3])
{
	float zero[3] = {0,0,0};
	Matrix SHmin(mSHmin), SHnew(mSHnew);
	Matrix invSHmin;
	SHmin.setTranslation(zero); SHnew.setTranslation(zero);
	// posouvat netreba (dostal bych stejny vysledek)
	SHmin.inverseRotateMatrix(invSHmin);
	SHnew.postMultiply(invSHmin);
	float alfa;
	SHnew.axisangle(omega, alfa);
	Vector::mul3(omega, alfa/m_deltaTime);
}

// transformuje SHF tak jako by se pohyboval rychlostmi linvel a angvel po dobu deltaTime:
void ColResp::applyVelocity(int h, const float linvel[3], const float angvel[3], Matrix &SHF, 
							float deltaTime)
{	
	SHF2SHC(h, SHF);
	float linpos[3]; cpvector(linpos, linvel);
	Vector::mul3(linpos, deltaTime);
	SHF.addTranslation(linpos);
	float respos[3], zero[3] = {0, 0, 0};
	SHF.getTranslation(respos);
	SHF.setTranslation(zero);
	float axis[3]; cpvector(axis, angvel);
	float alfa = norm(axis);
	unitize(axis);
	alfa *= deltaTime;
	Matrix rot;
	rot.setRotationAxisAngle(axis, alfa);
	SHF.preMultiply(rot);
	SHF.setTranslation(respos);
	SHC2SHF(h, SHF);
}

// vynuluje displacementy:
void ColResp::resetDisplacements()
{
	for (int n=0; n<2; n++)
		for (int m=0; m<3; m++) 
			m_linDisplac[n][m] = 0;		
	for (n=0; n<2; n++) m_angDisplac[n].loadIdentity();
}

// ma za ukol vyrobit m_angDisplac a m_linDisplac na zaklade casu od kolize 'timeSinceCol'
// pouziva matice m_noDef a m_maxDef a celkovy cas reakce m_unControlTime
// diplacementy jsou (stejne jako m_noDef a m_maxDef) relativni vzhledem k root.finalu
void ColResp::addMovement(int h, float timeSinceCol)
{
	// viz. 17.10.02 a 18.10.02
	// spocitam interpolacni parametr p mezi m_noDef a m_maxDef:
	float p = 1 - 1/((20*timeSinceCol)/m_unControlTime + 1);
	m_noDef[h].SLERP(m_angDisplac[h], m_maxDef[h], p);
	// prevod na displacementy:
	m_angDisplac[h].getTranslation(m_linDisplac[h]);
	m_angDisplac[h].setTranslation(zero3);
	// jeste interpoluju pozici lokte:
	vecinterp(m_elDisplac[h], m_EPnoDef[h], m_EPmaxDef[h], p);
}

// vraci vzdalenost tela hrace h a jeho zbrane; telo i zbran jsou aprox. valci, zbran je na ceste
// od SHzac do SHkon v delce 't' (21.12.02):
// (SHzac a SHkon jsou relativni vzhledem k root.finalu)
float ColResp::bodyWeaponDist(int h, float t, const Matrix SHzac, const Matrix SHkon, bool obracene)
{
	// osa valce prochazi root jointem, ktery je (vzhledem k sobe) v pocatku
	// udelam hrozne dlouhy valec; podle konvenci miri osa Y od rootu smerem k hlave:
	// 2 body urcujici osu valce, ktery aproximuje telo:
	float bodyCylA[3] = {0, 1000, 0}, bodyCylD[3] = {0, -1000, 0}; 
	
	int tot = m_weapon[h]->m_numVertices;
	interpWeapon(h, t, SHzac, SHkon, NULL, obracene);
	
	float weaponCylA[3], weaponCylD[3]; // 2 body urcujici osu valce, ktery aproximuje zbran
	// m_pActVers bylo nastaveno interpWeaponem:
	cpvector(weaponCylA, m_weapon[h]->m_pActVers[tot-vA].m_location);
	cpvector(weaponCylD, m_weapon[h]->m_pActVers[tot-vD].m_location);
	float dummy[2][3];
	return distSegSeg(bodyCylA, bodyCylD, weaponCylA, weaponCylD, dummy[0], dummy[1]);
}

// testuje jestli zbran hrace h protne jeho telo pri ceste ('obracene') mezi SHzac a SHkon:
bool ColResp::bodyWeaponCol(int h, const Matrix SHzac, const Matrix SHkon, bool obracene)
{
	float minDist = 1000000;
	// rychlosti zbrane pokud pohyb trva 1 sekundu:
	float linvel[3], angvel[3];
	getLinVelocity(h, SHzac, SHkon, 1.0, linvel);
	getAngVelocity(SHzac, SHkon, 1.0, angvel);
	// maximalni okamzita rychlost zbrane:
	float maxVel = weaponMaxVel(h, linvel, angvel, SHzac);
	float step;
	if (maxVel < GEOM_EPSILON) step = 1.0; // max. okamz. rychlost je tak mala, ze nemusim nic trasovat
	else step = (m_Rbody[h] - m_RnBody[h]) / maxVel; // 21.12.02
//	TRACE("bodyWeaponCol, hrac %i, step = %f -> %f kroku\n", h, step, 1/step);
	for (float t=0; t<1; t+=step)
	{
		float dist = bodyWeaponDist(h, t, SHzac, SHkon, obracene);
		if (dist < minDist) minDist = dist;
	}
	//if (!obracene && h==0) TRACE("MODEL0, primy pruchod, hrac %i: minimalni vzdalenost mece od tela = %f\n", h, minDist);
	if (minDist < m_Rbody[h] + m_R) return true; // protly se
	return false; // neprotly se 
}

// prida k orientaci&posunuti zbrane a k pozici lokte (vse vzhledem k root.finalu, tj. tak jak leze z MIDu)
// displacementy teziste, ktere si spocita pomoci addMovement
// vraci true jestli se or nebo pos zmenily:
bool ColResp::applyDisplacements(int h, Matrix &or, float pos[3], float elPos[3])
{
	if (m_unControlTime==-1) 	
		return false; // zbran plne pod kontrolou, nic nemenime	

	float time = m_sinceCol.getTime()/1000;	
	m_unControl = time/m_unControlTime;
	if (m_unControl>=1)
	{
		m_unControlTime = -1;
		return false;
	}
	assert(m_unControl >= 0);

	// da do displacu pohyb za cas time:
	addMovement(h, time);		
	// provedeme SLERP:
	Matrix Modraz(m_angDisplac[h]); Modraz.setTranslation(m_linDisplac[h]);
	Matrix Mmys(or); Mmys.setTranslation(pos);
	// tady by se to chtelo rozhodnout jestli mec protne telo a vybrat smer SLERPu...

	bool obracene; // jestli pojedu SLERP ostrehu uhlu (primo) nebo tupeho (obracene), 21.12.02
	bool test1 = bodyWeaponCol(h, Modraz, Mmys, false);
	if (test1)
	{ // primy SLERP vede ke kolizi:
		//TRACE("******* Primy SLERP vede ke kolizi\n");
		bool test2 = bodyWeaponCol(h, Modraz, Mmys, true);
		if (test2)
		{ // obraceny taky, takze pouzijem radsi primy
			//TRACE(":-((((( obraceny taky -> pouzivam primy\n");
			obracene = false;
		}
		else
		{ // obraceny ne -> pouziju ho
			//TRACE(":-))))) obraceny ne -> pouziju ho\n");
			obracene = true;
		}
	}
	else // primy SLERP nevede ke kolizi:
		obracene = false; 

	Matrix res;
	Modraz.SLERP(res, Mmys, m_unControl, obracene);
	or.set(res.getMatrix());
	or.setTranslation(zero3);
	res.getTranslation(pos);	
	// jeste interpoluji elPos (se stejnym parametrem):
	float MIDelPos[3]; cpvector(MIDelPos, elPos);
	linterpvect(elPos, m_elDisplac[h], MIDelPos, m_unControl);

	return true;
/*
	// vyrobim z or a pos SHF
	or.setTranslation(pos);
	SHF2SHC(h, or);	
	or.addTranslation(m_linDisplac[h]);
	or.preRotMultiply(m_angDisplac[h]);
	SHC2SHF(h, or);
	or.getTranslation(pos);
	float zero[3]={0,0,0};
	or.setTranslation(zero);
*/
}

// konvence: teleso A = 0, teleso B = 1

// nastavi nejblizsi body dotyku usecek 'us' transformovanych SHF a taky normalu (ve smeru od telesa B)
// vraci vzdalenost valcu
float ColResp::getColPoints(const Matrix SHF[2], int us, float p[2][3], float normal[3])
{
	/*
	int tot[2] = {m_weapon[0]->m_numVertices,
		   	      m_weapon[1]->m_numVertices};
	Vector v0[2], v1[2];    
	if (us == W0W1)
	{
		// cepel0:
		v0[0].set(m_weapon[0]->m_pVertices[tot[0]-vA].m_location);		
		v1[0].set(m_weapon[0]->m_pVertices[tot[0]-vD].m_location);		
		// cepel1:
		v0[1].set(m_weapon[1]->m_pVertices[tot[1]-vA].m_location);		
		v1[1].set(m_weapon[1]->m_pVertices[tot[1]-vD].m_location);		
	}
	else if (us == W0Z1)
	{
		// cepel0:
		v0[0].set(m_weapon[0]->m_pVertices[tot[0]-vA].m_location);		
		v1[0].set(m_weapon[0]->m_pVertices[tot[0]-vD].m_location);		
		// zastita1:
		v0[1].set(m_weapon[1]->m_pVertices[tot[1]-vE].m_location);		
		v1[1].set(m_weapon[1]->m_pVertices[tot[1]-vF].m_location);		
	}
	else if (us == Z0W1)
	{
		// zastita0:
		v0[0].set(m_weapon[0]->m_pVertices[tot[0]-vE].m_location);		
		v1[0].set(m_weapon[0]->m_pVertices[tot[0]-vF].m_location);
		// cepel1:
		v0[1].set(m_weapon[1]->m_pVertices[tot[1]-vA].m_location);		
		v1[1].set(m_weapon[1]->m_pVertices[tot[1]-vD].m_location);		
	}

	for (int hrac=0; hrac<2; hrac++)
	{		
		//Apos[hrac].set(m_weapon[hrac]->m_pVertices[tot[hrac]-vA].m_location);
		v0[hrac].transform( SHF[hrac] ); // transformuju
		if (m_weapon[hrac]->m_leftHanded) v0[hrac].mul(1, 1, -1);	

		//Dpos[hrac].set(m_weapon[hrac]->m_pVertices[tot[hrac]-vD].m_location);
		v1[hrac].transform( SHF[hrac] ); // transformuju
		if (m_weapon[hrac]->m_leftHanded) v1[hrac].mul(1, 1, -1);			
	}
	*/
	float v0[2][3], v1[2][3];
	transformVertices(0, SHF[0]);
	transformVertices(1, SHF[1]);
	setSegVertices(us, v0, v1);
	// body v0[i], v1[i] tvori vrcholy usecky patrici hraci i (bud cepel, nebo zastita)

	float B[2][3];
	float dist = distSegSeg(v0[0], v1[0], v0[1], v1[1], B[0], B[1]);	
	// oba body dotyku jsou stejne a je to stred pricky usecek:
	cpvector(p[0], B[0]);
	Vector::add3(p[0], B[1]);
	Vector::mul3(p[0], 0.5);
	cpvector(p[1], p[0]);
	// normala je jednotkova smernice pricky ve smeru od telesa B (1):
	cpvector(normal, B[0]);
	Vector::sub3(normal, B[1]);
	unitize(normal);	

	//testik:
//	float norm2[3];
//	Dpos[0].sub(Apos[0]); Dpos[1].sub(Apos[1]);
//	float dpos[2][3];
//	Dpos[0].get(dpos[0]); Dpos[1].get(dpos[1]);	
//	crossproduct(norm2, dpos[0], dpos[1]);
//	unitize(norm2);
	//TRACE("normal: %f %f %f, norm2: %f %f %f\n", normal[0], normal[1], normal[2],
	//	norm2[0], norm2[1], norm2[2]);
	// Pozn. normaly se nemusi rovnat, protoze valec ma polokulovite zakonceni (tam je jina normala)

//	if (fabs(fabs(norm2[0])-fabs(normal[0]))>0.05 || fabs(fabs(norm2[1])-fabs(normal[1]))>0.05 ||
//		fabs(fabs(norm2[2])-fabs(normal[2]))>0.05)
//	{
//		TRACE("bacha, normaly si nejak neodpovidaj!\n");
//	}

	return dist;
}

// pocita rychlosti po aplikaci impulsu: (vraci false pokud se telesa vzdaluji)
// predpoklada kolizi typu 'us'
bool ColResp::applyImpulse(const Matrix SHF[2], int us, const float vPre[2][3], const float omegaPre[2][3],  
						   float vPost[2][3], float omegaPost[2][3])
{
	float colPoint[2][3], normal[3];
	float dist = getColPoints(SHF, us, colPoint, normal);
//	TRACE("getColPoints: dist = %f\n", dist);
//	TRACE("normal: %f %f %f\n", normal[0], normal[1], normal[2]);
	if (fabs(dist-2*m_R) > 0.1) TRACE("zajimavy, zbrane jsou v podezrele vzdalenosti (%f)\n", dist);
	Matrix SHFc[2] = {SHF[0], SHF[1]}; // SHF posunuty do teziste
	// odecteni teziste:
	float teziste[2][3];
	for (int h=0; h<2; h++)
	{
		SHFc[h].set(SHF[h].getMatrix());
		SHF2SHC(h, SHFc[h]);
		SHFc[h].getTranslation(teziste[h]);
		Vector::sub3(colPoint[h], teziste[h]);
	}
//	TRACE("bodA: %f, %f, %f   bodB: %f, %f, %f\n", colPoint[0][0], colPoint[0][1], 
//		colPoint[0][2], colPoint[1][0], colPoint[1][1], colPoint[1][2]);	
	// aktualni hmotnosti zbrane:
	Matrix actinvIbody[2]; float actmass[2];
//	TRACE("toughness[0] = %f, toughness[1] = %f\n", m_fmodel[0]->m_toughness, m_fmodel[1]->m_toughness);
	compActMass(0, actmass[0], actinvIbody[0]);
	compActMass(1, actmass[1], actinvIbody[1]);

	// spocitam aktualni tenzor setrvacnosti (inverzni)
	Matrix invSHFc[2], invIact[2];
	for (h=0; h<2; h++)  // vzorec (2-40)
	{
		SHFc[h].setTranslation(zero3);
		SHFc[h].inverseRotateMatrix(invSHFc[h]);
		invIact[h].set(SHFc[h].getMatrix());
		invIact[h].postMultiply(actinvIbody[h]);
		invIact[h].postMultiply(invSHFc[h]);
	}
	// rychlosti obou bodu dotyku (ackoliv body jsou (skoro/uplne) stejne, rychlosti se dost ruzni!)
	float pvel[2][3];
	for (h=0; h<2; h++)
	{		
		crossproduct(pvel[h], omegaPre[h], colPoint[h]);  // vzorec (8-1)
		Vector::add3(pvel[h], vPre[h]);	
	}
//	TRACE("Pa: %f %f %f, Pb: %f %f %f\n", pvel[0][0], pvel[0][1], pvel[0][2], pvel[1][0], pvel[1][1], pvel[1][2]);	
//	TRACE("normal: %f %f %f\n", normal[0], normal[1], normal[2]);
	// relativni rychlost bodu dotyku z pohledu telesa B (1) promitla do smeru 'normal'
	float vRel,tmp[3], tmp2[3];
	cpvector(tmp, pvel[0]);
	Vector::sub3(tmp, pvel[1]);
//	TRACE("relativni rychlost: %f %f %f\n", tmp[0], tmp[1], tmp[2]);
	vRel = DOT(normal, tmp);
//	TRACE("vRel: %f\n", vRel);
	if (vRel >=0 )
	{
		TRACE("Resting or retrating contact\n");
		return false;
	}
	// spocitame 'j':
	float numerator = -(1+m_epsilon)*vRel;
	float denominator = 0;
	Vector v;
	for (h=0; h<2; h++)
	{
		denominator += 1/actmass[h];
		crossproduct(tmp, colPoint[h], normal);		
		v.set(tmp);
		v.transform3(invIact[h]); // je jen 3x3
		crossproduct(tmp2, v.getVector(), colPoint[h]);
		float vys = DOT(normal, tmp2);
		denominator += vys;
	}
	float j = numerator/denominator;
	//TRACE("j=%f\n", j);
	// nove linearni a uhlove rychlosti:
	for (h=0; h<2; h++)
	{
		// na teleso A pusobi +j a na B zas -j
		float actj;
		if (h==1) actj = -j;
		else actj = +j; // h==A
		// linearni:
		cpvector(vPost[h], normal);
		Vector::mul3(vPost[h], actj/actmass[h]);
		Vector::add3(vPost[h], vPre[h]);
		// uhlova:
		cpvector(tmp, normal);
		Vector::mul3(tmp, actj);
		crossproduct(tmp2, colPoint[h], tmp);
		v.set(tmp2);
		v.transform3(invIact[h]);
		v.get(omegaPost[h]);
		Vector::add3(omegaPost[h], omegaPre[h]);
	}
	// pro force-feedback:
	cpvector(m_lastImpulse, normal);
	Vector::mul3(m_lastImpulse, j);

	// pro kontrolu:
	// rychlosti obou bodu dotyku po srazce:
	float Pvel[2][3];
	for (h=0; h<2; h++)
	{		
		crossproduct(Pvel[h], omegaPost[h], colPoint[h]);  // vzorec (8-1)
		Vector::add3(Pvel[h], vPost[h]);	
	}
//	TRACE("PostA: %f %f %f, PostB: %f %f %f\n", Pvel[0][0], Pvel[0][1], Pvel[0][2], Pvel[1][0], Pvel[1][1], Pvel[1][2]);
	cpvector(tmp, Pvel[0]);
	Vector::sub3(tmp, Pvel[1]);
	float vRelPlus = DOT(normal, tmp);
//	TRACE("vRelPlus: %f, -m_epsilon*vRel = %f\n", vRelPlus, -m_epsilon*vRel);
	// melo by platit vRelPlus = -m_epsilon*vRel
	if (fabs(vRelPlus/vRel + m_epsilon) > 0.1) 
	{
		TRACE("Pozor! neplati vRelPlus = -m_epsilon*vRel\n");
		assert(false);		
	}
	return true;
}

// pocita rychlosti po aplikaci impulsu, jako applyImpulse, akorat nepocita s uhlovou rychlosti:
bool ColResp::applyLinImpulse(const Matrix SHF[2], const float vPre[2][3], float vPost[2][3])
{
	assert(false&&"Pozor, v linImpulse se nepocita s pevnosti ruky");


	float colPoint[2][3], normal[3];
	float dist = getColPoints(SHF, W0W1, colPoint, normal);
	if (fabs(dist-2*m_R) > 0.1) TRACE("zajimavy, zbrane jsou v podezrele vzdalenosti (%f)\n", dist);
	Matrix SHFc[2] = {SHF[0], SHF[1]}; // SHF posunuty do teziste
	// odecteni teziste:
	float teziste[2][3];
	for (int h=0; h<2; h++)
	{
		SHFc[h].set(SHF[h].getMatrix());
		SHF2SHC(h, SHFc[h]);
		SHFc[h].getTranslation(teziste[h]);
		Vector::sub3(colPoint[h], teziste[h]);
	}
	//TRACE("bodA: %f, %f, %f   bodB: %f, %f, %f\n", colPoint[0][0], colPoint[0][1], 
	//	colPoint[0][2], colPoint[1][0], colPoint[1][1], colPoint[1][2]);	
	for (h=0; h<2; h++) SHFc[h].setTranslation(zero3);
	// rychlosti obou bodu dotyku (ackoliv body jsou (skoro/uplne) stejne, rychlosti se dost ruzni!)
	float pvel[2][3];
	for (h=0; h<2; h++) cpvector(pvel[h], vPre[h]);
	//TRACE("Pa: %f %f %f, Pb: %f %f %f\n", pvel[0][0], pvel[0][1], pvel[0][2], pvel[1][0], pvel[1][1], pvel[1][2]);	
	//TRACE("normal: %f %f %f\n", normal[0], normal[1], normal[2]);
	// relativni rychlost bodu dotyku z pohledu telesa B (1) promitla do smeru 'normal'
	float vRel,tmp[3];
	cpvector(tmp, pvel[0]);
	Vector::sub3(tmp, pvel[1]);
	vRel = DOT(normal, tmp);
	if (vRel >= 0 )
	{
		TRACE("Lin: Resting or retrating contact\n");
		return false;
	}
	//TRACE("vRel: %f\n", vRel);
	// spocitame 'j':
	float numerator = -(1+m_epsilon)*vRel;
	float denominator = 0;
	Vector v;
	for (h=0; h<2; h++)	denominator += 1/m_mass[h];
	float j = numerator/denominator;
	//TRACE("j=%f\n", j);
	// nove linearni rychlosti:
	for (h=0; h<2; h++)
	{
		// na teleso A pusobi +j a na B zas -j
		float actj;
		if (h==1) actj = -j;
		else actj = +j; // h==A
		// linearni:
		cpvector(vPost[h], normal);
		Vector::mul3(vPost[h], actj/m_mass[h]);
		Vector::add3(vPost[h], vPre[h]);	
	}

	// pro kontrolu:
	// rychlosti obou bodu dotyku po srazce:
	float Pvel[2][3];
	for (h=0; h<2; h++) cpvector(Pvel[h], vPost[h]);
	//TRACE("PostA: %f %f %f, PostB: %f %f %f\n", Pvel[0][0], Pvel[0][1], Pvel[0][2], Pvel[1][0], Pvel[1][1], Pvel[1][2]);
	cpvector(tmp, Pvel[0]);
	Vector::sub3(tmp, Pvel[1]);
	float vRelPlus = DOT(normal, tmp);
	//TRACE("vRelPlus: %f, -m_epsilon*vRel = %f\n", vRelPlus, -m_epsilon*vRel);
	// melo by platit vRelPlus = -m_epsilon*vRel
	if (fabs(vRelPlus/vRel + m_epsilon) > 0.1) 
	{
		TRACE("Pozor! neplati vRelPlus = -m_epsilon*vRel\n");
		assert(true);		
	}
	return true;
}

// spocita o kolik je minimalne potreba odstrcit usecky 'us' ve smeru 'dir',
// aby vzdalenost mezi nimi byla alespon D; vzdy odstrkavam usecku 2. hrace
float ColResp::sepDir(Matrix SHres[2], float dir[3], int us, float D)
{
	float dist0, dist1; // vzdalenost v pocatecni poloze a v poloze horniho odhadu
	float v1[2][3], v2[2][3]; // vrcholy usecek v pocatecni poloze (dist0 <= D)
	float dMax; // horni odhad posunuti (nasobku 'dir')
	float dummy1[3], dummy2[3];
	float dist;

	transformVertices(0, SHres[0]);
	transformVertices(1, SHres[1]);
	setSegVertices(us, v1, v2);
	dist0 = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy1, dummy2);
	if (dist0 > D) 
		TRACE("ColResp::sepDir: podezrely, dist0 = %f, D = %f\n", dist0, D);			
	assert(dist0 <= D);
	// ted spoctu horni odhad posunuti (pmci obalovych kouli, viz. 12.2.03)
	float c[2][3]; // stredy usecek
	float R[2]; // polomery kouli, tj. polovina delky usecek
	for (int h=0; h<2; h++)
	{
		cpvector(c[h], v1[h]);
		Vector::add3(c[h], v2[h]);
		Vector::mul3(c[h], 0.5);

		float tmp[3];
		cpvector(tmp, v1[h]);
		Vector::sub3(tmp, c[h]);
		R[h] = sqrt(DOT(tmp, tmp));
	}
	dMax = sepPoints(c[0], c[1], dir, R[0] + R[1] + D + GEOM_EPSILON); // epsilon na zaruceni >D
	assert(dMax >= 0); // reseni by melo existovat
	// testik, jestli to fakt dost odpreparovalo druhou usecku:
	float pos[3];
	cpvector(pos, dir);
	Vector::mul3(pos, dMax);
	Matrix SHpos;
	SHpos.set(SHres[1].getMatrix());
	SHpos.addTranslation(pos);
	transformVertices(1, SHpos);
	setSegVertices(us, v1, v2);
	dist1 = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy1, dummy2);
	//TRACE("dist0 = %f, vzdalenost po posunuti o dMax: dist1 = %f\n", dist0, dist1);
	assert(dist1 > D);
	// mam meze, nasleduje binarni dohledani minimalniho posunuti splnujiciho dist(us) > D
	float lo = 0.0; float hi = dMax; // meze posunuti	
	int iter = 30; bool ok = false;
	while (iter-- > 0)
	{
		if (hi-lo < GEOM_EPSILON) { ok = true; break; }
		float d = (lo + hi)/2;

		cpvector(pos, dir);
		Vector::mul3(pos, d);
		SHpos.set(SHres[1].getMatrix());
		SHpos.addTranslation(pos);
		transformVertices(1, SHpos);
		setSegVertices(us, v1, v2);
		dist = distSegSeg(v1[0], v2[0], v1[1], v2[1], dummy1, dummy2);

		if (dist <= D) lo = d;
		else /* dist > D*/ hi = d;
	}
	//TRACE("ColResp::sepDir - bylo potreba %i iteraci\n", 30-iter);
	// invariant: v lo je vzdy parametr davajici dist <= D, v hi je vzdy > D
	if (!ok)
	{
		TRACE("ColResp::sepDir - nepodarilo se presne urcit velikost posunu\n");
		assert(false);
	}		
	return hi;
}	

// posune SHF tak, aby jim transformovane valce byly ve vzdalenosti aspon 2*m_Rb
// vraci jestli se separace podarila; do 'vzd' (je-li neNULL) ulozi o jakou vzdalenost 
// bylo treba separovat
bool ColResp::separate(Matrix SHres[2], float *vzd)
{		
	float dist, dummy[2][3], normal[3][3], posun, vysNorm[3];
	float dist1 = getColPoints(SHres, W0W1, dummy, normal[0]);
	float dist2 = getColPoints(SHres, W0Z1, dummy, normal[1]);
	float dist3 = getColPoints(SHres, Z0W1, dummy, normal[2]);
	bool k1 = false, k2 = false, k3 = false;
	if (dist1 < 2*m_Rb) k1 = true;
	if (dist2 < 2*m_Rb) k2 = true;
	if (dist3 < 2*m_Rb) k3 = true;
	dist = min3(dist1, dist2, dist3);	
	if (!k1 && !k2 && !k3) // zadna koliz, vzdalenost je bepecna
		{ if (vzd) *vzd=0.0; return true; } 
	int typKolize;
	//TRACE("\nColResp::separate\n");
	//TRACE("dist1 = %f, dist2 = %f, dist3 = %f\n",dist1,dist2,dist3);	
	if (k1&&k2&&k3) 
	{
		TRACE("******************* 3 kolize najednou! ************************\n");		
		// zkusim prumer ze vsech normal
		cpvector(vysNorm, normal[0]);		
		Vector::add3(vysNorm, normal[1]);
		Vector::add3(vysNorm, normal[2]);		
		unitize(vysNorm); // to aby mel smysl 'posun'

		float dir[3]; cpvector(dir, vysNorm); 
		Vector::mul3(dir, -1.0); // smer ukazuje tam kam se strka teleso B
		// ted pmci sepDiru najdu nejmensi vzdalenost posunuti pro obe dvojice usecek:
		float sepDist1, sepDist2, sepDist3;
		sepDist1 = sepDir(SHres, dir, W0W1, 2*m_Rb);
		sepDist2 = sepDir(SHres, dir, W0Z1, 2*m_Rb);
		sepDist3 = sepDir(SHres, dir, Z0W1, 2*m_Rb);

		posun = max3(sepDist1, sepDist2, sepDist3) / 2;
	}
	else if ((k1&&k2) || (k1&&k3) || (k2&&k3))
	{
		//TRACE("********************** 2 kolize najednou ************************\n");
		// zkusim prumer z obou normal
		cpvector(vysNorm, zero3);
		if (k1) Vector::add3(vysNorm, normal[0]);
		if (k2) Vector::add3(vysNorm, normal[1]);
		if (k3) Vector::add3(vysNorm, normal[2]);		
		unitize(vysNorm); // to aby mel smysl 'posun'

		float dir[3]; cpvector(dir, vysNorm); 
		Vector::mul3(dir, -1.0); // smer ukazuje tam kam se strka teleso B
		// ted pmci sepDiru najdu nejmensi vzdalenost posunuti pro obe dvojice usecek:
		float sepDist1 = 0.0, sepDist2 = 0.0, sepDist3 = 0.0;
		if (k1) sepDist1 = sepDir(SHres, dir, W0W1, 2*m_Rb);
		if (k2) sepDist2 = sepDir(SHres, dir, W0Z1, 2*m_Rb);
		if (k3) sepDist3 = sepDir(SHres, dir, Z0W1, 2*m_Rb);

		posun = max3(sepDist1, sepDist2, sepDist3) / 2;
	}	
	else
	{ // jenom jedna koliz		
		typKolize = min3arg(dist1, dist2, dist3);
		//TRACE(" ** 1 kolize **, typ = %i\n", typKolize);
		cpvector(vysNorm, normal[typKolize-1]);
		posun = (2*m_Rb - dist)/2; // o kolik posunout kazdy valec

		float dir[3]; cpvector(dir, vysNorm); 
		Vector::mul3(dir, -1.0); // smer ukazuje tam kam se strka teleso B
		float sepDist = sepDir(SHres, dir, typKolize, 2*m_Rb);
		//TRACE("posun = %f, sepDist/2 = %f\n", posun, sepDist/2);
		if (fabs(posun - sepDist/2) > 0.01) 
		{
			TRACE("podezrele velky rozdil posunu a sepDistu/2\n");
			assert(false);
		}
	}
	//TRACE("nebezpecna vzdalenost (%f), separuji o posun %f\n", dist, posun);	
	if (posun > m_maxSeparationTrans)
	{
		TRACE("################### Separace se nepodarila #####################\n");
		TRACE(" -- byl by potreba prilis velky posun: %f\n", posun);
		return false;
	}
	float pos[2][3];
	cpvector(pos[0], vysNorm);
	Vector::mul3(pos[0], posun); // normala jde ve smeru od B
	cpvector(pos[1], vysNorm);
	Vector::mul3(pos[1], -posun); // B v protismeru
	SHres[0].addTranslation(pos[0]);
	SHres[1].addTranslation(pos[1]);

	if (vzd) *vzd = posun*2; // o kolik se posouvalo
//test, jestli se separace podarila:
	dist1 = getColPoints(SHres, W0W1, dummy, dummy[0]);
	dist2 = getColPoints(SHres, W0Z1, dummy, dummy[0]);
	dist3 = getColPoints(SHres, Z0W1, dummy, dummy[0]);
	dist = min3(dist1, dist2, dist3);
	if (dist < 2*m_Rb - 0.001) 
	{
		TRACE("dist1 = %f, dist2 = %f, dist3 = %f\n",dist1,dist2,dist3);
		TRACE("################### Separace se nepodarila (%f) #####################\n", dist);
		//assert(false);
		return false;
	}
	return true;
}

// separace po castech: simuluje pohyb s rychlostmi linvel, angvel po cas time s tim,
// ze kdyz dojde ke kolizi tak si pomaha separaci:
// vraci jestli se poovedlo separovat (a tedy jestli SHres ma smysl)
bool ColResp::separatePerPartes(Matrix SHres[2], float EPres[2][3], const float linvel[2][3], const float angvel[2][3], float time)
{
	float timeLeft = time;
	int h, iter=1;
	Matrix SHaft[2], SHbef[2];
	while (timeLeft > 0)
	{		
		TRACE("sepPerPartes, iter: %i, zbyva cas %f\n", iter, timeLeft);
		// provedu pohyb za cas timeLeft s danymi rychlostmi
		for (h=0; h<2; h++)		
		{
			SHbef[h].set(SHres[h].getMatrix());	// vychozi poloha
			SHaft[h].set(SHres[h].getMatrix());	// "vchozi" poloha :-)
			applyVelocity(h, linvel[h], angvel[h], SHaft[h], timeLeft);		
		}
		int typKolize;
		float tCol = findCollision(SHbef, SHaft, &typKolize);
		if (tCol == -1) 
		{ // nejni kolize, vratim SHaft a koncim:
			if (iter == 1)
				TRACE("byl volan separatePerPartes bez kolize (viz. 5.12.02)\n");
			SHres[0].set(SHaft[0].getMatrix());
			SHres[1].set(SHaft[1].getMatrix());
			break;
		}
		// je kolize, takze rollBack:
		rollBack(tCol, SHbef, SHaft, SHres, EPres); // vraci na vzdalenost m_R
		timeLeft = (1-tCol)*timeLeft;
		float sepDist;
		bool ok = separate(SHres, &sepDist); // odseparuje na vzdalenost m_Rb; dulezite: m_Rb > m_R
		if (!ok) return false;
		float sepTime = (sepDist>0) ? (sepDist/100) : 0; // empiricky vztah viz. 6.12.02
		//TRACE("sepDist = %f, sepTime = %f\n", sepDist, sepTime);

		timeLeft -= sepTime; // odectu cas za separaci
		




				// sejvovani animace: zapsat zmeny do stavu modelu:
	for (int h=0; h<2; h++)
	{
		Matrix or, inv; float pos[3];
		or.set(SHres[h].getMatrix());
		m_fmodel[h]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(inv);
		or.preMultiply(inv); // podle konvence do stavu modelu ma prijit matice vzhledem k root.final
		or.getTranslation(pos);
		or.setTranslation(zero3);
		m_fmodel[h]->updateModelState(or, pos, m_oldElPos[h]);
	}
	m_as->saveState(2); // a zasejvovat









		iter++;
	}
	assert(iter<20 &&"toto neni chyba, ale upozorneni ze pocet iteraci v sepPP prekrocil 19");
	return true; // povedlo se
}

// spocita zadne a hlavne maximalni vychyleni obou zbrani z velikosti uhlove rychlosti (vzhledem k root.finalu)
// pouziva tez m_unControlTime pro pripad skoro nulove uhlove rychlosti:
void ColResp::compDeflection(const Matrix SHF[2], const float linvel[2][3], const float angvel[2][3],
					Matrix noDef[2], Matrix maxDef[2], float EPnoDef[2][3], float EPmaxDef[2][3])
{
	int h, maxAng;
	Matrix invFinal[2];	
	float angmag[2];

	for (h=0; h<2; h++)
	{
		m_fmodel[h]->setFinalMatrices(); // budu potrebovat root.final (a jen to!)
		// zadne vychyleni: relativizuji vzhledem k root.final, tj. prenasobim zleva inverzi root.final
		noDef[h].set(SHF[h].getMatrix());	
		m_fmodel[h]->m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(invFinal[h]);		
		// maxDef spocitam ve svetovych souradnicich:
		maxDef[h].set(SHF[h].getMatrix());				
	}
	// spoctu pozici lokte pri zadnem vychyleni:
	for (h=0; h<2; h++)
	{
		Matrix or; or.set(noDef[h].getMatrix());
		float pos[3];
		or.getTranslation(pos);
		or.setTranslation(zero3);
		m_ikan[h]->compElbowPos(pos, or, EPnoDef[h]);
	}
	noDef[0].preMultiply(invFinal[0]);
	noDef[1].preMultiply(invFinal[1]);
	// vyberu vetsi z uhlovych rychlosti (maxAng):
	angmag[0] = norm(angvel[0]);
	angmag[1] = norm(angvel[1]);
	if (angmag[0] > angmag[1]) maxAng = 0;
	else maxAng = 1;
	if (angmag[maxAng] < GEOM_EPSILON) // obe uhlove rychlosti jsou moc male
	{
		assert(!"Toto neni chyba, jen velmi nepravdepodobna moznost, ke ktere zatim nedoslo");
		TRACE("Warning: mala vetsi uhlova rychlost angmag[angMax] = %f\n", angmag[maxAng]);
		// jiny model: jednoduse necham linvel pusobit cas m_unControlTime/2
		applyVelocity(0, linvel[0], zero3, maxDef[0], m_unControlTime/2);
		applyVelocity(1, linvel[1], zero3, maxDef[1], m_unControlTime/2);
	}
	else // vetsi uhlova rychlost je dost velka
	{
		// podle angmag[maxAng] urcim uhel maximalniho vychyleni (alfa):
		float alfa = M_PI/3*(1 - 1/(angmag[maxAng]/6 + 1));
		// dopocitam cas po ktery musi ta vetsi uhlova rychlost pusobit aby dosahla uhlu alfa:
		float t = alfa/angmag[maxAng];
		// necham vsechny rychlosti pusobit cas t:
		applyVelocity(0, linvel[0], angvel[0], maxDef[0], t);
		applyVelocity(1, linvel[1], angvel[1], maxDef[1], t);
	}
	// spoctu pozici lokte pri maximalnim vychyleni:
	for (h=0; h<2; h++)
	{
		Matrix or; or.set(maxDef[h].getMatrix());
		float pos[3];
		or.getTranslation(pos);
		or.setTranslation(zero3);
		m_ikan[h]->compElbowPos(pos, or, EPmaxDef[h]);
	}
	// nakonec vratim maxDefy do souradnic vzhledem k root.finalu:
	maxDef[0].preMultiply(invFinal[0]);
	maxDef[1].preMultiply(invFinal[1]);
}

// pomoci m_fmodel[h]->m_toughness (pevnost drzeni mece od 0 do 1) spocita aktualni hmotnost
// a inverzni tenzor setrvacnosti zbrane:
void ColResp::compActMass(int h, float &mass, Matrix &invI)
{
	float ro = 1 + m_fmodel[h]->m_toughness*7; // hustota mece od 1 do 8
	mass = m_mass[h]*ro;
	m_invIbody[h].mulScal(1/ro, invI);
}

// vraci informace o odrazeni zbrane, vraci jestli v poslednim volani go byl nastaven odraz:
bool ColResp::getDeflection(int h, Matrix &noDef, Matrix &maxDef)
{
	noDef = m_noDef[h];
	maxDef = m_maxDef[h];
	return m_odraz;
}