/* 
    ColResp.h

	Collision response evaluation

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "LocalRemoteModel.h"
#include "ColDet.h"
#include "Matrix.h"
#include "ikan.h"
#include "myvec.h"
#include "Win32Timer.h"

#include "AnimSequence.h"

#ifndef COLRESP_H
#define COLRESP_H

const int	W0W1 = 1, W0Z1 = 2, Z0W1 = 3; // typ kolize, zastita1-zastita2 neuvazuju

class ColResp  {	
protected:	

	AnimSaver	*m_as;

	FencerModel *m_fmodel[2];
	Model	*m_model[2];
	Model	*m_weapon[2];
	CDModel	*m_cdw[2]; // obaly zbrani
	Matrix  m_SHmin[2]; // minule swordholder.final pro oba modely
	Matrix  m_SHnew[2]; // aktualni swordholder.final pro oba modely
	Matrix	m_rootMin[2]; // minuly root.final	
	IKAN	*m_ikan[2]; // IK solver pro oba modely
	float	m_oldElPos[2][3]; // predchozi poloha loktu (z IKANu)
	Win32Timer m_timer;
//	bool	m_lastCol; // doslo pri minulem volani 'go' ke kolizi?
	float	m_lastTime, m_deltaTime;
//	float	m_timeEstimate; // odhad jak dlouho potrva nez dojde k dalsimu zavolani go (5.12.02)
	float	m_R; // polomer valce aproximujicicho zbran
	float	m_Rb; // polomer "bezpecneho" valce (13.10.02)
	float	m_Rn; // polomer nebezpecneho, vnitrniho valce (13.10.02) 
	// ma platit m_Rn < m_R < m_Rb
// dynamicke vlastnosti zbrane:
	float	m_mass[2]; // hmotnost
	float	m_com[2][3]; // teziste
	Matrix  m_invIbody[2]; // inverze tenzoru setrvacnosti (translacni cast matice je 0)

	float	m_Rbody[2]; // polomery valcu aproximujicich tela
	float	m_RnBody[2]; // "nebezpecne" polomery valcu aproximujicich tela
	float	m_maxSeparationTrans; // o kolik muze separace maximalne hnout 1 zbrani

	float	m_linvel[2][3]; // rychlost (linearni) 
	float	m_angvel[2][3]; // uhlova rychlost
	float	m_minLinvelMagn[2], m_minAngvelMagn[2]; // velikost minule linearni a uhlove rychlosti
	float	m_linDisplac[2][3]; // kumulovane posunuti
	Matrix	m_angDisplac[2]; // kumulovana rotace
	float	m_elDisplac[2][3]; // pozice lokte souhlasici s xxxDisplac
	float	m_EPDisplac[2][3]; // posunuti pozice lokte

	Matrix  m_noDef[2]; // zadne vychyleni (deflekce), tj. pozice po rollBacku a applyVelocity
	Matrix	m_maxDef[2]; // maximalni vychyleni, stanovi se ve fci compDeflection
	float	m_EPnoDef[2][3]; // pozice lokte pro zadne vychyleni (presneji jen po rollBacku)
	float	m_EPmaxDef[2][3]; // pozice lokte pro maximalni vychyleni, spocita se taky v compDef
	float	m_epsilon;	// refrakcni koeficient

	bool	m_odraz; // kvuli force-feedbacku - odrazena zbran v poslednim volani go?
	bool    m_appliedImpulse; // byl v poslednim volani go aplikovan impuls?
	float	m_lastImpulse[3]; // posledni aplikovany impuls

public: // vratit na protected!

	float	m_unControl; // jak moc mec neni pod kontrolou; -1: je plne pod kontrolou
	float	m_unControlTime; // jak dlouho se bude vymykat kontrole
	Win32Timer m_sinceCol; // pocita cas od posledni kolize

public:	
	ColResp(FencerModel *fmodelA, FencerModel *fmodelB, IKAN *ikanA, IKAN *ikanB);
	~ColResp();

	// 'go' se vola kazdy animacni krok;
	bool GO(AnimSaver *as);
	// nacte ze souboru hmotnost, teziste a tenzor setrvacnosti:
	bool loadMassProp(int h, const char *fname);
	// posune translacni cast matice z SHF do stredu hmotnosti (pouziva m_com):
	void SHF2SHC(int h, Matrix &shf);
	// posune translacni cast matice ze stredu hmotnosti zbrane do pocatku, tj. SHF (pouziva m_com):
	void SHC2SHF(int h, Matrix &shf);
	// vynuluje displacementy:
	void resetDisplacements();	
	// prida k orientaci&posunuti zbrane a k pozici lokte displacementy teziste, ktere si spocita pomoci addMovement
	// vraci true jestli se or nebo pos zmenily:
	bool applyDisplacements(int h, Matrix &or, float pos[3], float elPos[3]);
	// pro FF: vraci informace o odrazeni zbrane, vraci jestli v poslednim volani go byl nastaven odraz:
	bool getDeflection(int h, Matrix &noDef, Matrix &maxDef);
	// pro FF: vraci informace o poslednim aplikovanem impulsu
	const float *getLastImpulse() { return m_lastImpulse; }
	bool appliedImpulse() { return m_appliedImpulse; }

protected:	
	// vrati maximalni okamzitou rychlost zbrane pohybujici se s danymi rychlostmi v poloze SHF:
	float weaponMaxVel(int h, const float linvel[3], const float angvel[3], const Matrix SHF);
	// nastavi m_weapon[h]->m_pActVers[A, D] na hodnoty zbrane interpolovane mezi polohami SHmin a SHnew
	// interpoluje se s parametrem t bud normalne, nebo proti smeru 'obracene'
	// je-li interp neNULL, tak do nej zapise vyslednou interpolovanou matici v IKANskych konvencich:
	void interpWeapon(int h, float t, const Matrix SHmin, const Matrix SHnew, 
							   ikMatrix interp = 0, bool obracene = false);
	// vraci vzdalenost usecek interpolovanych v case t
	// do interp nastavuje (jsou-li neNULL) interpolovane swordholder.finaly
	// vraci vzdalenost usecek v case t
	float interpSegDist(float t, const Matrix SHmin[2], const Matrix SHnew[2], int *typ = 0, ikMatrix interp0 = 0, ikMatrix interp1 = 0);
	// hleda okamzik kontaktu mezi SHmin a SHnew; pokud neexistuje vraci -1:
	// vstupni predpoklad: v m_SHmin neni kolize
	float findCollision(const Matrix SHmin[2], const Matrix SHnew[2], int *typ);
	// tc je okamzik kontaktu; nastavuje vysledny SHF a ElPos: (pozor na kumulaci chyby)
	void rollBack(float tc, const Matrix SHmin[2], const Matrix SHnew[2], Matrix SHres[2], float ElPosRes[2][3]);
	// vraci displacementy:
	void getLinDisplac(int h, float l[3]) { cpvector(l, m_linDisplac[h]); }
	void getAngDisplac(int h, Matrix &m) { m.set(m_angDisplac[h].getMatrix()); }	
//	// kompenzuje rollBack: - obsolette
//	void rollForward(Matrix SHres[2], float linvel[2][3]);		
	// nastavi linearni rychlost zbrane; pozaduje m_com:	
	void getLinVelocity(int h, const Matrix &mSHmin, const Matrix &mSHnew, 
	  				    float m_deltaTime, float vel[3]);
	// nastavi uhlovou rychlost v [rad/s]:
	void getAngVelocity(const Matrix &mSHmin, const Matrix &mSHnew, float m_deltaTime, 
							 float omega[3]);
	// transformuje SHF tak jako by se pohyboval rychlostmi linvel a angvel po dobu deltaTime:
	void applyVelocity(int h, const float linvel[3], const float angvel[3], Matrix &SHF, 
							float deltaTime);
	// nastavi nejblizsi body zbrani transformovanych SHF: (vraci vzdalenost usecek)
	float getColPoints(const Matrix SHF[2], int us, float p[2][3], float normal[3]);
	// pocita rychlosti po aplikaci impulsu:
	bool applyImpulse(const Matrix SHF[2], int us, const float vPre[2][3], const float omegaPre[2][3],  
						   float vPost[2][3], float omegaPost[2][3]);
	// totez co applyImpulse, ale nepocita s uhlovymi rychlostmi:
	bool applyLinImpulse(const Matrix SHF[2], const float vPre[2][3], float vPost[2][3]);
	// posune SHF tak, aby jim transformovane valce byly ve vzdalenosti aspon 2*m_Rb	
	bool separate(Matrix SHres[2], float *vzd = 0);
	// vraci minimalni vzdalenost mezi interpolovanymi useckami (pro ladici ucely):
	float getMinDist(const Matrix SHmin[2], const Matrix SHnew[2]);
	// vola IK ktera propaguje zmeny v SHF a pozice lokte (elPos) do cele paze:
	void updateArm(const Matrix SHF[2], const float elPos[2][3]);
	// spocita zadne a hlavne maximalni vychyleni obou zbrani z velikosti uhlove rychlosti (vzhledem k root.finalu)
	// pouziva tez m_unControlTime pro pripad skoro nulove uhlove rychlosti:
	void compDeflection(const Matrix SHF[2], const float linvel[2][3], const float angvel[2][3],
					Matrix noDef[2], Matrix maxDef[2], float EPnoDef[2][3], float EPmaxDef[2][3]);
	// ma za ukol vyrobit m_angDisplac a m_linDisplac na zaklade casu od kolize 'timeSinceCol'
	// pouziva matice m_noDef a m_maxDef a celkovy cas reakce m_unControlTime:
	void addMovement(int h, float timeSinceCol);
	// separace po castech: simuluje pohyb s rychlostmi linvel, angvel po cas time s tim,
	// ze kdyz dojde ke kolizi tak si pomaha separaci:
	bool separatePerPartes(Matrix SHres[2], float EPres[2][3], const float linvel[2][3], const float angvel[2][3], float time);
	// spocita maximalni velikost casu deltaT za ktery se valcoidy priblizi maximalne o R-Rn (19.12.02):
	float maxStep(const float linvel[2][3], const float angvel[2][3], const Matrix SHF[2]);
	// vraci vzdalenost tela hrace h a jeho zbrane; telo i zbran jsou aprox. valci, zbran je na ceste
	// od SHzac do SHkon v delce 't' (21.12.02):
	float bodyWeaponDist(int h, float t, const Matrix SHzac, const Matrix SHkon, bool obracene);
	// testuje jestli zbran hrace h protne jeho telo pri ceste ('obracene') mezi SHzac a SHkon:
	bool bodyWeaponCol(int h, const Matrix SHzac, const Matrix SHkon, bool obracene);

	// spocita o kolik je minimalne potreba odstrcit usecky 'us' ve smeru 'dir',
	// aby vzdalenost mezi nimi byla alespon D
	float sepDir(Matrix SHres[2], float dir[3], int us, float D);

	// 'us' urcuje dve usecky, u1 a u2. Do 'v1[i]' a 'v2[i]' se zapisi aktualni
	// koncove body usecky ui
	void setSegVertices(int us, float v1[2][3], float v2[2][3]);

	// vezme dulezite vrcholy (A,D,E,F) z m_weapon[h]->m_pVertices, transformuje 'cinterp'
	// a vysledek zapise do m_weapon[h]->m_pActVers
	void transformVertices(int h, const Matrix cinterp);


private:
	void compActMass(int h, float &mass, Matrix &invI);
};

#endif