/* 
    Communication.cpp

	Routines for network communication using windows sockets.	

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "gl1.h"
#include "Communication.h"
#include "MainFrm.h"
#include "ModeViewer.h"
#include "ChildView.h"
#include <assert.h>

int msgLengths[LAST_MESSAGE] = {sizeof(ModelState)*2, sizeof(InputEvent), 1, 6, sizeof(JudgeMessage)};

void CServerSocket::OnAccept(int nError)
{
//	TRACE("CServerSocket::OnAccept - threadId = %i\n", GetCurrentThreadId());

	if (nError) 
	{
		TRACE("Error %i\n",nError);
		MessageBox(NULL, "Accept connection failed (1)", "Error", MB_ICONSTOP);
		return;
	}
	int ff = -1;
	for (int n=0; n<cMaxClients; n++) 
		if (!m_conn[n]) { ff=n; break; }	
	if (ff == -1)
	{
		// moc klientu, akceptujem, poslem error a zavrem:
		CCommSocket tmp;
		if (Accept(tmp))
		{
			tmp.sendInfo(ERRMSG_TOOMANYCLIENTS);
			tmp.Close();
		}
		// else se to vyresilo samo kvuli jine chybe :-)
		return;
	}

	// vyrobim si komunikacni socket:
	m_conn[ff] = new CCommSocket(0, this, ff);
	if (Accept(*m_conn[ff])) 
	{		
		TRACE("Accepted connection nr. %i OK\n", ff);
		bool nodelay = true;
		if (!SetSockOpt(TCP_NODELAY, &nodelay, sizeof(bool), IPPROTO_TCP))
		{
			TRACE("CServerSocket::OnAccept Nepovedlo se SetSockOpt na TCP_NODELAY = true\n");
		}
		else TRACE("CServerSocket::OnAccept SetSockOpt TCP_NODELAY = true succeeded\n");
		m_newClient = true;
		// posleme zpravu MainFramu na pridani klienta do obou men:
		CString txt;
		getClientString(ff, txt);
		((CMainFrame *)m_owner)->addToMenus(ff, txt);
		// nove prisedsimu klientovi posleme init:
		if (m_quality == 0) m_conn[ff]->sendInfo(INITMSG_QUALITY0);
		else if (m_quality == 1) m_conn[ff]->sendInfo(INITMSG_QUALITY1);
		else assert(false);
	}	
	else
	{
		MessageBox(NULL, "Accept connection failed (2)", "Error", MB_ICONSTOP);	
		delete m_conn[ff];
		m_conn[ff] = 0;
	}
}

void CServerSocket::startServer(int port)
{
	int err=0;

	// byl-li otevreny -> zavrit
	for (int n=0; n<cMaxClients; n++) 
	{
		if (m_conn[n]) m_conn[n]->Close();
		m_conn[n] = 0;
	}
	Close();
	
	if (!Create(port, SOCK_STREAM))
	{		
		err=GetLastError();
		TRACE("Create error %i\n",err);
		MessageBox(0, "Can't create server socket", "Error", MB_ICONSTOP);
		return;
	}	

	bool nodelay = true;
	if (!SetSockOpt(TCP_NODELAY, &nodelay, sizeof(bool)))
	{
		TRACE("CServerSocket: Nepovedlo se SetSockOpt na TCP_NODELAY = true\n");
	}
	else TRACE("CServerSocket: SetSockOpt na TCP_NODELAY = true succeeded\n");

	if (!Listen(1))
	{
		TRACE("Listening on port %i failed\n", port);
		err=GetLastError();
		TRACE("Listen error %i\n",err);
		MessageBox(0, "Listening failed", "Error", MB_ICONSTOP);		
	}	
}

void CServerSocket::clientDied(int nr)
{
	assert(m_conn[nr]);
	delete m_conn[nr];
	m_conn[nr] = 0;
	((CMainFrame *)m_owner)->delFromMenus(nr);
}

void CServerSocket::killServer()
{
	TRACE("CServerSocket::killServer\n");
	for (int n=0; n<cMaxClients; n++) 
		if (m_conn[n]) 
		{
			m_conn[n]->Close(); // nevola OnClose, zavolame si ho sami:
			clientDied(n);
			//delete m_conn[n];
			//m_conn[n]=0;
		}
	Close(); // zavru se
}

CServerSocket::~CServerSocket()
{
	killServer();
}

// nastavi textovy popis i-teho klienta; vraci jestli je vubec nejaky pripojen:
bool CServerSocket::getClientString(int i, CString &txt) const
{
	if (!m_conn[i]) return false;
	CString ip; unsigned int port;
	m_conn[i]->GetPeerName(ip, port);
	char buf[64];
	sprintf(buf, "%i", port);
	txt = ip + CString(":") + CString(buf);	
	return true;
}



void CCommSocket::reset()
{
	isReady = false;	
	bi = 0; // prazdny buffer
	wrkMsg.type = MSG_INFO;
	wrkMsg.body[0] = 0;
	receiving = TYPE;
}

CCommSocket::CCommSocket(void *childView, CServerSocket *ss, int sn)
{
	for (int n=0; n<LAST_MESSAGE; n++)
		assert(cBUFLEN >= msgLengths[n]);

	reset();
	m_childView = childView;
	m_serverSocket = ss;
	m_serverNr = sn;
	m_validModelState = false;
	m_model[0] = 0;
	m_model[1] = 0;
	m_locModel[0] = m_locModel[1] = 0;
	m_viewer = 0;
}

CCommSocket::~CCommSocket()
{
	Close();
}

void CCommSocket::OnSend(int nError)
{
//	TRACE("CCommSocket::OnSend - threadId = %i\n", GetCurrentThreadId());

	if (nError) TRACE("OnSend Error %i\n",nError);
	TRACE("OnSend\n");
	isReady = true;
}

void CCommSocket::sendMessage(char typ, const char *p)
{
	int sent = Send(&typ, 1);
	if (sent == SOCKET_ERROR)
	{
		int err = GetLastError();
		TRACE("sendMessage: type send error %i\n", err);
		//if (err == WSAECONNRESET || err == WSAECONNABORTED) return; // spojeni umrelo			
		return; // nema ceny posilat dal
	}
	else
	{
		assert(sent>0); // je to tak v dokumentaci MFC!
		//TRACE("Sent message typ %i (1 byte)\n", typ);	
	}
	assert(0 <= typ && typ < LAST_MESSAGE); // povoleny typ
	int i = 0, length = msgLengths[typ];	
	do
	{
		sent = Send(p+i, length-i);
		if (sent == SOCKET_ERROR)
		{
			int err = GetLastError();
			TRACE("Send error %i\n", err);
			if (err == WSAECONNRESET || err == WSAECONNABORTED || err == WSAENOTSOCK ||
				err == WSAENOTCONN) break; // spojeni umrelo
		}
		else
		{
			assert(sent>0); // je to tak v dokumentaci MFC!
			//TRACE("Sent %i bytes\n", sent);
			i += sent;		
		}
	} while (i < length);
}

// posila novy stav sermiru:
void CCommSocket::sendModelState(const ModelState *ms[2])
{
	char buf[sizeof(ModelState)*2];
	memcpy(buf, ms[0], sizeof(ModelState));
	memcpy(buf + sizeof(ModelState), ms[1], sizeof(ModelState));
	sendMessage(MSG_MODELSSTATE, buf);
}

void CCommSocket::sendInfo(char err)
{
	sendMessage(MSG_INFO, &err);
}

// posila zpravu o zasazeni trojuhelnicku 'tri' hrace 'player':
void CCommSocket::sendTriangleHit(char player, int tri, byte value)
{
	char buf[6];
	buf[0] = player;
	buf[1] = value;
	*(int *)(&(buf[2])) = tri;
	sendMessage(MSG_TRIANGLEHIT, buf);
}

void CCommSocket::sendInputEvent(const InputEvent &ie)
{
	sendMessage(MSG_INPUTEVENT, (const char *)&ie);
}

// nastavi posledni kompletni stav sermire 'n' do promenne 'ms'
bool CCommSocket::setLastModelState(int n, ModelState &ms)
{
	CSingleLock lock(&m_mutex);
	m_mutex.Lock();
	ms = m_modelState[n];
	m_mutex.Unlock();
	return m_validModelState;
}

void CCommSocket::sendJudgeMsg(int score0, int score1, const char *msg)
{
	JudgeMessage jm;
	jm.score0 = score0; jm.score1=score1;
	strncpy(jm.msg, msg, cMaxJudgeMessage);
	sendMessage(MSG_JUDGEMSG, (const char *)&jm);
	TRACE("CCommSocket::sendJudgeMsg %i %i %s\n", score0, score1, msg);
}

// predpokladam, ze v OnReceivu nemuze prijit dalsi OnReceive
void CCommSocket::OnReceive(int nErrorCode)
{
//	TRACE("OnReceive\n");
//	TRACE("CCommSocket::OnReceive - threadId = %i\n", GetCurrentThreadId());

	if (nErrorCode) TRACE("OnReceive Error %i\n",nErrorCode);	
	else
	{		
		if (receiving == TYPE)
		{ // cekam na typ zpravy
			int r=Receive(&wrkMsg.type, 1);
			if (r == 0) 
			{
				TRACE("OnReceive connection reset (receiving type)\n");
				return; // connection closed
			}
			else if (r == SOCKET_ERROR)
			{
				TRACE("Type receive error %i\n", GetLastError());
			}
			else
			{
				assert(0 <= wrkMsg.type && wrkMsg.type < LAST_MESSAGE);
				//TRACE("Received type %i\n", wrkMsg.type);
				receiving = BODY;
			}
		}
		if (receiving == BODY)
		{ // uz prijimam telo zpravy
			int r=Receive(wrkMsg.body+bi, msgLengths[wrkMsg.type]-bi);
			if (r == 0) 
			{
				TRACE("OnReceive connection reset (bi = %i)\n", bi);
				return; // connection closed
			}
			else if (r == SOCKET_ERROR)
			{
				int err = GetLastError();
				if (err != WSAEWOULDBLOCK) // mohlo se stat, ze dalsi paket jeste nedorazil
					TRACE("Receive error %i\n", GetLastError());
			}
			else
			{
				//TRACE("Received %i bytes\n", r);
				bi += r;
				assert(bi <= msgLengths[wrkMsg.type]);
				if (bi == msgLengths[wrkMsg.type])
				{															
					// dorucena cela zprava; podle typu ji zpracuju:
					switch (wrkMsg.type)
					{
					case MSG_INFO:						
						if (wrkMsg.body[0] == ERRMSG_TOOMANYCLIENTS) // vypisu chybu:
						{
							if (m_childView) ((CChildView *)m_childView)->onClientFailed();
							MessageBox(NULL, "Connection refused: too many clients\n", "Error", MB_ICONSTOP);
						}
						else if (wrkMsg.body[0] == INITMSG_QUALITY0)
						{
							if (m_childView) ((CChildView *)m_childView)->onClientInit(0);
						}
						else if (wrkMsg.body[0] == INITMSG_QUALITY1)
						{
							if (m_childView) ((CChildView *)m_childView)->onClientInit(1);
						}
						else assert(false); 
						break;
					case MSG_MODELSSTATE:
						// prepisu posledni stavy modelu:
						{
							CSingleLock lock(&m_mutex);
							lock.Lock();
							memcpy(&(m_modelState[0]), wrkMsg.body, sizeof(ModelState));
							memcpy(&(m_modelState[1]), wrkMsg.body + sizeof(ModelState), sizeof(ModelState));
							lock.Unlock();
							m_validModelState = true; // uz je tu
						}
						break;
					case MSG_TRIANGLEHIT:
						{
							int player = wrkMsg.body[0];
							assert(player == 0 || player ==1); // cislo hrace
							byte value = wrkMsg.body[1];
							int tri = *(int *)&(wrkMsg.body[2]);
							if (m_model[player]) m_model[player]->setTriangleHit(tri, value);
						}
						break;
					case MSG_INPUTEVENT:
						{
							InputEvent *ie = (InputEvent *)wrkMsg.body;							
							int pl = ie->player;
							assert(pl == 0 || pl == 1);
							//TRACE("InputEvent typu %i pro hrace %i\n", ie->type, pl);
							// kriticka sekce je v feedNewPos, feedButton a KeyboardEvent
							if (ie->type == cPosition)
							{								
								if (m_locModel[pl]) m_locModel[pl]->m_mid->feedNewPos(ie->x, ie->y);
							}
							else if (ie->type == cButton)
							{
								if (m_locModel[pl]) m_locModel[pl]->m_mid->feedButton(ie->data);
							}
							else if (ie->type == cKey)
							{
								if (m_locModel[pl]) m_locModel[pl]->m_kid->KeyboardEvent(ie->flag, ie->data);
							}
							else assert(false && "Spatny typ ve zprave InputEvent");
						}
						break;
					case MSG_JUDGEMSG:
						{
							JudgeMessage *jm = (JudgeMessage *)wrkMsg.body;
							TRACE("Prisla judgeMessage %i:%i, %s\n", jm->score0, jm->score1, jm->msg);
							CString score;
							if (jm->score0 < 0) score.Empty();
							else score.Format("%i : %i", jm->score0, jm->score1);
							if (m_viewer) ((CModeViewer *)m_viewer)->setTexts(score, jm->msg);
						}
						break;
					default: assert(false&&"Prisla spatna zprava");
					}					
					bi = 0;
					receiving = TYPE;					
				}			
			}
		}
	}
}

void CCommSocket::OnConnect(int nErrorCode)
{
	if (nErrorCode) 
	{
		if (m_childView) ((CChildView *)m_childView)->onClientFailed();
		MessageBox(NULL, "Connection failed", "Error", MB_ICONSTOP);
		TRACE("Connection error %i\n",nErrorCode);	
	}
	else
	{
		bi = 0; // prazdny buffer
		receiving = TYPE;
		TRACE("Connected OK");
		// init pro CChildView tady jeste ne, ten poslu az dostanu prvni zpravu
	}
}

void CCommSocket::OnClose(int nErrorCode)
{
	TRACE("OnClose\n");
	if (nErrorCode) TRACE("OnClose error %i\n",nErrorCode);
	// pokud patri k servru, tak by se server mel dozvedet, ze spojeni
	// umrelo, aby uvolnil soket pro dalsiho klienta:
	Close();
	if (m_serverSocket)
	{
		m_serverSocket->clientDied(m_serverNr);	
		// POZOR, v clientDied se provadi delete this objektu!		
	}
	else
	{ // jinak jsem soket na strane klienta:
		MessageBox(NULL, "Server has closed connection", "Client panic", MB_ICONSTOP);
		if (m_childView) ((CChildView *)m_childView)->Disconnect();		
	}
}

void CCommSocket::startClient(const char *host, int port)
{
	int err;	
	Close(); // byl-li otevren
	if (!Create(0, SOCK_STREAM))
	{
		MessageBox(0, "Can't create client socket", "Error", MB_ICONSTOP);
		err=GetLastError();		
		TRACE("Create error %i\n",err);
		return;
	}	

	bool nodelay = true;
	if (!SetSockOpt(TCP_NODELAY, &nodelay, sizeof(bool), IPPROTO_TCP))
	{
		TRACE("CCommSocket: Nepovedlo se SetSockOpt na TCP_NODELAY = true\n");
	}
	else TRACE("CCommSocket: SetSockOpt TCP_NODELAY = true succeeded\n");

	if (!Connect(host, port))
	{
		err=GetLastError();
		if (err != WSAEWOULDBLOCK)
		{
			MessageBox(0, "Connection failed", "Error", MB_ICONSTOP);
			TRACE("Connect error %i\n",err);
		}
	}
}