/* 
    DirectInput.cpp

	DirectInput class implementation

    Author: Ladislav Kavan, 2002 (based on DirectX 7.0a SDK sample)
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "gl1.h"
#include <assert.h>
#include "DirectInput.h"
#include "MainFrm.h"

// pocitani joysticku:
BOOL CALLBACK EnumJoysticksCallbackCnt( const DIDEVICEINSTANCE* pdidInstance,
			 	                     VOID* pContext )
{
	int hm=pdidInstance->dwSize; hm++; //antiwarning
	CDirectInput *toto = (CDirectInput *)pContext;
	toto->m_numJoysticks++;
    return DIENUM_CONTINUE;
}

// vytvareni joystikovych objektu a jejich ukladani do pole:
BOOL CALLBACK EnumJoysticksCallbackWrite( const DIDEVICEINSTANCE* pdidInstance,
			 	                     VOID* pContext )
{
    HRESULT hr;
	CDirectInput *toto = (CDirectInput *)pContext;

    // Obtain an interface to the enumerated joystick.
	assert(toto->m_counter < toto->m_numJoysticks);
    hr = toto->g_pDI->CreateDeviceEx( pdidInstance->guidInstance, IID_IDirectInputDevice2,
		                 (VOID**)&(toto->m_pJoystick[toto->m_counter].g_pJoystick), NULL );

	toto->m_pJoystick[toto->m_counter].text = CString(pdidInstance->tszInstanceName);

	if (!FAILED(hr)) toto->m_counter++;
	return DIENUM_CONTINUE;
}

BOOL CALLBACK EnumAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi,
                                VOID* pContext )
{	
	CDirectInput *toto = (CDirectInput *)pContext;    

    DIPROPRANGE diprg; 
    diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
    diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
    diprg.diph.dwHow        = DIPH_BYOFFSET; 
    diprg.diph.dwObj        = pdidoi->dwOfs; // Specify the enumerated axis
    diprg.lMin              = -1000; 
    diprg.lMax              = +1000; 
    
    // Set the range for the axis
    if( FAILED( toto->m_pJoystick[toto->m_counter].g_pJoystick->SetProperty( DIPROP_RANGE, &diprg.diph ) ) )
        return DIENUM_STOP;

    return DIENUM_CONTINUE;
}


bool CDirectInput::initDirectInput()
{
    HRESULT hr;	

    // Register with the DirectInput subsystem and get a pointer
    // to a IDirectInput interface we can use.
    hr = DirectInputCreateEx( m_hInst, DIRECTINPUT_VERSION,IID_IDirectInput7, (LPVOID*)&g_pDI, NULL );
    if( FAILED(hr) ) 
        return false;

	// spocitani joysticku:
	m_numJoysticks = 0;
	hr = g_pDI->EnumDevices( DIDEVTYPE_JOYSTICK, EnumJoysticksCallbackCnt,
                             (VOID *)this, DIEDFL_ATTACHEDONLY );
	if( FAILED(hr) ) return false;

	m_pJoystick = new JoystickInfo[m_numJoysticks];
	/* -- presunuto do konstruktoru
	for (int n=0; n<m_numJoysticks; n++) 
	{
		m_pJoystick[n].g_pJoystick = NULL;		
		m_pJoystick[n].isFF = 0;
		m_pJoystick[n].initialized = false;
		m_pJoystick[n].g_effConstantForce = NULL;
	}
	*/
    m_counter = 0;
	// vytvoreni a zapsani do pole:
    hr = g_pDI->EnumDevices( DIDEVTYPE_JOYSTICK, EnumJoysticksCallbackWrite,
                             (VOID *)this, DIEDFL_ATTACHEDONLY );
    if( FAILED(hr) ) return false;
	m_numJoysticks = m_counter; // nektere mohly odpadnout

    // Make sure we got a joystick
    if( m_numJoysticks == 0 )
    {
        MessageBox( NULL, "No joystick found", "Error", 
                    MB_ICONERROR | MB_OK );
        return false;
    }

	if (m_numJoysticks >= 8) TRACE("Warning: mam vyhrazeno jen 8 ID do menu pro joysticky a joysticku je %i\n", m_numJoysticks);

	// pridame do m_menu:
	CMenu *inputMenu0 = ((CMainFrame *)m_mainfrm)->getSubMenu(ID_INPUT0_NOTHING);
	assert(inputMenu0);
	CMenu *inputMenu1 = ((CMainFrame *)m_mainfrm)->getSubMenu(ID_INPUT1_NOTHING);
	assert(inputMenu1);
	for (int n=0; n<m_numJoysticks; n++)
	{		
		inputMenu0->AppendMenu(MF_STRING, ID_INPUT0_RESERVED0 + n, m_pJoystick[n].text);
		inputMenu1->AppendMenu(MF_STRING, ID_INPUT1_RESERVED0 + n, m_pJoystick[n].text);		
	}
	inputMenu0->AppendMenu(MF_SEPARATOR);
	inputMenu1->AppendMenu(MF_SEPARATOR);

    return true;
}

void CDirectInput::freeJoystick(int j)
{
	assert(0<=j && j<m_numJoysticks);
	assert(m_pJoystick[j].initialized);
	m_pJoystick[j].g_pJoystick->Unacquire();
	m_pJoystick[j].initialized=false;
}

//BOOL CALLBACK DIEnumEffectsCallback(  LPCDIEFFECTINFO pdei,  LPVOID pvRef )
//{
//	return DIENUM_CONTINUE;
//}



bool CDirectInput::initJoystick(int j)
{
	HRESULT hr;
	assert(0<=j && j<m_numJoysticks);
	if (m_pJoystick[j].initialized) freeJoystick(j);
	m_pJoystick[j].initialized = false; // kdyby se nepovedlo inicializovat
	// Set the data format to "simple joystick" - a predefined data format 
	//
	// A data format specifies which controls on a device we are interested in,
	// and how they should be reported. This tells DInput that we will be
	// passing a DIJOYSTATE structure to IDirectInputDevice::GetDeviceState().
	hr = m_pJoystick[j].g_pJoystick->SetDataFormat( &c_dfDIJoystick );
	if( FAILED(hr) ) return false;
	
	// Set the cooperative level to let DInput know how this device should
	// interact with the system and with other DInput applications.
	// kvuli FF je potreba exkluzivni pristup:
	hr = m_pJoystick[j].g_pJoystick->SetCooperativeLevel( m_winHandle, DISCL_EXCLUSIVE|DISCL_BACKGROUND );
	if( FAILED(hr) ) return false;
	
	// Determine how many axis the joystick has (so we don't error out setting
	// properties for unavailable axis)
	m_pJoystick[j].g_diDevCaps.dwSize = sizeof(DIDEVCAPS);
	hr = m_pJoystick[j].g_pJoystick->GetCapabilities(&(m_pJoystick[j].g_diDevCaps));
	if( FAILED(hr) ) return false;	
	
	m_pJoystick[j].isFF = m_pJoystick[j].g_diDevCaps.dwFlags&&DIDC_FORCEFEEDBACK;
	
	// Enumerate the axes of the joyctick and set the range of each axis. Note:
	// we could just use the defaults, but we're just trying to show an example
	// of enumerating device objects (axes, buttons, etc.).
	m_counter = j;
	m_pJoystick[j].g_pJoystick->EnumObjects( EnumAxesCallback, (VOID*)this, DIDFT_AXIS );

	if (m_pJoystick[j].isFF)
	{ // inicializujeme force-feedback
		// Since we will be playing force feedback effects, we should disable the
		// auto-centering spring.
		DIPROPDWORD dipdw;
		dipdw.diph.dwSize       = sizeof(DIPROPDWORD);
		dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		dipdw.diph.dwObj        = 0;
		dipdw.diph.dwHow        = DIPH_DEVICE;
		dipdw.dwData            = FALSE;

		hr = m_pJoystick[j].g_pJoystick->SetProperty( DIPROP_AUTOCENTER, &dipdw.diph );
		if( FAILED(hr) ) m_pJoystick[j].isFF = false;

		// vyrobim prvni efekt: konstatni sila:
		DWORD    rgdwAxes[2] = { DIJOFS_X, DIJOFS_Y };
		LONG     rglDirection[2] = { 1, 0 };
		DICONSTANTFORCE cf = { 0 }; // nulova velikost na zacatku

		DIEFFECT eff;
		ZeroMemory( &eff, sizeof(eff) );
		eff.dwSize                  = sizeof(DIEFFECT);
		eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
		eff.dwDuration              = INFINITE;
		eff.dwSamplePeriod          = 0;
		eff.dwGain                  = DI_FFNOMINALMAX;
		eff.dwTriggerButton         = DIEB_NOTRIGGER;
		eff.dwTriggerRepeatInterval = 0;
		eff.cAxes                   = 2;
		eff.rgdwAxes                = rgdwAxes;
		eff.rglDirection            = rglDirection;
		eff.lpEnvelope              = 0;
		eff.cbTypeSpecificParams    = sizeof(DICONSTANTFORCE);
		eff.lpvTypeSpecificParams   = &cf;
		eff.dwStartDelay            = 0;

		// Create the prepared effect
		hr = m_pJoystick[j].g_pJoystick->CreateEffect( GUID_ConstantForce, &eff, 
			&m_pJoystick[j].g_effConstantForce, NULL );
		if( FAILED(hr) ) TRACE("nepovedlo se vytvorit efekt konstatni sily\n");
		
		// druhy efekt: treni
		//m_pJoystick[j].g_pJoystick->EnumEffects(DIEnumEffectsCallback, this, DIEFT_ALL);

		DICONDITION cond[2];
		cond[0].lOffset = 0;
		cond[0].dwNegativeSaturation = 10000;
		cond[0].dwPositiveSaturation = 10000;
		cond[0].lDeadBand = 0;
		cond[0].lNegativeCoefficient = 0;//10000;
		cond[0].lPositiveCoefficient = 0;//10000;
		cond[1] = cond[0];

		ZeroMemory( &eff, sizeof(eff) );
		eff.dwSize                  = sizeof(DIEFFECT);
		eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
		eff.dwDuration              = INFINITE;
		eff.dwSamplePeriod          = 0;
		eff.dwGain                  = DI_FFNOMINALMAX;
		eff.dwTriggerButton         = DIEB_NOTRIGGER;
		eff.dwTriggerRepeatInterval = 0;
		eff.cAxes                   = 2;
		eff.rgdwAxes                = rgdwAxes;
		eff.rglDirection            = rglDirection;
		eff.lpEnvelope              = 0;
		eff.cbTypeSpecificParams    = sizeof(DICONDITION)*2;
		eff.lpvTypeSpecificParams   = &cond;
		eff.dwStartDelay            = 0;
		hr = m_pJoystick[j].g_pJoystick->CreateEffect( GUID_Friction, &eff, 
			&m_pJoystick[j].g_effFriction, NULL );
		if( FAILED(hr) ) TRACE("nepovedlo se vytvorit efekt treni\n");

		// treti efekt: rampova sila

		DIRAMPFORCE rf;
		rf.lStart = 0;
		rf.lEnd = 0;

		ZeroMemory( &eff, sizeof(eff) );
		eff.dwSize                  = sizeof(DIEFFECT);
		eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
		eff.dwDuration              = 0; // mikrosekundy!
		eff.dwSamplePeriod          = 0;
		eff.dwGain                  = DI_FFNOMINALMAX;
		eff.dwTriggerButton         = DIEB_NOTRIGGER;
		eff.dwTriggerRepeatInterval = 0;
		eff.cAxes                   = 2;
		eff.rgdwAxes                = rgdwAxes;
		eff.rglDirection            = rglDirection;
		eff.lpEnvelope              = 0;
		eff.cbTypeSpecificParams    = sizeof(DIRAMPFORCE);
		eff.lpvTypeSpecificParams   = &rf;
		eff.dwStartDelay            = 0;
		hr = m_pJoystick[j].g_pJoystick->CreateEffect( GUID_RampForce, &eff, 
			&m_pJoystick[j].g_effRamp, NULL );
		if( FAILED(hr) ) TRACE("nepovedlo se vytvorit efekt ramp force\n");

		// ctvrty efekt: setrvacnost
		cond[0].lOffset = 0;
		cond[0].dwNegativeSaturation = 10000;
		cond[0].dwPositiveSaturation = 10000;
		cond[0].lDeadBand = 0;
		cond[0].lNegativeCoefficient = 10000;
		cond[0].lPositiveCoefficient = 10000;
		cond[1] = cond[0];

		ZeroMemory( &eff, sizeof(eff) );
		eff.dwSize                  = sizeof(DIEFFECT);
		eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
		eff.dwDuration              = INFINITE;
		eff.dwSamplePeriod          = 0;
		eff.dwGain                  = DI_FFNOMINALMAX;
		eff.dwTriggerButton         = DIEB_NOTRIGGER;
		eff.dwTriggerRepeatInterval = 0;
		eff.cAxes                   = 2;
		eff.rgdwAxes                = rgdwAxes;
		eff.rglDirection            = rglDirection;
		eff.lpEnvelope              = 0;
		eff.cbTypeSpecificParams    = sizeof(DICONDITION)*2;
		eff.lpvTypeSpecificParams   = &cond;
		eff.dwStartDelay            = 0;
		hr = m_pJoystick[j].g_pJoystick->CreateEffect( GUID_Inertia, &eff, 
			&m_pJoystick[j].g_effInertia, NULL );
		if( FAILED(hr) ) TRACE("nepovedlo se vytvorit efekt setrvacnosti\n");		

		if (!m_pJoystick[j].isFF) TRACE("selhala inicializace force-feedbacku\n");
	}
	if (!acquire(j)) return false;
	m_pJoystick[j].initialized = true;

	if (m_pJoystick[j].isFF)
	{ // musi se udelat po acquire:
		setConstantForce(j, 0, -1, 3000); // defaultni sila - gravitace

		hr = m_pJoystick[j].g_effInertia->Start(1, 0);
		if (FAILED(hr)) TRACE("nepovedlo se spustit efekt setrvacnosti\n");
	}

	return true;
}

// ziskame (zpet nebo poprve) zarizeni:
bool CDirectInput::acquire(int j)
{
	assert(0<=j && j<m_numJoysticks);
	if (m_pJoystick[j].g_pJoystick)
	{
		HRESULT hr = m_pJoystick[j].g_pJoystick->Acquire();
		if (FAILED(hr)) return false;
		else 
		{
			if (m_pJoystick[j].isFF)
			{ // v pripade FF zapneme constant force effect:
				assert(m_pJoystick[j].g_effConstantForce);
				HRESULT hr = m_pJoystick[j].g_effConstantForce->Start( 1, 0 ); // Start the effect
				if (FAILED(hr)) TRACE("Nepodarilo se spustit konstatni silu\n");
				assert(m_pJoystick[j].g_effFriction);
				hr = m_pJoystick[j].g_effFriction->Start( 1, 0 ); // Start the effect
				if (FAILED(hr)) TRACE("Nepodarilo se spustit treni\n");
				assert(m_pJoystick[j].g_effRamp);
				hr = m_pJoystick[j].g_effRamp->Start( 1, 0 ); // Start the effect
				if (FAILED(hr)) TRACE("Nepodarilo se spustit ramp silu\n");
			}
			return true;
		}
	}
	else return false;
}

void CDirectInput::setConstantForce(int j, long x, long y, int magnitude)
{
	assert(m_pJoystick[j].initialized);
	if (!m_pJoystick[j].isFF) return;
	// Modifying an effect is basically the same as creating a new one, except
	// you need only specify the parameters you are modifying
    LONG rglDirection[2] = { -x, y }; // konverze na standardni souradnice (na rozdil od DirectX udavam smer sily joysticku, ne uzivatele)

    DICONSTANTFORCE cf = {magnitude};    

    DIEFFECT eff;
    ZeroMemory( &eff, sizeof(eff) );
    eff.dwSize                = sizeof(DIEFFECT);
    eff.dwFlags               = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
    eff.cAxes                 = 2;
    eff.rglDirection          = rglDirection;
    eff.lpEnvelope            = 0;
    eff.cbTypeSpecificParams  = sizeof(DICONSTANTFORCE);
    eff.lpvTypeSpecificParams = &cf;

    // Now set the new parameters and start the effect immediately.
    m_pJoystick[j].g_effConstantForce->SetParameters( &eff, DIEP_DIRECTION|
		                             DIEP_TYPESPECIFICPARAMS|DIEP_START );
}

void CDirectInput::setFrictionCoefficient(int j, int coef)
{
	assert(m_pJoystick[j].initialized);
	if (!m_pJoystick[j].isFF) return;

    DICONDITION cond[2];
	cond[0].lOffset = 0;
	cond[0].dwNegativeSaturation = 10000;
	cond[0].dwPositiveSaturation = 10000;
	cond[0].lDeadBand = 0;
	cond[0].lNegativeCoefficient = coef;
	cond[0].lPositiveCoefficient = coef;
	cond[1] = cond[0];

	DIEFFECT eff;
	ZeroMemory( &eff, sizeof(eff) );
	eff.dwSize                  = sizeof(DIEFFECT);
	eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
	eff.cAxes                   = 2;	
	eff.cbTypeSpecificParams    = sizeof(DICONDITION)*2;
	eff.lpvTypeSpecificParams   = &cond;	

    // Now set the new parameters and start the effect immediately.
    m_pJoystick[j].g_effFriction->SetParameters( &eff,
		                             DIEP_TYPESPECIFICPARAMS|DIEP_START );
}

// nastavi ramp silu s pocatecni silou 'startF', koncovou 'endF' a trvajici 'duration' mikrosekund:
void CDirectInput::setRampForce(int j, int x, int y, int startF, int endF, int duration)
{
	assert(m_pJoystick[j].initialized);
	if (!m_pJoystick[j].isFF) return;
/* zoufale zjistovani proc to neprochazi ... bylo potreba nastavi eff.cAxes na 2
	DIEFFECTINFO info;
	info.dwSize = sizeof(info);
	HRESULT hr1 = m_pJoystick[j].g_pJoystick->GetEffectInfo(&info, GUID_ConstantForce);

	HRESULT err1 = DIERR_NOTINITIALIZED;
	HRESULT err2 = DIERR_INCOMPLETEEFFECT;
	HRESULT err3 = DIERR_INPUTLOST ;
	HRESULT err4 = DIERR_INVALIDPARAM ;
	HRESULT err5 = DIERR_EFFECTPLAYING ;	
*/

	LONG rglDirection[2] = { -x, y }; // konverze na standardni souradnice
	DIRAMPFORCE rf; rf.lStart = startF; rf.lEnd = endF;	
	DIEFFECT eff;
	ZeroMemory( &eff, sizeof(eff) );
	eff.dwSize                  = sizeof(DIEFFECT);
	eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
	eff.cAxes					= 2;
	eff.dwDuration              = duration;				
	eff.rglDirection            = rglDirection;	
	eff.cbTypeSpecificParams    = sizeof(DIRAMPFORCE);
	eff.lpvTypeSpecificParams   = &rf;	

	HRESULT hr;
	hr = m_pJoystick[j].g_effRamp->SetParameters(&eff, DIEP_DIRECTION|DIEP_DURATION|
		DIEP_TYPESPECIFICPARAMS|DIEP_START);	
	if (hr)
	{
		TRACE("CDirectInput::setRampForce returns %i\n", hr);
	}
}

CDirectInput::CDirectInput(HWND winHandle, HINSTANCE hInst, CFrameWnd *mainfrm)
{
	m_winHandle = winHandle;
	m_hInst = hInst;
	m_mainfrm = mainfrm;
	if (!initDirectInput())
	{
		MessageBox(0, "DirectInput initialization failed", "Error", MB_OK);
	}
}

// zapise do X, Y aktualni polohu (od -1 do 1) joysticku 'j' a do B1, B2 stav tlacitek
// vraci jestli je vse OK
bool CDirectInput::getCurrentState(float *X, float *Y, bool *B1, bool *B2, int j)
{
    HRESULT     hr;
	//    CHAR        strText[128]; // Device state text
    DIJOYSTATE  js;           // DInput joystick state 
	//    CHAR*       str;

	assert(0<=j && j<m_numJoysticks);
    assert(m_pJoystick[j].initialized);
    
	do
	{
		// Poll the device to read the current state
		hr = m_pJoystick[j].g_pJoystick->Poll();
		if( FAILED(hr) ) return false;
		
		// Get the input's device state
		hr = m_pJoystick[j].g_pJoystick->GetDeviceState( sizeof(DIJOYSTATE), &js );
		
		if( hr == DIERR_INPUTLOST )
		{
			TRACE("zajimavy, DIRECT INPUT LOST\n");
			// DInput is telling us that the input stream has been
			// interrupted. We aren't tracking any state between polls, so
			// we don't have any special reset that needs to be done. We
			// just re-acquire and try again.			
			if (!acquire(j)) return false;			
		}
	}
	while( DIERR_INPUTLOST == hr );
	
	if( FAILED(hr) ) return false;

	*X = (float)js.lX/1000;
	*Y = -(float)js.lY/1000;
	assert(m_pJoystick[j].attackButton>=0 && m_pJoystick[j].attackButton<32);
	assert(m_pJoystick[j].parryButton>=0 && m_pJoystick[j].parryButton<32);
	*B1 = js.rgbButtons[m_pJoystick[j].attackButton] >= 0x80;
	*B2 = js.rgbButtons[m_pJoystick[j].parryButton] >= 0x80;

	return true;
} 

void CDirectInput::selectButton(int j, int akce)
{
   HRESULT     hr;	
   DIJOYSTATE  js;           // DInput joystick state 
   int button = -1;

	assert(0<=j && j<m_numJoysticks);
    assert(m_pJoystick[j].initialized);
    
	do
	{
		// Poll the device to read the current state
		hr = m_pJoystick[j].g_pJoystick->Poll();
		if( FAILED(hr) ) return;
		
		// Get the input's device state
		hr = m_pJoystick[j].g_pJoystick->GetDeviceState( sizeof(DIJOYSTATE), &js );
		
		if( hr == DIERR_INPUTLOST )
		{
			TRACE("zajimavy, DIRECT INPUT LOST\n");			
			if (!acquire(j)) return;
		}

		for (int n=0; n<32; n++)
		{
			if (js.rgbButtons[n] >= 0x80)
			{
				button = n;
				break;
			}
		}
	}
	while( button == -1 );
		
	if (akce == cAttack) m_pJoystick[j].attackButton = button;
	else if (akce == cParry) m_pJoystick[j].parryButton = button;
	else assert(false);
}

CDirectInput::~CDirectInput()
{
	for (int n=0; n<m_numJoysticks; n++)
		if (m_pJoystick[n].initialized) freeJoystick(n);
	delete[] m_pJoystick;
}
