/* 
    DirectInput.h

	DirectInput class

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef MYDIRECTINPUT_H
#define MYDIRECTINPUT_H

#include <dinput.h>
#include <assert.h>

const int cAttack = 1, cParry = 2;

//enum { cEffConstantForce = 0,
//	   cEffLast
//};

typedef class CJoystickInfo {
public:
	LPDIRECTINPUTDEVICE2 g_pJoystick;
	CString text;
	bool initialized; // true byl-li inicializovan pmci initJoystick
// nasledujici je validni pouze pro zvoleny joystick:
	DIDEVCAPS            g_diDevCaps;
	bool	isFF; // podporuje force-feedback?	
	LPDIRECTINPUTEFFECT g_effConstantForce,
						g_effFriction,
						g_effRamp,
						g_effInertia; // efekty, pouze pokud je isFF = true
	int attackButton, parryButton; // nastaveni butonu

	CJoystickInfo() 
	{
		g_pJoystick = NULL;		
		isFF = false;
		initialized = false;
		g_effConstantForce = NULL;
		attackButton = 0;
		parryButton = 1;
	}
} JoystickInfo;

// zamerena specialne na joystick:
class CDirectInput
{
private:
	HWND	m_winHandle;
	HINSTANCE m_hInst;
	CFrameWnd *m_mainfrm; // kam pridat seznam joysticku

	LPDIRECTINPUT7       g_pDI;         
	int m_numJoysticks, m_counter;
	JoystickInfo *m_pJoystick; // seznam vsech joysticku	

	// pujci si objekt directInput a joystick:
	bool initDirectInput();

	// na spocitani joysticku, abych moh alokovat pole m_pJoystick
	friend BOOL CALLBACK EnumJoysticksCallbackCnt( const DIDEVICEINSTANCE* pdidInstance,
			 					                VOID* pContext );

	// ulozeni info o joystikach 
	friend BOOL CALLBACK EnumJoysticksCallbackWrite( const DIDEVICEINSTANCE* pdidInstance,
			 					                VOID* pContext );

	// nastavi spravne meritka os:
	friend BOOL CALLBACK EnumAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi,
					                       VOID* pContext );

	//friend BOOL CALLBACK DIEnumEffectsCallback(  LPCDIEFFECTINFO pdei,  LPVOID pvRef );
	// ziskame (zpet nebo poprve) joystick cislo m_pJoystick[j]:
	bool acquire(int j);

public:
	CDirectInput(HWND winHandle, HINSTANCE hInst, CFrameWnd *mainfrm);

	// inicializuje a nastavi j-ty joystick z pole m_pJoystick
	bool initJoystick(int j);

	// uvolni j-ty joystick:
	void freeJoystick(int j);
	
	// zapise do X, Y aktualni polohu (od -1 do 1) joysticku 'j' a do B1, B2 stav tlacitek
	// vraci jestli je vse OK
	bool CDirectInput::getCurrentState(float *X, float *Y, bool *B1, bool *B2, int j);
	~CDirectInput();

	// nastavi silu pro joystick 'j' ve smeru 'x','y' velikosti 'magnitude' (-10000 az 10000):	
	// smer sily [x, y] je v standardnich kartezskych souradnicich (ne jako v DirectX):
	void setConstantForce(int j, long x, long y, int magnitude);

	// nastavi "koeficient treni" (od -10000 do 10000) pro joystick 'j':
	void setFrictionCoefficient(int j, int coef);

	// nastavi ramp silu ve smeru [x, y] s pocatecni silou 'startF', koncovou 'endF' a trvajici 
	// 'duration' mikrosekund:
	void setRampForce(int j, int x, int y, int startF, int endF, int duration);

	int getNumJoysticks() { return m_numJoysticks; }	

	// tuhle funkci zavola CButtonConfigDialog az se klikne na volbu buttonu;
	// pocka na stisknuti butonu a pak ho priradi k akci 'akce' joysticku 'j':
	void selectButton(int j, int akce);

	const CString getJoystickName(int j)
	{
		assert(0<=j && j<m_numJoysticks);
		assert(m_pJoystick[j].initialized);
		return m_pJoystick[j].text;
	}

	int getAttackButton(int j)
	{
		assert(0<=j && j<m_numJoysticks);
		assert(m_pJoystick[j].initialized);
		return m_pJoystick[j].attackButton;
	}

	int getParryButton(int j)
	{
		assert(0<=j && j<m_numJoysticks);
		assert(m_pJoystick[j].initialized);
		return m_pJoystick[j].parryButton;
	}
};

// toto je zpusob jak dostat do callbacku ukazatel na volajici tridu:
typedef struct SBlbejCallback
{
	CDirectInput *toto;
	void *pContect;
} BlbejCallback;



#endif