/* 
    IK.cpp

	Inverse Kinematics interface for animating model's arm.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "IK.h"
#include "Model.h"
#include "Vector.h"
#include "Matrix.h"

IK::IK(Model *m) : m_model(m)  {
};

IK::~IK()  {
};

// aby to fungovalo, musi byt spocitane m_finaly
void IK::getEEpos(Vector &e)
{	
	Matrix &mSwordholder = m_model->m_pJoints[m_model->m_swordholder].m_final;
	e.reset();
	e.transform( mSwordholder );
}

void IK::getEEor1(Vector &e)
{	
	float f[3];
	m_model->m_pJoints[m_model->m_swordholder].m_final.getColumn(2, f);
	e.set(f);
	e.mul(-1, -1, -1);
	e.normalize();
}

void IK::getEEor2(Vector &e)
{
	float f[3];
	m_model->m_pJoints[m_model->m_swordholder].m_final.getColumn(1, f);
	e.set(f);
	e.mul(-1, -1, -1);
	e.normalize();
}
/*

void JIK::cmpJacobian()
{
	int c;
	int shoulder = m_model->m_ikj[SHOULDER], elbow = m_model->m_ikj[ELBOW],
		hand = m_model->m_ikj[HAND], swordholder = m_model->m_swordholder;

	//m_model->setFinalMatrices( root_pos );
	//Matrix &mShoulder = m_model->m_pJoints[shoulder].m_final;
	//Matrix &mElbow = m_model->m_pJoints[elbow].m_final;
	//Matrix &mHand = m_model->m_pJoints[hand].m_final;
	Matrix m[3],mSwordholder;
	// prepocitam si finalni matice u kloubu paze:
	for (c = 0; c < 3; c++)
    {
		int j = m_model->m_ikj[c];
		m[c].set( m_model->m_pJoints[m_model->m_pJoints[j].m_parent].m_final.getMatrix() );
		m[c].postMultiply( m_model->m_pJoints[j].m_relative );
		Matrix transform;
		transform.setTranslation( m_model->m_pJoints[j].trans );
		transform.setRotationRadians( m_model->m_pJoints[j].rot );
		m[c].postMultiply( transform );
	}
	m_model->m_pJoints[shoulder].m_final.set(m[SHOULDER].getMatrix());
	m_model->m_pJoints[elbow].m_final.set(m[ELBOW].getMatrix());
	m_model->m_pJoints[hand].m_final.set(m[HAND].getMatrix());
	// jeste spocitam swordholder (EE)
	m_model->m_pJoints[swordholder].m_final.set(m[HAND].getMatrix());
	m_model->m_pJoints[swordholder].m_final.postMultiply(m_model->m_pJoints[swordholder].m_relative);
	// transformace ve swordholderu nejsou
	mSwordholder.set(m_model->m_pJoints[swordholder].m_final.getMatrix());

	Vector axis[DOF],or1,or2;
	float a[3];

// nastaveni os rotaci
    m[SHOULDER].getColumn(0, a); axis[0].set(a);
	m[SHOULDER].getColumn(1, a); axis[1].set(a);
	m[SHOULDER].getColumn(2, a); axis[2].set(a);
	m[ELBOW].getColumn(0, a); axis[3].set(a);
	m[ELBOW].getColumn(1, a); axis[4].set(a);
	m[HAND].getColumn(0, a); axis[5].set(a);
	m[HAND].getColumn(1, a); axis[6].set(a);
	m[HAND].getColumn(2, a); axis[7].set(a);
	for (int n = 0; n < DOF; n++) axis[n].normalize();
	// 1. orientacni vektor je ve smeru -z (smer ostri)
	mSwordholder.getColumn(2, a); or1.set(a); or1.mul(-1, -1, -1); or1.normalize(); //del
	getEEor1(or1); // pro kontrolu
	// 2. orientacni vektor je ve smeru -x (smer pricky nahoru)
	mSwordholder.getColumn(0, a); or2.set(a); or2.mul(-1, -1, -1); or2.normalize(); //del
	getEEor2(or2);

	Vector EE, posHand, posElbow, posShoulder;
	getEEpos(EE); // finalni matice nastaveny
	posShoulder.transform( m[SHOULDER] );
	posElbow.transform( m[ELBOW] );
	posHand.transform( m[HAND] ); 

	// pro Shoulder:
	for (c = 0; c < 3; c++)
	{  // c-ty sloupec
		Vector r(EE);
		r.sub(posShoulder);
		//p.normalize();
		Vector p(axis[c]);
		p.cross(r);				
		J[0][c] = p[0]; J[1][c] = p[1]; J[2][c] = p[2];
	}

	// pro Elbow
	for (c = 3; c < 5; c++)
	{  // c-ty sloupec
		Vector r(EE);
		r.sub(posElbow);
		//p.normalize();
		Vector p(axis[c]);
		p.cross(r);				
		J[0][c] = p[0]; J[1][c] = p[1]; J[2][c] = p[2];
	}

	// pro Hand
	for (c = 5; c < 8; c++)
	{  // c-ty sloupec
		Vector r(EE);
		r.sub(posHand);
		//p.normalize();
		Vector p(axis[c]);
		p.cross(r);
		J[0][c] = p[0]; J[1][c] = p[1]; J[2][c] = p[2];
	}

	// ted spocitam orientacni cast Jakobianu
	for (c = 0; c < 8; c++)
	{
		Vector p(axis[c]);
		p.cross(or1);
		J[3][c] = p[0]; J[4][c] = p[1]; J[5][c] = p[2];
		p.set(axis[c].getVector());
		p.cross(or2);
		J[6][c] = p[0]; J[7][c] = p[1]; J[8][c] = p[2];
	}
}

void JIK::Jmul(float *f, float *out)
{
	int n;
	for (n = 0; n < DOF; n++)
	{
		out[0] += J[0][n]*f[n];
		out[1] += J[1][n]*f[n];
		out[2] += J[2][n]*f[n];
	}
}

void JTIK::posGoal( Vector &goal )
{
	int shoulder = m_model->m_ikj[SHOULDER], elbow = m_model->m_ikj[ELBOW],
		hand = m_model->m_ikj[HAND];	
	int r;
	float dq[DOF]; // zmeny otoceni kloubu (rotace kolem os)

	TRACE("goal = [%f, %f, %f]\n",goal[0],goal[1],goal[2]);

	int iter = 100;

	for (int i = 0; i < iter; i++ )
	{
		cmpJacobian();

		Vector force(goal), EE;
		getEEpos(EE);
		force.sub(EE); // rozdil
		force.homog();
		//if (force.length() < 1) break; // uz dost presne

		for (r = 0; r < DOF; r++)
		{
			dq[r] = J[0][r]*force[0] + J[1][r]*force[1] + J[2][r]*force[2];
			dq[r] /= iter*100;
		}		

		m_model->m_pJoints[shoulder].rot[0] += dq[0];
		m_model->m_pJoints[shoulder].rot[1] += dq[1];
		m_model->m_pJoints[shoulder].rot[2] += dq[2];
		m_model->m_pJoints[elbow].rot[0] += dq[3];
		m_model->m_pJoints[elbow].rot[1] += dq[4];
		m_model->m_pJoints[hand].rot[0] += dq[5];
		m_model->m_pJoints[hand].rot[1] += dq[6];
		m_model->m_pJoints[hand].rot[2] += dq[7];
	}

}

void JTIK::posorGoal( Vector &posgoal, Matrix &or )
{
	int shoulder = m_model->m_ikj[SHOULDER], elbow = m_model->m_ikj[ELBOW],
		hand = m_model->m_ikj[HAND];	
	int r;
	float dq[DOF]; // zmeny otoceni kloubu (rotace kolem os)

	TRACE("posgoal = [%f, %f, %f], or1goal = [%f, %f, %f]\n",posgoal[0],posgoal[1],posgoal[2],or1goal[0],or1goal[1],or1goal[2]);

	int iter = 200;

	for (int i = 0; i < iter; i++ )
	{
		cmpJacobian();

		Vector force(posgoal), twist(or1goal), EE, or1, or2;
		getEEpos(EE);getEEor1(or1);
		force.sub(EE); // rozdil
		force.homog();
		twist.sub(or1);
		twist.homog();
		//if (force.length() < 1) break; // uz dost presne

		for (r = 0; r < DOF; r++)
		{
			dq[r] = J[0][r]*force[0] + J[1][r]*force[1] + J[2][r]*force[2];
			dq[r] += (J[3][r]*twist[0] + J[4][r]*twist[1] + J[5][r]*twist[2]) * 500;
			dq[r] /= iter*1000;
		}		

		m_model->m_pJoints[shoulder].rot[0] += dq[0];
		m_model->m_pJoints[shoulder].rot[1] += dq[1];
		m_model->m_pJoints[shoulder].rot[2] += dq[2];
		m_model->m_pJoints[elbow].rot[0] += dq[3];
		m_model->m_pJoints[elbow].rot[1] += dq[4];
		m_model->m_pJoints[hand].rot[0] += dq[5];
		m_model->m_pJoints[hand].rot[1] += dq[6];
		m_model->m_pJoints[hand].rot[2] += dq[7];		
	}

	// normovani hodnot kloubu na <0, 2*pi)
	for (int k = 0; k < 3; k++)
	{
		int j = m_model->m_ikj[k];
		for (int o = 0; o < 3; o++)
		{
			if (o==2 && j==elbow) continue; // tou se nehybe
			while (m_model->m_pJoints[j].rot[o] < -M_PI)
				m_model->m_pJoints[j].rot[o] += 2*M_PI;
			while (m_model->m_pJoints[j].rot[o] > M_PI)
				m_model->m_pJoints[j].rot[o] -= 2*M_PI;
		}
	}

	float jointBounds[DOF+1][2] = {{-60, 90}, {-130, 90}, {-130, 30}, {-90, 90}, {-100, 40},
	{0, 0}, {-180, 180}, {-30, 30}, {-60, 60}}; // vcetne nehybne osy u lokte

	for (r = 0; r < DOF; r++)  {
		jointBounds[r][0] *= M_PI;jointBounds[r][0] /= 180;
		jointBounds[r][1] *= M_PI;jointBounds[r][1] /= 180;
	}

	// clamping na povolene hodnoty kloubu
	for (k = 0; k < 3; k++)
	{
		int j = m_model->m_ikj[k];
		for (int o = 0; o < 3; o++)
		{
			if (o==2 && j==elbow) continue; // tou se nehybe
			if (m_model->m_pJoints[j].rot[o] < jointBounds[k*3 + o][0])
				m_model->m_pJoints[j].rot[o] = jointBounds[k*3 + o][0];
			else if (m_model->m_pJoints[j].rot[o] > jointBounds[k*3 + o][1])
				m_model->m_pJoints[j].rot[o] = jointBounds[k*3 + o][1];
		}
	}
}

*/
