/* 
    IK.h

	Inverse Kinematics interface for animating model's arm.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef IK_H
#define IK_H

#include "Model.h"
#include "Vector.h"

//#ifndef M_PI
//#define M_PI 3.141592653F
//#endif

// interface for IK
class IK {
protected:
	Model *m_model; // model k animaci

public:
// constructor accepting model for animation:
	IK(Model *m);

// destructor:
	virtual ~IK();

// get EE position
	void getEEpos(Vector &e);

// get EE first orientation vector
	void getEEor1(Vector &e);

// get EE second orientation vector
	void getEEor2(Vector &e);

// run IK; position goal
	virtual void posGoal( Vector &goal ) = 0;

// run IK; position and orientation goal
	virtual void posorGoal( Vector &posgoal, Matrix &or ) = 0;
};


// extends IK with Jacobian (for Jacobian-based methods)
const int DOF = 8;
class JIK : public IK  {		
protected:
	float J[9][DOF]; // pozice a 2 vektory orientace; 8 DOF

public:
// constructor accepting model for animation:
	JIK(Model *m) : IK(m) {}

// destructor:
	virtual ~JIK() {}

// compute Jacobian for current model position (m_finals in joints)
	void cmpJacobian();

// apply Jacobian to vector f (DOF float array), result add to out (3 float array)
	void Jmul(float *f, float *out);
};

// Jacobian transpose method
/* obsolette: replaced with IKAN
class JTIK : public JIK  {
public:
// constructor accepting model for animation:
	JTIK(Model *m) : JIK(m) {}

// destructor:
	virtual ~JTIK() {}

// run IK - implementation by JT; position goal
	virtual void posGoal( Vector &goal );

// run IK; position and orientation goal
	virtual void posorGoal( Vector &posgoal, Matrix &or );
};
*/

#endif