/* 
    InputDevice.cpp

	Interace for input devices

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "InputDevice.h"
#include "myvec.h"
//#include "MilkshapeModel.cpp"
#include "KeyframeInterpolator.h"
#include <math.h>
#include "ikan.h"
#include <assert.h>

#pragma warning( disable : 4244 4305 )  

void CMouseInputDevice::getMousePos(float &x, float &y)
{
	POINT mpos;
	GetCursorPos(&mpos);
	CPoint cp(mpos.x, mpos.y);

	CPoint center;
	float width2 = (float)m_rect.Width() / 2;
	float height2 = (float)m_rect.Height() / 2;
	center = m_rect.CenterPoint();
	CPoint pos = cp - center;
	x = (float)pos.x / width2;
	y = -(float)pos.y / height2;
	if (x < -1) x = -1;
	else if (x > 1) x = 1;
	if (y < -1) y = -1;
	else if (y > 1) y = 1;
}

void C2DInputDevice::feedNewPos(float x, float y)
{
	float xpos = x - m_mx;
	float ypos = y - m_my;
	float d = sqrt(xpos*xpos + ypos*ypos); // urazena vzdalenost
	xpos /= d; ypos /= d; // normalizace
	float t = m_timer.getTime() / 1000; m_timer.reset();
	float maxD = m_maxSpeed*t;
	//TRACE("t = %f, maxD = %f\n", t, maxD);
	if (d <= maxD)
	{ // stiha se to s mensi nez maxSpeed rychlosti:
		m_mx = x;
		m_my = y;
	}
	else
	{ // kvuli omezeni rychlosti urazime mene:
		//TRACE("omezeni rychlosti: stara poloha [%f,%f], zadana [%f, %f], dosazitelna [%f,%f]\n",
		//	m_mx, m_my, x, y, m_mx + xpos*maxD, m_my + ypos*maxD);
		m_mx = m_mx + xpos*maxD;
		m_my = m_my + ypos*maxD;
	}	
}

KeyframeMID::KeyframeMID(Model *m, int gridX, int gridY)
{
	m_model = m;
 // inicializace mrizky
	m_gridX = gridX;
	m_gridY = gridY;
	m_keyframes = new Matrix[gridX * gridY];
	m_elbowPos = new Point3[gridX * gridY];
}

KeyframeMID::~KeyframeMID()
{
	delete[] m_keyframes;
	delete[] m_elbowPos;
}

bool KeyframeMID::loadKeyframe(const char *fname, int x, int y, double sx, double sy, double sz)
{
	KeyframeInterpolator *anim;
	
	assert( x>=0 && x<m_gridX && y>=0 && y<m_gridY );

	m_elbowPos[x + y*m_gridX].x = sx;
	m_elbowPos[x + y*m_gridX].y = sy;
	m_elbowPos[x + y*m_gridX].z = sz;

	anim = new MilkshapeKeyframeInterpolator(m_model);
	if ( anim->loadKeyframeData( fname ) == false )
	{
		TRACE("Can't load keyframe data %s\n", fname);
		return false;
	}
	anim->setTransform(0);
	float rp[3] = {0,0,0};
	m_model->setFinalMatrices(rp);
	m_keyframes[x + y*m_gridX].set( 
		m_model->m_pJoints[m_model->m_swordholder].m_final.getMatrix() );
	// jeste podle konvence zrusim transformaci root.final:
	Matrix invRF;
	m_model->m_pJoints[0].m_final.inverseHomogenousMatrix(invRF);
	m_keyframes[x + y*m_gridX].preMultiply(invRF);

	delete anim;

	return true;
}

bool KeyframeMID::loadElbowPos(const char *fname)
{
	FILE *f;
	f = fopen(fname, "rt");
	if (!f) return false;
	for (int n=0; n<m_gridX*m_gridY; n++)
	{
		if (feof(f)) return false;
		float x, y, z;
		fscanf(f, "%f %f %f\n", &x, &y, &z);
		m_elbowPos[n].x = x; m_elbowPos[n].y = y; m_elbowPos[n].z = z;
	}
	fclose(f);
	return true;
}

bool KeyframeMID::saveAllKeyframes(const char *fname)
{
	FILE *f;
	f = fopen(fname, "wb");
	if (!f) return false;
	fwrite(&m_gridX, sizeof(int), 1, f);
	fwrite(&m_gridY, sizeof(int), 1, f);
	for (int n=0; n<m_gridX*m_gridY; n++)
	{		
		fwrite(m_keyframes[n].getMatrix(), 16*sizeof(float), 1, f); 
	}
	fclose(f);
	return true;
}

bool KeyframeMID::loadAllKeyframes(const char *fname)
{
	FILE *f;
	float mat[16];
	int gridX, gridY;

	f = fopen(fname, "rb");
	if (!f) return false;
	fread(&gridX, sizeof(int), 1, f);
	if (feof(f)) return false;
	fread(&gridY, sizeof(int), 1, f);
	if (feof(f)) return false;

	//assert(gridX == m_gridX && gridY == m_gridY);
	if (gridX != m_gridX || gridY != m_gridY) return false;

	for (int n=0; n<m_gridX*m_gridY; n++)
	{		
		fread(mat, 16*sizeof(float), 1, f); 
		if (feof(f)) return false;
		m_keyframes[n].set(mat);
	}
	fclose(f);
	return true;
}

void KeyframeMID::getGoal(Matrix &or, Vector &opos)
{
	Matrix res;
	ikMatrix ikRes;
	float dummy[3];
	getikMatrixEl(ikRes, dummy);

	ikMatrix2Matrix(ikRes, res);
	float p[3];
	res.getColumn(3, p);
	opos.set(p);	
	or.set( res.getMatrix() );
	float nil[3] = {0,0,0};
	or.setColumn(3, nil);
}

void flinterpPoint3(float res[3], Point3 A, Point3 B, double t)
{
	res[0] = A.x*(1-t) + B.x*t;
	res[1] = A.y*(1-t) + B.y*t;
	res[2] = A.z*(1-t) + B.z*t;
}
void linterpPoint3(Point3 &res, Point3 A, Point3 B, double t)
{
	res.x = A.x*(1-t) + B.x*t;
	res.y = A.y*(1-t) + B.y*t;
	res.z = A.z*(1-t) + B.z*t;
}
void linterpfloat3(float elpos[3], float a[3], float b[3], float t)
{
	elpos[0] = a[0]*(1-t) + b[0]*t;
	elpos[1] = a[1]*(1-t) + b[1]*t;
	elpos[2] = a[2]*(1-t) + b[2]*t;
}

/*
void KeyframeMID::getElbow(float pos[3])
{
	double dcx = 2 / (double)(m_gridX - 1); // delka ctverecku
	double dcy = 2 / (double)(m_gridY - 1);

	double sxd = floor(	(m_mx+1) / dcx );
	double syd = floor( (m_my+1) / dcy );
	double x = ((m_mx+1) - sxd*dcx) / dcx, y = ((m_my+1) - syd*dcy) / dcy;
	int sx = sxd, sy = syd;
	if (sx == m_gridX-1) { sx--; x=1; }
	if (sy == m_gridY-1) { sy--; y=1; }
	assert(sx >= 0 && sy >= 0 && (sx+1) < m_gridX && (sy+1) < m_gridY);
	Point3 r0,r1,r2,r3;
	r0 = m_elbowPos[sx + m_gridX*sy];
	r1 = m_elbowPos[sx + m_gridX*(sy+1)];
	r2 = m_elbowPos[sx+1 + m_gridX*(sy+1)];
	r3 = m_elbowPos[sx+1 + m_gridX*sy];

	Point3 A, B, res;
	linterpPoint3(A, r0, r1, y);
	linterpPoint3(B, r3, r2, y);
	linterpPoint3(res, A, B, x);
	pos[0] = res.x; pos[1] = res.y; pos[2] = res.z;
}
*/

void KeyframeMID::getikMatrixEl(ikMatrix m, float pos[3])
{
	double dcx = 2 / (double)(m_gridX - 1); // delka ctverecku
	double dcy = 2 / (double)(m_gridY - 1);

/* 3) bilinearni kvaternionova interpolace na pravidelne ctvercove mrizce (velikosti gridSize): */
	// pripravim si vstupni data:
	double sxd = floor(	(m_mx+1) / dcx );
	double syd = floor( (m_my+1) / dcy );
	double x = ((m_mx+1) - sxd*dcx) / dcx, y = ((m_my+1) - syd*dcy) / dcy;
	int sx = sxd, sy = syd;
	if (sx == m_gridX-1) { sx--; x=1; }
	if (sy == m_gridY-1) { sy--; y=1; }
	assert(sx >= 0 && sy >= 0 && (sx+1) < m_gridX && (sy+1) < m_gridY);
	//TRACE("m_mx = %f, m_my = %f, sx = %i, sy = %i, x = %f, y = %f\n",m_mx,m_my,sx,sy,x,y);
	ikMatrix rohy[4];
	Matrix2ikMatrix(m_keyframes[sx + m_gridX*sy], rohy[0]);
	Matrix2ikMatrix(m_keyframes[sx + m_gridX*(sy+1)], rohy[1]);
	Matrix2ikMatrix(m_keyframes[sx+1 + m_gridX*(sy+1)], rohy[2]);	
	Matrix2ikMatrix(m_keyframes[sx+1 + m_gridX*sy], rohy[3]);	

	ikMatrix A, B;
	linterpikMatrix(A, rohy[0], rohy[1], y);
	linterpikMatrix(B, rohy[3], rohy[2], y);
	linterpikMatrix(m, A, B, x);

// jeste elbow position:
	Point3 r0,r1,r2,r3;
	r0 = m_elbowPos[sx + m_gridX*sy];
	r1 = m_elbowPos[sx + m_gridX*(sy+1)];
	r2 = m_elbowPos[sx+1 + m_gridX*(sy+1)];
	r3 = m_elbowPos[sx+1 + m_gridX*sy];

	Point3 pA, pB;
	linterpPoint3(pA, r0, r1, y);
	linterpPoint3(pB, r3, r2, y);
	flinterpPoint3(pos, pA, pB, x);	
}

ActionMID::ActionMID(Model *m, int nsx, int nsy, int ssx, int ssy, int ksx, int ksy) : 
	m_naprahy(m,nsx,nsy), m_seky(m,ssx,ssy), m_kryty(m,ksx,ksy), m_timer()
{
	m_model = m;
	m_stav = cNaprah;
	m_dobaNaprahKryt = 250;
	m_dobaNaprahSek = 400;
	m_dobaSNK = 250;
	m_dobaSNKN = 250;
	m_maxDobaSeku = 500;
}

// zjisti v jake zone se vstupni zarizeni (m_mx, m_my) prave ted nachazi (3.3.03):
int ActionMID::getZone()
{
	// zatim jen cZonaNic a cZonaNaprah, jeste bude cZonaBod
	TRACE("ActionMID::getZone  m_mx = %f, m_my = %f\n", m_mx, m_my);
	if (m_mx < - 0.8 || m_mx > 0.8 || m_my < -0.8 || m_my > 0.8) return cZonaNaprah;
	return cZonaMimo;
}

void ActionMID::testTime()
{
	double realTime = m_timer.getTime()*m_timerMul + m_timerAdd;
	switch (m_stav)
	{	
		case cNaprahKryt:
				if (realTime >= m_dobaNaprahKryt) m_stav = cKryt;				
			break;
		case cKrytNaprah:
				if (realTime < 0) m_stav = cNaprah;
			break;
		case cNaprahSek:
				if (realTime >= m_dobaNaprahSek) 
				{
					m_stav = cSek;
					// od 3.3.03 se nesmi v seku zustavat nekonecne dlouhu -> natahnu timer:
					m_timer.reset();
					m_timerMul = 1; m_timerAdd = 0;
				}
			break;
		case cSekNaprah:
				if (realTime < 0) m_stav = cNaprah;
			break;
		case cSek: // automaticke zruseni seku (3.3.03):
				if (realTime >= m_maxDobaSeku) 
				{
					m_timerMul = -1; m_timerAdd = m_dobaNaprahSek;
					m_timer.reset();
					m_stav = cSekNaprah;
				}
			break;
		case cSNK:
				if (realTime >= m_dobaSNK) m_stav = cKryt;
			break;
		case cSNKN:
				if (realTime >= m_dobaSNKN) m_stav = cNaprah;
			break;
		default:
			break;
	}

	//TRACE("stav = %i\n", m_stav);
}

void ActionMID::feedButton(int a)
{
	CSingleLock lock(&m_mutex); // vzajemne vylouceni!
	lock.Lock();
	TRACE("Buton %i, stav = %i \n", a, m_stav);

	testTime();

	double realTime = m_timer.getTime()*m_timerMul + m_timerAdd;
	switch (m_stav)
	{
		case cNaprah:
			if (a==WM_LBUTTONDOWN && getZone() == cZonaNaprah) // sek jen v naprahove zone (3.3.03)
			{
				m_timerMul = 1; m_timerAdd = 0;
				m_timer.reset();
				m_stav = cNaprahSek;
			}
			else if (a==WM_RBUTTONDOWN)
			{
				m_timerMul = 1; m_timerAdd = 0;
				m_timer.reset();
				m_stav = cNaprahKryt;
			}
			break;

		case cKryt:	
			if (a==WM_RBUTTONUP)
			{
				m_timerMul = -1; m_timerAdd = m_dobaNaprahKryt;
				m_timer.reset();
				m_stav = cKrytNaprah;
			}
			break;

		case cSek:
			if (a==WM_LBUTTONUP)
			{
				m_timerMul = -1; m_timerAdd = m_dobaNaprahSek;
				m_timer.reset();
				m_stav = cSekNaprah;
			}
			break;

		case cNaprahKryt:
			if (a==WM_RBUTTONUP)
			{
				m_timerMul = -1; m_timerAdd = realTime; // nevracim cele, jen odkud prave jsem
				m_timer.reset();
				m_stav = cKrytNaprah;
			}
			break;
		case cKrytNaprah:
			if (a==WM_RBUTTONDOWN)
			{
				m_timerMul = 1; m_timerAdd = realTime;
				m_timer.reset();
				m_stav = cNaprahKryt;
			}
			break;
		case cNaprahSek:
			if (a==WM_LBUTTONUP)
			{
				m_timerMul = -1; m_timerAdd = realTime; // nevracim cele, jen odkud prave jsem
				m_timer.reset();
				m_stav = cSekNaprah;
			}
			break;
		case cSekNaprah:
			if (a==WM_LBUTTONDOWN && getZone() == cZonaNaprah)
			{
				// 3.3.03: chyba - misto m_timerAdd = m_dobaNaprahSek - m_timer.getTime(); ma byt:
				m_timerMul = 1; m_timerAdd = realTime;
				m_timer.reset();
				m_stav = cNaprahSek;
			}
			if (a==WM_RBUTTONDOWN)
			{
				ikMatrix pom; float p3[3];
				getikMatrixEl(pom, p3); // zapamatuju si matici a pozici lokte
				cpikMatrix(m_from, pom); cpvector(m_pfrom, p3);
				// pokud se tam nahodou provedla zmena ve stavu, tak to nevadi, protoze
				// se vzapeti nastavi novy:
				m_timerMul = 1; m_timerAdd = 0;
				m_timer.reset();				
				m_stav = cSNK;
			}
			break;
		case cSNK: // nouzove kryti
			if (a==WM_RBUTTONUP)
			{
				ikMatrix pom; float p3[3];
				getikMatrixEl(pom, p3);	// neni dobry psat rovnou do m_from a m_pfrom - ty se totiz uvnitr getikMatrixEl pouzivaji...
				cpikMatrix(m_from, pom); cpvector(m_pfrom, p3);
				m_timerMul = 1; m_timerAdd = 0;
				m_timer.reset();				
				m_stav = cSNKN;
			}
			break;
		
		default: assert(m_stav == cSNKN);
	}
	lock.Unlock();
}

void ActionMID::feedNewPos(float x, float y)
{
	CSingleLock lock(&m_mutex);
	lock.Lock();

	C2DInputDevice::feedNewPos(x, y);
	m_naprahy.feedNewPos(x, y);
	m_seky.feedNewPos(x, y);
	m_kryty.feedNewPos(x, y);

	lock.Unlock();
}

void ActionMID::setPosOr(ikMatrix ikRes, Matrix &or, Vector &opos)
{
	Matrix res;
	ikMatrix2Matrix(ikRes, res);
	float p[3];
	res.getColumn(3, p);
	opos.set(p);	
	or.set( res.getMatrix() );
	float nil[3] = {0,0,0};
	or.setColumn(3, nil);
}

void ActionMID::getElbow(float elpos[3])
{
	ikMatrix m;
	getikMatrixEl(m, elpos);	
}

void ActionMID::getGoal(Matrix &or, Vector &opos)
{
	ikMatrix m; float dummy[3];
	getikMatrixEl(m, dummy);
	setPosOr(m, or, opos);
}

// vraci pevnost drzeni zbrani od 0 do 1:
float ActionMID::getToughness()
{
	CSingleLock lock(&m_mutex); lock.Lock();

	float ret;
	// jednoduchy model: pri krytu nejpevnejsi drzeni 1, jinak 0; 
	if (m_stav == cKryt) ret = 1.0;
	if (m_stav == cNaprahKryt)
	{ //pri pohybu do krytu linterp:
		float realTime = m_timer.getTime()*m_timerMul + m_timerAdd;
		float t = realTime / m_dobaNaprahKryt;
		if (t<0) t=0; else if (t>1) t=1;
		ret = t;
	}
	ret = 0.0;

	lock.Unlock();
	return ret;
}

void ActionMID::getikMatrixEl(ikMatrix m, float elpos[3])
{
	CSingleLock lock(&m_mutex); lock.Lock();

	testTime();

	double realTime = m_timer.getTime()*m_timerMul + m_timerAdd, t;
	//TRACE("%f\n", realTime);
	switch (m_stav)
	{
		case cNaprah:
			m_naprahy.getikMatrixEl(m, elpos);
			break;
		case cKryt:			
			m_kryty.getikMatrixEl(m, elpos);
			break;
		case cSek:			
			m_seky.getikMatrixEl(m, elpos);
			break;
		case cNaprahKryt:
		case cKrytNaprah:
			{
				ikMatrix nap, kryt;
				float pnap[3], pkryt[3];
				t = realTime / m_dobaNaprahKryt;
				if (t<0) t=0;
				else if (t>1) t=1;
				//if (m_stav == cKrytNaprah) t = 1 - t;
				m_naprahy.getikMatrixEl(nap, pnap);
				m_kryty.getikMatrixEl(kryt, pkryt);
				linterpikMatrix(m, nap, kryt, t);
				linterpfloat3(elpos, pnap, pkryt, t);				
			}
			break;
		case cNaprahSek:
		case cSekNaprah:
			{
				ikMatrix nap, sek;
				float pnap[3], psek[3];
				t = realTime / m_dobaNaprahSek;
				if (t<0) t=0;
				else if (t>1) t=1;				
				m_naprahy.getikMatrixEl(nap, pnap);
				m_seky.getikMatrixEl(sek, psek);
				linterpikMatrix(m, nap, sek, t);
				linterpfloat3(elpos, pnap, psek, t);				
			}
			break;
		case cSNK:
			{
				ikMatrix cil;
				float pcil[3];
				t = realTime / m_dobaSNK;
				if (t<0) t=0;
				else if (t>1) t=1;
				m_kryty.getikMatrixEl(cil, pcil);
				linterpikMatrix(m, m_from, cil, t);
				linterpfloat3(elpos, m_pfrom, pcil, t);
			}	
			break;
		case cSNKN:
			{
				ikMatrix cil;
				float pcil[3];
				t = realTime / m_dobaSNKN;
				if (t<0) t=0;
				else if (t>1) t=1;
				m_naprahy.getikMatrixEl(cil, pcil);
				linterpikMatrix(m, m_from, cil, t);
				linterpfloat3(elpos, m_pfrom, pcil, t);
			}	
			break;
		default: assert(false);
	}
	lock.Unlock();
}




/*
void Sphere1MID::toCircle(float &icx, float &icy, bool &inner)
{
	float rr = m_mx*m_mx + m_my*m_my;
	float x = m_mx, y = m_my;
	if (rr > m_v*m_v)
	{ // je mimo velky kruh
		float t = m_v/sqrt(rr);
		x = t*m_mx; y = t*m_my; // promitnu na velkou kruznici
		rr = x*x + y*y;
	}
	if (rr > 1)
	{ // je mimo maly (jednotkovy) kruh
		inner = false;
		float nl = 2-sqrt(rr);
		float t = 1/sqrt(rr);
		x = t*x; y = t*y; // promitnu na jednotkovou kruznici
		icx = x*nl; icy = y*nl; // zmensim
	} else {
		inner = true;
		icx = x; icy = y;
	}
//	TRACE("icx = %f, icy = %f, inner = %i\n",icx,icy,inner);
}

// vraci matici orientace a pozice koncoveho efektoru
void Sphere1MID::getGoal(Matrix &or, Vector &opos)
{
	float icz, icy;
	bool inner;
	toCircle(icz, icy, inner);
	float coef = -1;
	if (inner) coef = 1;
	float pos[3] = {coef*sqrt(1 - icz*icz - icy*icy), icy, icz};
	float n[3] = {pos[0], pos[1], pos[2]};
	float c = n[0] / DOT(n, n); // v = (1, 0, 0)
	float z[3] = {1 - c*n[0], -c*n[1], -c*n[2]};
	float d[3];
	crossproduct(d, z, n);
	Vector::norm3(n); Vector::norm3(z); Vector::norm3(d); Vector::norm3(pos);
	pos[0]*=m_armLength; pos[1]*=m_armLength; pos[2]*=m_armLength;
	opos.set(pos);
	TRACE("pos = [%f, %f, %f]\n",pos[0],pos[1],pos[2]);

	n[0]*=-1; n[1]*=-1; n[2]*=-1;
	or.loadIdentity();
	or.setColumn(2, n); // normala urcuje smer cepele (osa -Z)
	or.setColumn(0, z); // zastita ve smeru ostri (osa X)
	or.setColumn(1, d);
}
*/