/* 
    InputManager.cpp

	The purpose of this class is to translate the physical input from the actual input
	devices to the logical input of Keyboard and Action Input Devices (see InputDevice.h).

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "InputManager.h"
#include <assert.h>

KeyboardProfile::KeyboardProfile(int hrac)
{
	setProfile(hrac);
}

// nastavi defaultni profil:
void KeyboardProfile::setProfile(int hrac)
{
	if (hrac == 0)
	{
		m_vk[cStreh] = 'N';
		m_vk[cOdsun] = 'H';
		m_vk[cPosun] = 'J';
		m_vk[cVypad] = 'K';
	}
	else if (hrac == 1)
	{
		m_vk[cStreh] = 'Z';
		m_vk[cOdsun] = 'A';
		m_vk[cPosun] = 'S';
		m_vk[cVypad] = 'D';
	}
	else assert(false);
}

int KeyboardProfile::inProfile(int key)
{
	for (int n=0; n<cPocetAkci; n++)
		if (m_vk[n] == key) return n;
	return -1;
}

InputManager::InputManager(CDirectInput *di, CMouseInputDevice *mi, LocalModel *locModel0, 
						   LocalModel *locModel1)
{
	// nastavim vstupy:
	m_di = di;
	m_mi = mi;
	// vystupy:
	m_locModel[0] = locModel0;
	m_locModel[1] = locModel1;
	m_cs[0] = m_cs[1] = 0; // na zacatku zadna sit
	// prirazeni mezi vstupy a vystupy:
	m_playerDevice[0] = -1;
	m_playerDevice[1] = -1;
	// (zatim natvrdo)
	//m_joyDriven = 1;
	//m_mouseDriven = 0;
	m_kbdProfil[0].setProfile(0);
	m_kbdProfil[1].setProfile(1);

	m_minB1[0] = m_minB1[1] = false; 
	m_minB2[0] = m_minB2[1] = false;
}

InputManager::~InputManager()
{
}

// osaha zarizeni (nektera nevysilaji udalosti) a rozesle zpravy dal:
void InputManager::pollDevices()
{
	float X, Y;

	for (int player = 0; player<2; player++)
	{
		if (m_playerDevice[player] == -2 || m_playerDevice[player] == -3); // nothing, nebo remote
		else if (m_playerDevice[player] == -1)
		{ // smys
			m_mi->getMousePos(X, Y);
			//if (m_locModel[player]) m_locModel[player]->m_mid->feedNewPos(X, Y);
			feedPos(player, X, Y);
		}
		else
		{ // nektery z joysticku, prip. dalsich zarizeni:
			assert(m_playerDevice[player] < m_di->getNumJoysticks());
			bool B1, B2;
			if (!m_di->getCurrentState(&X, &Y, &B1, &B2, m_playerDevice[player])) 
				TRACE("nak nejde cist joystick cislo %i\n", m_playerDevice[player]);
			else
			{ // jde cist joystick:				
				/*
				m_locModel[player]->m_mid->feedNewPos(X, Y);
				// na rozdil od mysi se tady musi vyslat i zpravy o zmene stavu buttonu:
				if (!m_minB1[player] && B1) m_locModel[player]->m_mid->feedButton(WM_LBUTTONDOWN);
				if (m_minB1[player] && !B1) m_locModel[player]->m_mid->feedButton(WM_LBUTTONUP);
				if (!m_minB2[player] && B2) m_locModel[player]->m_mid->feedButton(WM_RBUTTONDOWN);
				if (m_minB2[player] && !B2) m_locModel[player]->m_mid->feedButton(WM_RBUTTONUP);				
				*/
				feedPos(player, X, Y);
				if (!m_minB1[player] && B1) feedButton(player, WM_LBUTTONDOWN);
				if (m_minB1[player] && !B1) feedButton(player, WM_LBUTTONUP);
				if (!m_minB2[player] && B2) feedButton(player, WM_RBUTTONDOWN);
				if (m_minB2[player] && !B2) feedButton(player, WM_RBUTTONUP);				
			}
			// ulozim minulost:
			m_minB1[player] = B1; m_minB2[player] = B2;
		}		
	}
}

void InputManager::setConstantForce(int h, long x, long y, int magnitude)
{
	if (m_playerDevice[h] < 0) return;
	assert(0<=m_playerDevice[h] && m_playerDevice[h]<m_di->getNumJoysticks());
	m_di->setConstantForce(m_playerDevice[h], x, y, magnitude);
}

void InputManager::setFrictionCoefficient(int h, int coef)
{
	if (m_playerDevice[h] < 0) return;
	assert(0<=m_playerDevice[h] && m_playerDevice[h]<m_di->getNumJoysticks());
	m_di->setFrictionCoefficient(m_playerDevice[h], coef);
}

void InputManager::setRampForce(int h, int x, int y, int startF, int endF, int duration)
{
	if (m_playerDevice[h] < 0) return;
	assert(0<=m_playerDevice[h] && m_playerDevice[h]<m_di->getNumJoysticks());
	m_di->setRampForce(m_playerDevice[h], x, y, startF, endF, duration);
}

// zprava o stisknuti/pusteni tlacitka mysi 'button':
void InputManager::mouseEvent(int button)
{
	for (int pl=0; pl<2; pl++)
	{
		if (m_playerDevice[pl] != -1) continue;
		// hrac 'pl' je ovladan mysi, poslu mu zpravu:
		//if (m_locModel[pl]) m_locModel[pl]->m_mid->feedButton(button);
		feedButton(pl, button);
	}
}

// prevodni tabulka z indexu v kbdProfajlu na kody, ktere prijima KID (z historickych duvodu nejsou stejne)
int naKid[cPocetAkci] = {VK_SPACE, 'H', 'J', 'K'};

// zprava o stisknuti/pusteni 'down' tlacitka 'nChar':
void InputManager::keyboardEvent(bool down, int nChar)
{
	int kcode;
	if ((kcode = m_kbdProfil[0].inProfile(nChar)) >= 0) 
		feedKey(0, naKid[kcode], down);
		//m_locModel[0]->m_kid->KeyboardEvent(down, naKid[kcode]);
	else if ((kcode = m_kbdProfil[1].inProfile(nChar)) >= 0) 
		//m_locModel[1]->m_kid->KeyboardEvent(down, naKid[kcode]);
		feedKey(1, naKid[kcode], down);
	else TRACE("Udalost na nehandlovane klavese %i\n", nChar);
}

void InputManager::setInputDevice(int player, int code)
{
	assert(-3 <= code && code < m_di->getNumJoysticks());
	m_playerDevice[player] = code;
	if (code >= 0)
	{
		m_di->initJoystick(code);
		m_minB1[player] = m_minB2[player] = false;
	}
}

void InputManager::feedButton(int pl, int button)
{
	if (m_locModel[pl]) m_locModel[pl]->m_mid->feedButton(button);
	if (m_cs[pl])

	{ // odeslu po siti:
		InputEvent ie; ie.player = pl;
		ie.type = cButton; ie.data = button;
		m_cs[pl]->sendInputEvent(ie);
	}
}

void InputManager::feedKey(int pl, int key, bool down)
{
	if (m_locModel[pl]) m_locModel[pl]->m_kid->KeyboardEvent(down, key);
	if (m_cs[pl])
	{ // odeslu po siti:
		InputEvent ie; ie.player = pl;
		ie.type = cKey; ie.data = key; ie.flag = down;
		m_cs[pl]->sendInputEvent(ie);
	}
}

void InputManager::feedPos(int pl, float x, float y)
{
	if (m_locModel[pl]) m_locModel[pl]->m_mid->feedNewPos(x, y);
	if (m_cs[pl])
	{ // odeslu po siti:
		InputEvent ie; ie.player = pl;
		ie.type = cPosition; ie.x = x; ie.y = y;
		m_cs[pl]->sendInputEvent(ie);
	}
}