/* 
    InputManager.h

	The purpose of this class is to translate the physical input from the actual input
	devices to the logical input of Keyboard and Action Input Devices (see InputDevice.h).

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef INPUTMANAGER_H
#define INPUTMANAGER_H

#include "InputDevice.h" // mys
#include "DirectInput.h" // joystick
#include "LocalRemoteModel.h" 
#include "Communication.h" // vystup pujde vzdalenemu locModelu pres sit

const cMouse = 1, cJoystick = 2;

const cPocetAkci = 4;
const cStreh = 0, cOdsun = 1, cPosun = 2, cVypad = 3;

// funkce klaves:
class KeyboardProfile
{
public:
	// virtual key-code (VK_xxx) pro jednotlive akce:
	int m_vk[cPocetAkci];

	KeyboardProfile(int hrac = 0);
	~KeyboardProfile() {}
	// vraci cislo klavesy 'key' v profilu, nebo -1 pokud tam neni:
	int inProfile(int key);
	// natahne defaultni profil:
	void setProfile(int hrac);
};

class InputManager
{
public:
	InputManager(CDirectInput *di, CMouseInputDevice *mi, LocalModel *locModel0, LocalModel *locModel1);
	virtual ~InputManager();

// 1) vstupy ze skutecnych vstupnich zarizeni:
// ... pozdeji sem muzu pridat ovladani po siti
protected:
	CDirectInput *m_di; // joystick
	bool m_minB1[2], m_minB2[2]; // minuly stav tlacitek joysticku pro oba hrace
	CMouseInputDevice *m_mi; // mys
public: // metody, ktere vlastnik teto tridy zavola aby ji informoval o udalostech:
	// rekne, ze hraci 'player' prislusi vstup se zarizeni 'code' (-1 je mys, 0,1,... joysticky):
	void setInputDevice(int player, int code);
	// zprava o stisknuti/pusteni 'down' tlacitka 'nChar':
	void keyboardEvent(bool down, int nChar);
	// zprava o stisknuti/pusteni tlacitka mysi 'button':
	void mouseEvent(int button);	

// dalsi metody:
	// osaha zarizeni (nektera nevysilaji udalosti) a rozesle zpravy modelum:
	void pollDevices();
	// vrati nastaveni klaves:
	int getKeyboardConfig(int h, int n) { return m_kbdProfil[h].m_vk[n]; }
	// nastavi klavesy:
	void setKeyboardConfig(int h, int n, int key) { m_kbdProfil[h].m_vk[n] = key; }
	// posle joysticku hrace 'h' pozadavak na nastaveni sily:
	void setConstantForce(int h, long x, long y, int magnitude);
	void setFrictionCoefficient(int h, int coef);
	void setRampForce(int h, int x, int y, int startF, int endF, int duration);

// 2) kam ma InputManager posilat vystupy: (jsou-li ukazatele = 0, tak je nepouziva)
public:
	void setLocModels(LocalModel *locModel0, LocalModel *locModel1)
		{ m_locModel[0] = locModel0; m_locModel[1] = locModel1; }
	void setCommSockets(CCommSocket *cs0, CCommSocket *cs1)
		{ m_cs[0] = cs0; m_cs[1] = cs1; }
protected:
	LocalModel	*m_locModel[2];
	CCommSocket *m_cs[2];

	// vnitrni funkce, ktere distribuuji zpravy neNULLovym m_locModelum a m_cskam:
	void feedButton(int pl, int button);
	void feedKey(int pl, int key, bool down);
	void feedPos(int pl, float x, float y);

// 3) vazby mezi vstupy a vystupy manazeru:
protected:
	//int m_joyDriven, m_mouseDriven; // index lok. modelu (0/1) kteremu se ma posilat vstup z joysticku a z mysi
	int m_playerDevice[2]; // kody vybranych vstupnich zarizeni pro hrace 0 a 1
	KeyboardProfile m_kbdProfil[2]; // profil pro model 0 a 1	
};

#endif INPUTMANAGER_H