/* 
    Judge.cpp

	Automatic fencing judge.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "gl1.h"
#include "Judge.h"
#include "ModeSimulation.h"
#include <assert.h>

//enum {
//	cJM_Nonec = 0, cJM_EnGuard, cJM_Go, cJM_Stop, cJM_DoubleHit, cJM_Avatar0Hit,
//	cJM_Avatar1Hit, cJM_InvalidHit, cJM_Last
//};
//
//char *JMtext[cJM_Last] = {0, "En Guard!", "Go!", "Stop! No hit", "Double hit!", 
//"Avatar 0 was hit!", "Avatar 1 was hit!", "Illegal hit! Go to guard first"};

CJudge::CJudge(void *mode)
{
	m_simulation = mode;
	m_state = cJudgeNotInit;
	m_text = (char *)0;
	m_score = (char *)0;
	m_strehy = 0;
	m_body[0] = -1;
}

CJudge::~CJudge()
{
}

const CString &CJudge::getText()
{
	testTime();
	return m_text;
}

const CString &CJudge::getScore()
{
	if (m_body[0] == -1) m_score.Empty();
	else m_score.Format("%i : %i", m_body[0], m_body[1]);
	return m_score;
}

// ma se spustit cinnost rozhodciho:
void CJudge::start(bool reset)
{
	m_state = cJudgeEnGuard;
	m_timer.reset();
	m_text = "En Guard!";
	m_strehy = 0;
	if (reset)
	{
		m_body[0] = m_body[1] = 0;
	}
}

void CJudge::inGuard(int h)
{
	m_strehy |= 1<<h;
}

void CJudge::bodybody()
{
	if (m_state == cJudgeGo)
	{
		m_state = cJudgeBodyBody;
		m_timer.reset();
		m_text = "Stop! No hit";
	}
}

void CJudge::hit(bool h0, bool h1)
{
	if (!(h0||h1)) return;
	int h;
	if (h0&&h1) h = -1;
	else if (h0) h = 0;
	else /* h1 */ h = 1;	

	if (m_state == cJudgeGo)
	{
		m_state = cJudgeHit;
		m_timer.reset();
		if (h == -1)
		{ // oba soucasne:
			m_body[0]++; m_body[1]++;
			//m_text = "Double hit! Score " + CString(m_body[0]) + " : " + CString(m_body[1]);
			m_text.Format("Double hit!");
		}
		else
		{ // jen jeden:
			m_body[1-h]++;			
			m_text.Format("Avatar %i was hit!", h);
		}
	}
	else if (m_state == cJudgeEnGuard)
	{
		m_text = "Illegal hit! Go to guard first!";
		m_timer.reset();
	}
}

void CJudge::testTime()
{
	switch (m_state)
	{
		case cJudgeEnGuard:
				if (m_timer.getTime() > cJudgeMessageTime)
				{
					m_text = (char *)0;
				}
				if (m_strehy == 3)
				{ // oba hraci ve strehu:
					m_state = cJudgeGo;
					m_timer.reset();
					m_text = "Go!";					
				}
			break;
		case cJudgeGo:
				if (m_timer.getTime() > cJudgeMessageTime)
					m_text = (char *)0;
			break;
		case cJudgeBodyBody:
				if (m_timer.getTime() > cJudgeMessageTime)
				{
					m_text = (char *)0;
					((CModeSimulation *)m_simulation)->retreat();					
					start(false); // restartuju se:
				}
			break;
		case cJudgeHit:
				if (m_timer.getTime() > cJudgeMessageTime)
				{
					m_text = (char *)0;
					((CModeSimulation *)m_simulation)->retreat();					
					start(false); // restartuju se:
				}
			break;
	}
	
}