/* 
    KeyboardInputDevice.cpp

	Interace for keyboard actions

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "InputDevice.h"
#include "myvec.h"
//#include "MilkshapeModel.cpp"
#include "KeyframeInterpolator.h"
#include <math.h>
#include <assert.h>
#include <direct.h>
#include "paths.h"

#pragma warning( disable : 4244 4305 )  

KeyboardInputDevice::KeyboardInputDevice(Model *m, CJudge *judge, int avatar)
{
	m_model = m;
	m_judge = judge;
	m_avatar = avatar;
    m_anim[0] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[0]->loadKeyframeData( DATA_PATH + "model.ms3d" ) == false )
	{ 
		::MessageBox( NULL, "Couldn't load the keyframe file 'model.ms3d'", "Error", MB_OK | MB_ICONERROR ); 
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Aktualni adresar", MB_OK);
	}
	m_anim[1] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[1] -> reverseKeyframeData( *m_anim[0] );
	m_anim[2] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[2]->loadKeyframeData( DATA_PATH + "posun.ms3d" ) == false )
	{
		::MessageBox( NULL, "Couldn't load the keyframe file 'posun.ms3d'", "Error", MB_OK | MB_ICONERROR );		
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Aktualni adresar", MB_OK);
	}
	m_anim[3] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[3] -> reverseKeyframeData( *m_anim[2] );
	m_anim[4] = new MilkshapeKeyframeInterpolator(m_model);
	if ( m_anim[4]->loadKeyframeData( DATA_PATH + "vypad.ms3d" ) == false )
	{
		::MessageBox( NULL, "Couldn't load the keyframe file 'posun.ms3d'", "Error", MB_OK | MB_ICONERROR );		
		char tst[128];	_getcwd(tst, 128);::MessageBox(0, tst, "Aktualni adresar", MB_OK);
	}
	m_anim[5] = new MilkshapeKeyframeInterpolator(m_model);
	m_anim[5] -> reverseKeyframeData( *m_anim[4] );

	resetAnimation();
}

void KeyboardInputDevice::resetAnimation()
{
	m_stav = nSraz;
	m_rootPos[0] = 0; m_rootPos[1] = 0; m_rootPos[2] = 0;
}

KeyboardInputDevice::~KeyboardInputDevice()
{
	for (int n=0; n<cNumAnims; n++) delete m_anim[n];
}


void KeyboardInputDevice::testTime()
{
	switch (m_stav)
	{
		case nSrazStreh:
			if (m_timer.getTime() >= m_anim[0]->getTotalTime())
			{
				if (m_judge) m_judge->inGuard(m_avatar);
				m_stav = nStreh;
				m_anim[0] -> addRootTrans(m_rootPos);
				m_anim[2] -> reset();
			}
			break;
		case nStrehSraz:
			if (m_timer.getTime() >= m_anim[1]->getTotalTime())
			{
				m_stav = nSraz;
				m_anim[1] -> addRootTrans(m_rootPos);
				m_anim[0] -> reset();
			}
			break;
		case nPosun:
			if (m_timer.getTime() >= m_anim[2]->getTotalTime())
			{
				m_stav = nStreh;
				m_anim[2] -> addRootTrans(m_rootPos);
				m_anim[2] -> reset();
			}
			break;
		case nOdsun:
			if (m_timer.getTime() >= m_anim[3]->getTotalTime())
			{
				m_stav = nStreh;
				m_anim[3] -> addRootTrans(m_rootPos);
				m_anim[2] -> reset();
			}
			break;
		case nStrehVypad:
			if (m_timer.getTime() >= m_anim[4]->getTotalTime())
			{
				if (!m_hnedZpatky)
				{
					m_stav = nVypad;
					m_anim[4] -> addRootTrans(m_rootPos);
					m_anim[5] -> reset();
				}
				else // doslo k uvolneni klavesy jeste behem vypadu
				{
					m_stav = nVypadStreh;
					m_anim[4] -> addRootTrans(m_rootPos);
					m_timer.reset();
					m_anim[5] -> reset();					
				}
			}
			break;
		case nVypadStreh:
			if (m_timer.getTime() >= m_anim[5]->getTotalTime())
			{
				m_stav = nStreh;
				m_anim[5] -> addRootTrans(m_rootPos);
				m_anim[2] -> reset();
			}
			break;
		default:
			break;
	}
}

void KeyboardInputDevice::KeyboardEvent(bool down, int nChar, int repCnt, int nFlags)
{
	CSingleLock lock(&m_mutex); // vzajemne vylouceni (vola se z CCommSocket.OnReceive)
	lock.Lock();
	nFlags = repCnt = 0; // anti-warning
	//if (repCnt>=2) return; // zabrani automatickemu opakovani

	testTime();
	if (down)
	{
		if (m_stav == nSraz && nChar == VK_SPACE)
		{			
			m_stav = nSrazStreh;
			m_timer.reset();
			m_anim[0]->reset();		
			return;
		}
		if (m_stav == nStreh && nChar == VK_SPACE)
		{
			m_stav = nStrehSraz;
			m_timer.reset();
			m_anim[1]->reset();
			return;
		}
		if (m_stav == nStreh)
		{
			if (nChar == 'J')
			{
				m_stav = nPosun;
				m_timer.reset();
				m_anim[2]->reset();
				return;
			}
			if (nChar == 'H')
			{
				m_stav = nOdsun;
				m_timer.reset();
				m_anim[3]->reset();
				return;
			}
			if (nChar == 'K')
			{
				m_stav = nStrehVypad;
				m_hnedZpatky = false;
				m_timer.reset();
				m_anim[4]->reset();
				return;
			}
		}
	}
	else // klavesa pustena
	{		
		if (m_stav == nVypad && nChar == 'K')
		{
			m_stav = nVypadStreh;
			m_timer.reset();
			m_anim[5]->reset();
			return;
		}
		if (m_stav == nStrehVypad && nChar == 'K')
		{
			m_hnedZpatky = true;
			return;
		}
	}
	lock.Unlock();
}

void KeyboardInputDevice::setLegsAnimation()
{
	CSingleLock lock(&m_mutex); lock.Lock();
	testTime();
	switch (m_stav)
	{
		case nSraz:
			anim_idx = 0; anim_time = 0;			
			break;
		case nStreh:
			anim_idx = 2; anim_time = 0;			
			break;
		case nSrazStreh:
			anim_idx = 0; anim_time = m_timer.getTime();			
			break;
		case nStrehSraz:
			anim_idx = 1; anim_time = m_timer.getTime();
			//m_anim[1] -> setTransform(m_timer.getTime());
			break;
		case nPosun:
			anim_idx = 2; anim_time = m_timer.getTime();
			//m_anim[2] -> setTransform(m_timer.getTime());
			break;
		case nOdsun:
			anim_idx = 3; anim_time = m_timer.getTime();
			//m_anim[3] -> setTransform(m_timer.getTime());
			break;
		case nStrehVypad:
			anim_idx = 4; anim_time = m_timer.getTime();
			//m_anim[4] -> setTransform(m_timer.getTime());
			break;
		case nVypad:
			anim_idx = 5; anim_time = 0;
			//m_anim[5] -> setTransform(0);
			break;
		case nVypadStreh:
			anim_idx = 5; anim_time = m_timer.getTime();
			//m_anim[5] -> setTransform(m_timer.getTime());
			break;
		default: assert(false);
	}
	m_anim[anim_idx] -> setTransform(anim_time);
	lock.Unlock();
}