/* 
    KeyframeInterpolator.h

	Compute interpolated joint transformations. Including
	extended class for loading keyframe data from Milkshape.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef KEYFRAMEINTERPOLATOR_H
#define KEYFRAMEINTERPOLATOR_H

#include "Model.h"

class KeyframeInterpolator {
public:
//	Animation keyframe information
		struct Keyframe
		{
			int m_jointIndex;
			float m_time;	// in milliseconds
			float m_parameter[3];
		};

// Joint keyframes information
		struct JointKeyframes  {
			int m_numRotationKeyframes, m_numTranslationKeyframes;
			Keyframe *m_pTranslationKeyframes;
			Keyframe *m_pRotationKeyframes;

			int m_currentTranslationKeyframe, m_currentRotationKeyframe;
		};
	
	JointKeyframes *m_pJointKeyframes; // pocet kloubu zjistim z m_model

	double m_totalTime; // total animation time
	double m_lastTime; // jaky posledni cas byl interpolatoru predan

	//int m_legs; // cislo kloubu kde zacinaji nohy

	Model *m_model; // model k animaci

	void setJointKeyframe( int jointIndex, int keyframeIndex, float time, float *parameter, bool isRotation );

public:
// constructor accepting model for animation:
	KeyframeInterpolator(Model *m);

// destructor:
	virtual ~KeyframeInterpolator();

// load keyframes from model file (will be implemented in extended class):
	virtual bool loadKeyframeData(const char *filename) = 0;

// copy keyframes from other instance with reverse order:
	void reverseKeyframeData( const KeyframeInterpolator &ki );

// sets current...Keyframes to zero
	void reset();

// write interpolated transformations for given time
// 'root_pos' is translation of root joint
	void setTransform(double time);

//	void setFinalMatrices(const float *root_pos);
  
	double getTotalTime() { return m_totalTime; }

// add total translation of root joint in this animation to 'root_pos' (3 float array)
	void addRootTrans(float *root_pos);	
protected:
// subtract root translation from all keyframes (move to origin):
	void normalizeRoot();
};


class MilkshapeKeyframeInterpolator : public KeyframeInterpolator  {
public:
// constructor accepting model for animation:
	MilkshapeKeyframeInterpolator(Model *m) : KeyframeInterpolator(m) {}

// destructor:
	virtual ~MilkshapeKeyframeInterpolator() {}

// load keyframes from model file:
	virtual bool loadKeyframeData(const char *filename);
};


#endif