/*
	Matrix.cpp

		Matrix manipulation.

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://rsn.gamedev.net/pl3d
	Copyright (C)2000, 2001, Brett Porter. All Rights Reserved.
	This source code is released under the LGPL. See license.txt for details.

	Created: 9 December 2000
	Last Edited: 8 January 2001

	Please see the file ChangeLog.html for a revision history.
*/

/*
	Modifications by Ladislav Kavan, 2002
*/

#include "stdafx.h"			// Header File For MFC
#include "Matrix.h"
#include "ikan.h"
#include "myvec.h"
#include "float.h"
#include <math.h>

Matrix::Matrix()
{
	loadIdentity();
}

Matrix::~Matrix()
{
}

void Matrix::getColumn(int idx, float *v)
{
	v[0] = m_matrix[idx*4];
	v[1] = m_matrix[idx*4 + 1];
	v[2] = m_matrix[idx*4 + 2];
}

void Matrix::setColumn(int idx, const float v[3])
{
	m_matrix[idx*4] = v[0];
	m_matrix[idx*4 + 1] = v[1];
	m_matrix[idx*4 + 2] = v[2];
}

void Matrix::postMultiply( const Matrix& matrix )
{
	float newMatrix[16];
	const float *m1 = m_matrix, *m2 = matrix.m_matrix;

	newMatrix[0] = m1[0]*m2[0] + m1[4]*m2[1] + m1[8]*m2[2];
	newMatrix[1] = m1[1]*m2[0] + m1[5]*m2[1] + m1[9]*m2[2];
	newMatrix[2] = m1[2]*m2[0] + m1[6]*m2[1] + m1[10]*m2[2];
	newMatrix[3] = 0;

	newMatrix[4] = m1[0]*m2[4] + m1[4]*m2[5] + m1[8]*m2[6];
	newMatrix[5] = m1[1]*m2[4] + m1[5]*m2[5] + m1[9]*m2[6];
	newMatrix[6] = m1[2]*m2[4] + m1[6]*m2[5] + m1[10]*m2[6];
	newMatrix[7] = 0;

	newMatrix[8] = m1[0]*m2[8] + m1[4]*m2[9] + m1[8]*m2[10];
	newMatrix[9] = m1[1]*m2[8] + m1[5]*m2[9] + m1[9]*m2[10];
	newMatrix[10] = m1[2]*m2[8] + m1[6]*m2[9] + m1[10]*m2[10];
	newMatrix[11] = 0;

	newMatrix[12] = m1[0]*m2[12] + m1[4]*m2[13] + m1[8]*m2[14] + m1[12];
	newMatrix[13] = m1[1]*m2[12] + m1[5]*m2[13] + m1[9]*m2[14] + m1[13];
	newMatrix[14] = m1[2]*m2[12] + m1[6]*m2[13] + m1[10]*m2[14] + m1[14];
	newMatrix[15] = 1;

	set( newMatrix );
}

void Matrix::preMultiply( const Matrix& matrix )
{
	float newMatrix[16];
	const float *m2 = m_matrix, *m1 = matrix.m_matrix;

	newMatrix[0] = m1[0]*m2[0] + m1[4]*m2[1] + m1[8]*m2[2];
	newMatrix[1] = m1[1]*m2[0] + m1[5]*m2[1] + m1[9]*m2[2];
	newMatrix[2] = m1[2]*m2[0] + m1[6]*m2[1] + m1[10]*m2[2];
	newMatrix[3] = 0;

	newMatrix[4] = m1[0]*m2[4] + m1[4]*m2[5] + m1[8]*m2[6];
	newMatrix[5] = m1[1]*m2[4] + m1[5]*m2[5] + m1[9]*m2[6];
	newMatrix[6] = m1[2]*m2[4] + m1[6]*m2[5] + m1[10]*m2[6];
	newMatrix[7] = 0;

	newMatrix[8] = m1[0]*m2[8] + m1[4]*m2[9] + m1[8]*m2[10];
	newMatrix[9] = m1[1]*m2[8] + m1[5]*m2[9] + m1[9]*m2[10];
	newMatrix[10] = m1[2]*m2[8] + m1[6]*m2[9] + m1[10]*m2[10];
	newMatrix[11] = 0;

	newMatrix[12] = m1[0]*m2[12] + m1[4]*m2[13] + m1[8]*m2[14] + m1[12];
	newMatrix[13] = m1[1]*m2[12] + m1[5]*m2[13] + m1[9]*m2[14] + m1[13];
	newMatrix[14] = m1[2]*m2[12] + m1[6]*m2[13] + m1[10]*m2[14] + m1[14];
	newMatrix[15] = 1;

	set( newMatrix );
}

void Matrix::preRotMultiply( const Matrix& matrix )
{
	float newMatrix[16];
	const float *m2 = m_matrix, *m1 = matrix.m_matrix;

	newMatrix[0] = m1[0]*m2[0] + m1[4]*m2[1] + m1[8]*m2[2];
	newMatrix[1] = m1[1]*m2[0] + m1[5]*m2[1] + m1[9]*m2[2];
	newMatrix[2] = m1[2]*m2[0] + m1[6]*m2[1] + m1[10]*m2[2];
	newMatrix[3] = 0;

	newMatrix[4] = m1[0]*m2[4] + m1[4]*m2[5] + m1[8]*m2[6];
	newMatrix[5] = m1[1]*m2[4] + m1[5]*m2[5] + m1[9]*m2[6];
	newMatrix[6] = m1[2]*m2[4] + m1[6]*m2[5] + m1[10]*m2[6];
	newMatrix[7] = 0;

	newMatrix[8] = m1[0]*m2[8] + m1[4]*m2[9] + m1[8]*m2[10];
	newMatrix[9] = m1[1]*m2[8] + m1[5]*m2[9] + m1[9]*m2[10];
	newMatrix[10] = m1[2]*m2[8] + m1[6]*m2[9] + m1[10]*m2[10];
	newMatrix[11] = 0;

	newMatrix[12] = m_matrix[12];
	newMatrix[13] = m_matrix[13];
	newMatrix[14] = m_matrix[14];
	newMatrix[15] = 1;

	set( newMatrix );
}

void Matrix::setTranslation( const float *translation )
{
	m_matrix[12] = translation[0];
	m_matrix[13] = translation[1];
	m_matrix[14] = translation[2];
}

void Matrix::addTranslation( const float *translation )
{
	m_matrix[12] += translation[0];
	m_matrix[13] += translation[1];
	m_matrix[14] += translation[2];
}
		
void Matrix::getTranslation( float *translation ) const
{
	translation[0] = m_matrix[12];
	translation[1] = m_matrix[13];
	translation[2] = m_matrix[14];
}

void Matrix::setInverseTranslation( const float *translation )
{
	m_matrix[12] = -translation[0];
	m_matrix[13] = -translation[1];
	m_matrix[14] = -translation[2];
}

void Matrix::setRotationDegrees( const float *angles )
{
	float vec[3];
	vec[0] = ( float )( angles[0]/180.0*PI );
	vec[1] = ( float )( angles[1]/180.0*PI );
	vec[2] = ( float )( angles[2]/180.0*PI );
	setRotationRadians( vec );
}

void Matrix::setInverseRotationDegrees( const float *angles )
{
	float vec[3];
	vec[0] = ( float )( angles[0]/180.0*PI );
	vec[1] = ( float )( angles[1]/180.0*PI );
	vec[2] = ( float )( angles[2]/180.0*PI );
	setInverseRotationRadians( vec );
}

void Matrix::setRotationRadians( const float *angles )
{
	double cr = cos( angles[0] );
	double sr = sin( angles[0] );
	double cp = cos( angles[1] );
	double sp = sin( angles[1] );
	double cy = cos( angles[2] );
	double sy = sin( angles[2] );

	m_matrix[0] = ( float )( cp*cy );
	m_matrix[1] = ( float )( cp*sy );
	m_matrix[2] = ( float )( -sp );

	double srsp = sr*sp;
	double crsp = cr*sp;

	m_matrix[4] = ( float )( srsp*cy-cr*sy );
	m_matrix[5] = ( float )( srsp*sy+cr*cy );
	m_matrix[6] = ( float )( sr*cp );

	m_matrix[8] = ( float )( crsp*cy+sr*sy );
	m_matrix[9] = ( float )( crsp*sy-sr*cy );
	m_matrix[10] = ( float )( cr*cp );
}

void Matrix::setInverseRotationRadians( const float *angles )
{
	double cr = cos( angles[0] );
	double sr = sin( angles[0] );
	double cp = cos( angles[1] );
	double sp = sin( angles[1] );
	double cy = cos( angles[2] );
	double sy = sin( angles[2] );

	m_matrix[0] = ( float )( cp*cy );
	m_matrix[4] = ( float )( cp*sy );
	m_matrix[8] = ( float )( -sp );

	double srsp = sr*sp;
	double crsp = cr*sp;

	m_matrix[1] = ( float )( srsp*cy-cr*sy );
	m_matrix[5] = ( float )( srsp*sy+cr*cy );
	m_matrix[9] = ( float )( sr*cp );

	m_matrix[2] = ( float )( crsp*cy+sr*sy );
	m_matrix[6] = ( float )( crsp*sy-sr*cy );
	m_matrix[10] = ( float )( cr*cp );
}

void Matrix::setRotationQuaternion( const Quaternion& quat )
{
	m_matrix[0] = ( float )( 1.0 - 2.0*quat[1]*quat[1] - 2.0*quat[2]*quat[2] );
	m_matrix[1] = ( float )( 2.0*quat[0]*quat[1] + 2.0*quat[3]*quat[2] );
	m_matrix[2] = ( float )( 2.0*quat[0]*quat[2] - 2.0*quat[3]*quat[1] );

	m_matrix[4] = ( float )( 2.0*quat[0]*quat[1] - 2.0*quat[3]*quat[2] );
	m_matrix[5] = ( float )( 1.0 - 2.0*quat[0]*quat[0] - 2.0*quat[2]*quat[2] );
	m_matrix[6] = ( float )( 2.0*quat[1]*quat[2] + 2.0*quat[3]*quat[0] );

	m_matrix[8] = ( float )( 2.0*quat[0]*quat[2] + 2.0*quat[3]*quat[1] );
	m_matrix[9] = ( float )( 2.0*quat[1]*quat[2] - 2.0*quat[3]*quat[0] );
	m_matrix[10] = ( float )( 1.0 - 2.0*quat[0]*quat[0] - 2.0*quat[1]*quat[1] );
}

void Matrix::setRotationAxisAngle(const float axis[3], float angle)
{
//	Quaternion q(cos(angle/2), sin(angle/2)*axis[0], sin(angle/2)*axis[1], sin(angle/2)*axis[2]);
//	setRotationQuaternion(q);
	float Sval[16] = {0, axis[2], -axis[1], 0,
					  -axis[2], 0, axis[0], 0,
					  axis[1], -axis[0], 0, 0,
					  0, 0, 0, 1};
	Matrix S, Ssin, S2; S.set(Sval); S2.set(Sval);
	loadIdentity();
	S.mulScal(sin(angle), Ssin);
	add(Ssin);
	S.postMultiply(S2);
	S.mulScal(1-cos(angle), S2);
	add(S2);
}

void Matrix::inverseRotateMatrix( Matrix &ir ) const
{
	float nm[16];
	for (int n=0;n<16;n++) nm[n]=0;
	nm[15]=1;
	for (int i=0;i<3;i++)
		for (int j=0;j<3;j++)
			nm[j + i*4] = m_matrix[i + j*4];
	ir.set(nm);
}

void Matrix::inverseHomogenousMatrix( Matrix &ir ) const
{
	ikMatrix my, inv;
	Matrix2ikMatrix(*this, my);
	inverthomoikMatrix(inv, my);
	ikMatrix2Matrix(inv, ir);
}

/*  Get orientation and translation component */
void Matrix::getOrPos( Matrix &or, Vector &pos) const
{
	float hm[3];
	hm[0] = m_matrix[12]; hm[1] = m_matrix[13]; hm[2] = m_matrix[14];
	pos.set(hm);
	or.set(m_matrix);
	or.setTranslation(zero3);
}

/*  Set orientation and translation component */
void Matrix::setOrPos( const Matrix &or, const Vector &pos)
{
	set(or.getMatrix());
	setTranslation(pos.getVector());
}

void Matrix::mulScal( float s, Matrix &res ) const
{
	// jednicku na [4][4] nechat byt!
	for (int n=0;n<15;n++) res.m_matrix[n] = m_matrix[n]*s;
}

void Matrix::add( const Matrix m )
{
	// jednicku na [4][4] nechat byt!
	for (int n=0;n<15;n++) m_matrix[n] += m.m_matrix[n];
}

// res = (1-t) * this  + t * m2
// 'obracene' vyzada interpolaci pres vetsi, tupy uhel (21.12.02):
void Matrix::SLERP( Matrix &res, const Matrix &m2, float t, bool obracene)
{
  ikMatrix i1, i2, ir;
  Matrix2ikMatrix(*this, i1);
  Matrix2ikMatrix(m2, i2);
  if (obracene) linterp2ikMatrix(ir, i1, i2, t);
  else linterpikMatrix(ir, i1, i2, t);
  ikMatrix2Matrix(ir, res);
}

void pprint(float m[16], float m2[16])
{
	int a,b;
	TRACE("\n");
		for (a=0; a<4; a++)
		{
			for (b=0; b<4; b++)
				TRACE("%.2f ", m[a+b*4]);
			TRACE("|");
			for (b=0; b<4; b++)
				TRACE("%.2f ", m2[a+b*4]);
			TRACE("\n");
		}
}

// inverze libovolne (nejen rotacni) matice:
bool Matrix::invert() 
{	
	float v[16], vys[16];
	int j, k;	
	for (int n=0; n<16; n++) v[n] = m_matrix[n];
	for (n=0; n<16; n++) vys[n] = (n%4)==(n/4) ? 1.0 : 0.0;
	// obycejna eliminace:
	for (int i=0; i<4; i++)
	{		
		float aa=v[i+4*i];
		if (fabs(aa) < 0.00001) return false; // spatne podminena matice
		for (j=0; j<4; j++)
		{
			v[i+4*j] /= aa;		
			vys[i+4*j] /= aa;
		}
		for (j=0; j<4; j++)
		{
			if (j==i) continue;
			float c = -v[j+i*4];
			for (k=0; k<4; k++)
			{
				v[j+4*k] += v[i+4*k]*c;
				vys[j+4*k] += vys[i+4*k]*c;
			}
		}		
	}
	set(vys);
	return true;
}

/* extract axis and angle */
void Matrix::axisangle( float axis[3], float &angle)
{
	const float eps = 1e-7;

    angle = acos((m_matrix[0] + m_matrix[5] + m_matrix[10] - 1) / 2.0);
    
    // Close to identity. Arbitrarily set solution to z axis rotation of 0 
    if (_isnan(angle) || fabs(angle) < eps || fabs(angle - M_PI) < eps)
    {
		angle = 0.0;
		axis[0] = axis[1] = 0.0; axis[2] = 1.0; 
    }
    else
    {
		axis[0] = m_matrix[2+1*4] - m_matrix[1+2*4];
		axis[1] = m_matrix[0+2*4] - m_matrix[2+0*4];
		axis[2] = m_matrix[1+0*4] - m_matrix[0+1*4];
		unitize(axis);
		if (norm(axis) < 1-eps)
		{ // taky skoro identita:
			angle = 0.0;
			axis[0] = axis[1] = 0.0; axis[2] = 1.0; 
		}
    }
}

float Matrix::l2norm( const Matrix m)
{
	float res = 0;
	for (int n=0; n<16; n++)
		res += (m.m_matrix[n] - m_matrix[n])*(m.m_matrix[n] - m_matrix[n]);
	return res;
}

void matrix_test()
{
	Matrix m1,m2;
//	float v[16] = {2, 0, 1, 1,
//				   1, 1, 2, 0,
//				   0, 2, 1, 2,
//				   0, 0, 1, 2};
	float v[16] = {2, 1, 0, 0,
				   1, 2, 1, 0,
				   0, 1, 2, 1,
				   0, 0, 1, 2};


	m1.set(v);
	if (!m1.invert()) TRACE("blbe podminena\n");
	m2.set(v);
	m2.postMultiply(m1);
}
