/* 
    ModeReplay.cpp

	Implementation of CMainMode abstract class for the replay of saved simulation.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "ModeReplay.h"
#include "gl1.h"
#include "MainFrm.h"
#include <assert.h>

void CModeReplay::init()
{
	//m_locModel[0] = m_locModel[1] = NULL;
	//m_inputManager = new InputManager(m_di, m_mid, NULL, NULL); // fakticky zbytecne, jen aby to nebylo NULL
	//((CMainFrame *)m_owner)->setInput(ID_INPUT0_MOUSE);
	//((CMainFrame *)m_owner)->setInput(ID_INPUT1_MOUSE);

	CFileDialog *fd = new CFileDialog(true, "sav", ".\\sequence.sav", OFN_NOCHANGEDIR, "Sequence Files (*.sav)|*.sav|All Files (*.*)|*.*||");
	fd->DoModal();
	m_animLoader = new AnimLoader(fd->GetPathName());
	delete fd;

	m_remoteModel[0] = new RemoteModel(0, 0, m_animLoader->getQuality());
	m_remoteModel[1] = new RemoteModel(1, 0, m_animLoader->getQuality());

	m_remoteModel[0]->setRootTrans(1);
	m_remoteModel[1]->setRootTrans(2);

	m_animLoader->setRemoteModels(m_remoteModel[0], m_remoteModel[1]);

	m_dbgDialog->m_slider2.SetRange(0, m_animLoader->getNumRecords()-1);
	m_dbgDialog->m_slider2.SetPos( 0 );	
	m_dbgDialog->m_slider2.EnableWindow();

	if (m_remoteModel[0]->m_model->getNumTriangles() != m_animLoader->getNumTri(0) || 
		m_remoteModel[1]->m_model->getNumTriangles() != m_animLoader->getNumTri(1))
	{
		if (m_animLoader->getNumRecords()) // je-li to 0, tak asi nastala chyba uz driv...
			MessageBox(0, "Saved animation incompatible with current settings (older version?)", "Error", MB_ICONSTOP);
		m_dbgDialog->m_slider2.EnableWindow(false);
	}		

	m_lastRec = -1;

	m_judgeText = 0; m_judgeScore = 0;
}

void CModeReplay::play() // spis "re"play :-)
{
// automaticky replay:
	//static int rec=0;
	//m_animLoader->loadRecord(rec++);
	//if (rec == m_animLoader->getNumRecords()) rec = 0;	
	int rec = m_lastRec;
	if (m_dbgDialog->posun == 1)
	{
		if (rec+1 < m_animLoader->getNumRecords()) rec++;
		while (m_animLoader->loadRecord(rec) != 0 && (rec+1<m_animLoader->getNumRecords())) rec++;
		m_dbgDialog->posun = 0;
		m_dbgDialog->m_slider2.SetPos(rec);
	}
	else if (m_dbgDialog->posun == -1)
	{
		if (rec>0) rec--;
		while (m_animLoader->loadRecord(rec) != 0 && rec>=0) rec--;
		m_dbgDialog->posun = 0;
		m_dbgDialog->m_slider2.SetPos(rec);
	}
	else rec = m_dbgDialog->m_slider2.GetPos();		

	if (rec != m_lastRec)
	{
		int tag = m_animLoader->loadRecord(rec);
		char pom[50];
		sprintf(pom, "%i", tag);
		m_dbgDialog->m_editTag.SetWindowText(pom);
		sprintf(pom, "%f", m_animLoader->getTime(rec) );
		m_dbgDialog->m_editDeltaTime.SetWindowText(pom);

		if (tag==0) glClearColor(0.0f, 1.0f, 0.0f, 0.0f);
		else if (tag==1) glClearColor(1.0f, 0.0f, 0.0f, 0.0f);
		else if (tag==2) glClearColor(0.0f, 0.0f, 1.0f, 0.0f);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		for (int h=0; h<2; h++)
		{
			//m_remoteModel[h]->changedModelState((void *)m_cr); -- to ne, to udela loadRecord
			//m_remoteModel[h]->m_model->setFinalMatrices(zero3); -- to taky
			m_remoteModel[h]->m_model->compActVers();
		}	
		m_lastRec = rec;

		m_judgeText = m_animLoader->getJudgeMsg(0, rec);
		m_judgeScore = m_animLoader->getJudgeMsg(1, rec);
	}
}

CModeReplay::~CModeReplay()
{
	delete m_remoteModel[0];
	delete m_remoteModel[1];
	delete m_animLoader;
}

void CModeReplay::draw(CCameraView *camera, int camId, const bool transpAvatar[2])
{
	float trval = (float)m_dbgDialog->m_slider.GetPos() / m_dbgDialog->m_slider.GetRangeMax();
	float tr[2] = {-1, -1};
	if (transpAvatar[0]) tr[0] = trval;
	if (transpAvatar[1]) tr[1] = trval;
	
	if (transpAvatar[0])
	{ // mohou byt i oba modely pruhledny; zbrane jsou urcite nepruhledny
		m_remoteModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);
		m_remoteModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);				
	}
	else
	{ // mohou byt i oba nepruhledny, coz neva
		m_remoteModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);	
		m_remoteModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);	
	}

	camera->setText(m_judgeText);
	camera->setScore(m_judgeScore);
}

void CModeReplay::reloadTextures(int camId)
{
	m_remoteModel[0]->m_model->reloadTextures(camId);
	m_remoteModel[1]->m_model->reloadTextures(camId);
}

FencerModel *CModeReplay::getModel(int i)
{
	if (i==0) return m_remoteModel[0];
	else if (i==1) return m_remoteModel[1];
	else assert(false);
	return m_remoteModel[0];
}