/* 
    ModeSimulation.cpp

	Implementation of CMainMode abstract class for the actual fencing simulation.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "ModeSimulation.h"
#include "gl1.h"
#include "MainFrm.h"
#include <assert.h>

void CModeSimulation::testCol(bool &t1t2, bool &t1z2, bool &t2z1, bool newClient)
{
	// nastaveni 2 pomocnych "fading" trojuhelniku:
	//Model::ActVertex *av1 = m_pModel[m_me]->getActVers(),
	//				 *av2 = m_pModel[m_me]2->getActVers();		

	if (m_dbgDialog->rebuild)
	{		
		m_locModel[0]->m_cdw->build();
		m_locModel[1]->m_cdw->build();
		m_locModel[0]->m_cdm->build();
		m_locModel[1]->m_cdm->build();		
		m_dbgDialog->rebuild = false;
	}
	else 
	{
		m_locModel[0]->m_cdw->refit();
		m_locModel[1]->m_cdw->refit();
		m_locModel[0]->m_cdm->refit();
		m_locModel[1]->m_cdm->refit();		
	}
	//for (int n=0; n<m_locModel[0]->m_model->getNumTriangles(); n++) m_pTr[0][n] = false;
	//for (n=0; n<m_locModel[1]->m_model->getNumTriangles(); n++) m_pTr[1][n] = false;
	t1t2 = m_locModel[0]->m_cdm->colTri(m_locModel[1]->m_cdm); // telo na telo
	t1z2 = m_locModel[0]->m_cdm->colTri(m_locModel[1]->m_cdw); // telo 1 na zbran 2
	t2z1 = m_locModel[1]->m_cdm->colTri(m_locModel[0]->m_cdw);	

	// rozeslu zpravy o vsech zasazenych trojuhelnicich:
	if (*m_ss)
	{
		for (int pl=0; pl<2; pl++)
			for (int n=0; n<m_locModel[pl]->m_model->getNumTriangles(); n++)
			{
				// posilam jen ty trojuhelniky, co nebyly zasazene minule nebo kdyz prisel novy klient
				if (newClient || (m_locModel[pl]->m_model->getTriangleHit(n) > m_pTr[pl][n])) 
				{						
					(*m_ss)->sendAllTriangleHit(pl, n, m_locModel[pl]->m_model->getTriangleHit(n));					
				}
			}	
	}

//	z1z2 = m_locModel[0]->m_cdw->colTri(m_locModel[1]->m_cdw, 0, 0);
// hleda se presneji v CelResp.findCol
}

float spina_x[2], spina_y[2]; // jen pro ucely ladeni a testovani

// nastavi silu pro hrace 'h':
void CModeSimulation::setForce(int h)
{
	const int cRampDuration = 1000; // ms

	float timeSinceFS = m_sinceForceSet.getTime();
	float actualForce;
	if (timeSinceFS > cRampDuration)
	{
		actualForce = m_lastForceMagnitude = 0; // sila skoncila
		//TRACE("m_lastForceMagnitude set to 0\n");
		m_sinceForceSet.reset(); // jen pro pomalejsi vypisy :-)
	}
	else
	{ // spocitam jak velka sila asi ted je (Ramp sila se interpoluje od m_lastForceMagnitude do 0)
		actualForce = m_lastForceMagnitude*
			((float)cRampDuration - timeSinceFS)/(float)cRampDuration;
	}

	if (m_cr->appliedImpulse())
	{
		float diff[3]; // impuls
		cpvector(diff, m_cr->getLastImpulse());
		// impuls pusobi pozitivne na teleso A (zbran hrace 0) a negativna na hrace 1:
		if (h==1) Vector::mul3(diff, -1.0);
		float magn = norm(diff);
		unitize(diff);
		
		//if (h==0) TRACE("norm(diff) = %f\n", magn);
		int imagn = (magn/2);
		if (imagn > 10000) imagn = 10000;	
		
		int x, y;
		x = diff[0]*1000; y = diff[1]*1000;
		spina_x[h] = diff[0]*imagn/50; spina_y[h] = diff[1]*imagn/50; // pro zobrazeni usecky
		if (h==1)
		{ // 2.hrac stoji obracene
			x = -x;
			//spina_x[h] = -spina_x[h];
			//pro zobrazovani neotacet - 2.avatar se na to bude divat obracene
		}		
		if (h==0) 
		{
			TRACE("Odraz - Hrac %i: velikost sily %i, smer [%i, %i]\n", h, imagn, x, y);
			// jen vypis:			
			TRACE("setForce - impulse = [%f, %f, %f]\n", diff[0], diff[1], diff[2]);
			TRACE("norm(diff) = %f\n", magn);
		}
		if (imagn > actualForce)
		{
			TRACE("nastavena nova sila %i\n", imagn);
			m_inputManager->setRampForce(h, x, y, imagn, 0, cRampDuration*1000);			
			m_lastForceMagnitude = imagn;
			m_sinceForceSet.reset();			
		}
		else TRACE("prilis mala sila (%i) oproti aktualni (%f)\n", imagn, actualForce);
	}	
	//m_inputManager->setConstantForce(h, x, y, imagn);		
}

void CModeSimulation::play()
{
	// starnuti zasazenych trojuhelniku:
	float elapsed = m_timer.getTime();
	m_timer.reset();
	byte sub = elapsed/20;	
	m_locModel[0]->m_model->healTri(sub);
	m_locModel[1]->m_model->healTri(sub);
	// okopiruju si aktualni stav zasazenosti trojuhelniku do m_pTr:
	for (int pl=0; pl<2; pl++)
		for (int n=0; n<m_locModel[pl]->m_model->getNumTriangles(); n++)
			m_pTr[pl][n] = m_locModel[pl]->m_model->getTriangleHit(n);

	m_inputManager->pollDevices();

	for (int h=0; h<2; h++)
	{
		m_locModel[h]->setRootTrans(h+1); // to je zbytecne nastavovat pokazde...
		m_locModel[h]->updateModel(h, (void *)m_cr);
	}
	bool z1z2 = m_cr->GO(m_animSaver); // vraci priznak kolize zbrani
	m_locModel[0]->m_model->compActVers();
	m_locModel[1]->m_model->compActVers();
	if (m_firstRun)
	{ // pokud zacinam, tak jeste jednou (predchozi hodnoty fading trojuhelnicku nemely smysl)
		m_locModel[0]->m_model->compActVers();
		m_locModel[1]->m_model->compActVers();
		m_firstRun = false;
	}
	
	bool newClient = false;
	if (*m_ss)
	{
		// hlasi-li server noveho klient, musim mu rozeslat pocatecni zpravy (i kdyby se od minule nic nezmenilo):
		newClient = (*m_ss)->m_newClient;
		if (newClient) (*m_ss)->m_newClient = false;
	}

	bool t1t2, t1z2, t2z1;
	testCol(t1t2, t1z2, t2z1, newClient);		
		
//	if (t1z2 || t2z1 || t1t2) MessageBeep(MB_ICONEXCLAMATION);

	// spocitani sil pro force-feedback:
	setForce(0); setForce(1);

	// krajeni tela :-)
	if (t2z1) m_inputManager->setFrictionCoefficient(0, 10000);
	else m_inputManager->setFrictionCoefficient(0, 0);
	if (t1z2) m_inputManager->setFrictionCoefficient(1, 10000);
	else m_inputManager->setFrictionCoefficient(1, 0);
	// kolize zbrani -> treni:
	if (z1z2) 
	{
		//TRACE("kolize zbrani, zapinam treni\n");

		//m_inputManager->setFrictionCoefficient(0, 10000);
		//m_inputManager->setFrictionCoefficient(1, 10000);
	}
//	else TRACE("nekolize zbrani, nezapinam treni\n");

/*
	if (t1z2 && t2z1)
		MessageBeep(MB_ICONEXCLAMATION);
	else if (t1z2)
		MessageBeep(MB_OK);
	else if (t2z1)
		MessageBeep(MB_ICONHAND);	
*/

	m_animSaver->saveState();	

	const ModelState *ms[2] = {(ModelState *)m_locModel[0]->getModelState(), 
		(ModelState *)m_locModel[1]->getModelState()};
	// odeslani stavu modelu, pokud se zmenily:
	bool changed = true;
	if (m_oldms0 == *(ms[0]) && m_oldms1 == *(ms[1])) changed = false;
	// posilam jen kdyz se neco zmenilo:
	if ((*m_ss)&&(changed||newClient)) (*m_ss)->sendAllModelState(ms);
	memcpy(&m_oldms0, ms[0], sizeof(ModelState));
	memcpy(&m_oldms1, ms[1], sizeof(ModelState));
	
	// odeslani hlasek rozhodciho, pokud se zmenily:
	int s1, s2;
	m_judge->getIntScore(s1, s2);
	if (m_oldJM.score0 != s1 || m_oldJM.score1 != s2 || 
		(CString)m_oldJM.msg != m_judge->getText() || newClient)
	{ // doslo ke zmene vystupu rozhodciho:
		if (*m_ss) (*m_ss)->sendAllJudgeMsg(s1, s2, m_judge->getText());
		m_oldJM.score0 = s1;
		m_oldJM.score1 = s2;
		strncpy(m_oldJM.msg, m_judge->getText(), cMaxJudgeMessage);
	}
	
	// posleme zpravy rozhodcimu (muze vyslat pokyn retreat, zpracuje se az v pristi iteraci):
	if (t1t2) m_judge->bodybody();
	if (t1z2 || t2z1) 
		m_judge->hit(t1z2, t2z1);
}


void CModeSimulation::init()
{
	m_firstRun = true;	
	m_judge = new CJudge(this);
	m_locModel[0] = new LocalModel(m_judge, 0, 0, m_quality); // to posledni (0/1) je kvalitka (subdiv)
	m_locModel[1] = new LocalModel(m_judge, 1, 0, m_quality);

	// rekneme inputManageru at to posila do nasich locModelu
	m_inputManager->setLocModels(m_locModel[0], m_locModel[1]);

	m_cr = new ColResp(m_locModel[0], m_locModel[1], m_locModel[0]->m_IKanim, m_locModel[1]->m_IKanim);

	m_locModel[0]->setRootTrans(0);

	m_pTr[0] = new byte[m_locModel[0]->m_model->getNumTriangles()];	
	m_pTr[1] = new byte[m_locModel[1]->m_model->getNumTriangles()];
	for (int n=0; n<m_locModel[0]->m_model->getNumTriangles(); n++) m_pTr[0][n] = 0;
	for (n=0; n<m_locModel[1]->m_model->getNumTriangles(); n++) m_pTr[1][n] = 0;

	m_animSaver = new AnimSaver(m_locModel[0], m_locModel[1], m_judge, "sequence.sav");	

	m_dbgDialog->rebuild = true;
	m_dbgDialog->m_slider2.EnableWindow(false);

	//for (n=0; n<cMaxCameras; n++)
	//	if (m_camera[n]) m_camera[n]->reloadTextures();

	m_oldms0 = zeroModelState;
	m_oldms1 = zeroModelState;	
}

void CModeSimulation::startJudge()
{	
	m_judge->start();
}

CModeSimulation::~CModeSimulation()
{
	delete m_locModel[0];
	delete m_locModel[1];
		
	delete m_cr;
		
	delete m_pTr[0];
	delete m_pTr[1];
		
	delete m_animSaver;
}

void CModeSimulation::draw(CCameraView *camera, int camId, const bool transpAvatar[2])
{
	float trval = (float)m_dbgDialog->m_slider.GetPos() / m_dbgDialog->m_slider.GetRangeMax();
	float tr[2] = {-1, -1};
	if (transpAvatar[0]) tr[0] = trval;
	if (transpAvatar[1]) tr[1] = trval;

	// nepruhledny modely musim kreslit jako prvni (se zapnutym Z-bufferem),
	// pruhledny jako druhy s read-only Z-bufferem
	if (transpAvatar[0])
	{ // mohou byt i oba modely pruhledny; zbrane jsou urcite nepruhledny
		m_locModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_locModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_locModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);
		m_locModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);				
	}
	else
	{ // mohou byt i oba nepruhledny, coz neva
		m_locModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);	
		m_locModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_locModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_locModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);	
	}

	// nastavim hlasky od rozhodciho:
	camera->setText(m_judge->getText());
	camera->setScore(m_judge->getScore());
	
	// kresleni force-feedbackove sily (nebo pripadne souradnych os):
	/*
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_LIGHTING);
	
	glColor3f(1.0, 0.0, 0.0);		
	glPointSize( 5.0 );
	glBegin(GL_POINTS);
	glVertex3f(0.0, 0.0, 0.0);
	glEnd();
	glLineWidth( 2.0 );
	glBegin(GL_LINES);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(spina_x[0], spina_y[0], 0.0);
	//glVertex3f(0.0, 10.0, 0.0);
	glColor3f(0.0, 0.0, 1.0);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(spina_x[1], spina_y[1], 0.0);
	glEnd();
	glEnable(GL_LIGHTING);
	/**/
}

void CModeSimulation::retreat()
{
	m_locModel[0]->m_kid->resetAnimation();
	m_locModel[1]->m_kid->resetAnimation();
}

void CModeSimulation::reloadTextures(int camId)
{
	m_locModel[0]->m_model->reloadTextures(camId);
	m_locModel[1]->m_model->reloadTextures(camId);
}

FencerModel *CModeSimulation::getModel(int i)
{
	if (i==0) return m_locModel[0];
	else if (i==1) return m_locModel[1];
	else assert(false);
	return m_locModel[0];
}