/* 
    ModeSimulation.h

	Implementation of CMainMode abstract class for the actual fencing simulation.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "MainMode.h"
#include "LocalRemoteModel.h"
#include "ColResp.h"
#include "AnimSequence.h"
#include "InputManager.h"
#include "gl1.h"
#include "DebugDialog.h"
#include "Communication.h"
#include "constants.h"

#ifndef MODESIMULATION_H
#define MODESIMULATION_H

class CModeSimulation : public CMainMode
{
protected:
// to si vyrobim:
	LocalModel	*m_locModel[2];
	ColResp		*m_cr;
	AnimSaver	*m_animSaver;
	ModelState  m_oldms0, m_oldms1;
	JudgeMessage m_oldJM;
	byte		*m_pTr[2]; // kolidujici trojuhelniky obou sermiru
	Win32Timer	m_sinceForceSet;
	int			m_lastForceMagnitude;
	bool		m_drawFadingTriangles;
	CJudge		*m_judge;
	bool		m_firstRun;
	int			m_quality;
	Win32Timer	m_timer;

// na to mam jen ukazatel:
	//CDirectInput *m_di;
	//CMouseInputDevice *m_mid;
	InputManager	*m_inputManager;
	CFrameWnd		*m_owner;
	DebugDialog		*m_dbgDialog;
	CServerSocket	**m_ss; // ukazuje na ukazatel m_ss v CChildViewu

protected: // pomocne funkce:
	void testCol(bool &t1t2, bool &t1z2, bool &t2z1, bool newClient);
	void setForce(int h);

public:
	CModeSimulation(InputManager *im, CFrameWnd *owner, 
		DebugDialog *dd, CServerSocket **ss, int quality) : m_inputManager(im), m_owner(owner), 
		m_dbgDialog(dd), m_ss(ss)
	{
		m_quality = quality;
		init();
		m_drawFadingTriangles = true; 
		m_lastForceMagnitude = 0;
	}
	virtual void init();
	virtual ~CModeSimulation();
	virtual void play();
	virtual void draw(CCameraView *camera, int camId, const bool transpAvatar[2]);
	virtual void reloadTextures(int camId);
	virtual FencerModel *getModel(int i);
	virtual int getMyState() { return cSimulate; }
//	virtual void changeQuality(int quality);

	// timto rozhodci rekne, ze se sermiri maji vzdalit
	void retreat();
	// zapne rozhodciho:
	void startJudge();
};

#endif MODESIMULATION_H