/* 
    ModeViewer.cpp
	
	Implementation of CMainMode abstract class for the viewing of remote simulation.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "ModeViewer.h"
#include "gl1.h"
#include "MainFrm.h"
#include <assert.h>

void CModeViewer::init(int quality)
{
	m_remoteModel[0] = new RemoteModel(0, 0, quality);
	m_remoteModel[1] = new RemoteModel(1, 0, quality);

	m_remoteModel[0]->setRootTrans(1);
	m_remoteModel[1]->setRootTrans(2);		
}

void CModeViewer::play() // spis "re"play :-)
{
// klientsky rezim - on-line prehravani dat prijimanych od serveru			
	// ridi se pouze daty co prisly po siti, jedinou vyjimkou je
	// starnuti zasazenych trojuhelniku:
	float elapsed = m_timer.getTime();
	m_timer.reset();
	byte sub = elapsed/20;	
	m_remoteModel[0]->m_model->healTri(sub);
	m_remoteModel[1]->m_model->healTri(sub);

	if (*m_cs)
	{
		// odesle na server informace o udalostech:
		m_inputManager->pollDevices();

		ModelState ms0, ms1;		
		bool valid = (*m_cs)->setLastModelState(0, ms0);
		valid &= (*m_cs)->setLastModelState(1, ms1);
		if (valid)
		{ // uz aspon jeden modelState dorazil			
			m_remoteModel[0]->setModelState((char *)&ms0);
			m_remoteModel[0]->changedModelState();
			m_remoteModel[1]->setModelState((char *)&ms1);
			m_remoteModel[1]->changedModelState();
		}
	}

	for (int h=0; h<2; h++)	
		m_remoteModel[h]->m_model->compActVers();	
}

CModeViewer::~CModeViewer()
{
	delete m_remoteModel[0];
	delete m_remoteModel[1];
}

void CModeViewer::draw(CCameraView *camera, int camId, const bool transpAvatar[2])
{
	float trval = (float)m_dbgDialog->m_slider.GetPos() / m_dbgDialog->m_slider.GetRangeMax();
	float tr[2] = {-1, -1};
	if (transpAvatar[0]) tr[0] = trval;
	if (transpAvatar[1]) tr[1] = trval;
	
	if (transpAvatar[0])
	{ // mohou byt i oba modely pruhledny; zbrane jsou urcite nepruhledny
		m_remoteModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);
		m_remoteModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);				
	}
	else
	{ // mohou byt i oba nepruhledny, coz neva
		m_remoteModel[0]->m_model->draw(camId, false, m_drawFadingTriangles, tr[0]);	
		m_remoteModel[0]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->getWeapon()->draw(camId, true, m_drawFadingTriangles);
		m_remoteModel[1]->m_model->draw(camId, false, m_drawFadingTriangles, tr[1]);	
	}
	camera->setText(m_msg);
	camera->setScore(m_score);
}

void CModeViewer::reloadTextures(int camId)
{
	m_remoteModel[0]->m_model->reloadTextures(camId);
	m_remoteModel[1]->m_model->reloadTextures(camId);
}

FencerModel *CModeViewer::getModel(int i)
{
	if (i==0) return m_remoteModel[0];
	else if (i==1) return m_remoteModel[1];
	else assert(false);
	return m_remoteModel[0];
}