/* 
    Model.cpp

		Model animation. Based on Brett Porter's tutorial, see below.

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/
/*
	Model.cpp

		Abstract base class for a model. The specific extended class will render the given model. 

	Author:	Brett Porter
	Email: brettporter@yahoo.com
	Website: http://www.geocities.com/brettporter/
	Copyright (C)2000, Brett Porter. All Rights Reserved.

	This file may be used only as long as this copyright notice remains intact.
*/

#include "stdafx.h"			// Header File For MFC
#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <assert.h>

#include "Model.h"
#include "MilkshapeModel.h" // kvuli zbrani
#include "Vector.h"
#include "texture.h"
#include "myvec.h"

Model::Model()
{
	m_numMeshes = 0;
	m_pMeshes = NULL;
	m_numMaterials = 0;
	m_pMaterials = NULL;
	m_numTriangles = 0;
	m_pTriangles = NULL;
	m_numVertices = 0;
	m_pVertices = NULL;
	// { NEW } 
	m_numJoints = 0;
	m_pJoints = NULL;

	m_pWeapon = NULL;

	m_leftHanded = false;
	m_swordholder = -1; 
	m_ikj[HAND] = -1; m_ikj[ELBOW] = -1; m_ikj[SHOULDER] = -1;
//	m_pTimer = Timer::getNewTimer();
//	m_looping = true;

//	restart();
	// { end NEW } 
//	m_BDdisabled = false;
	m_renderingMethod = BONES_BLENDING; // defaultni
}

Model::~Model()
{
	int i;
	for ( i = 0; i < m_numMeshes; i++ )
		delete[] m_pMeshes[i].m_pTriangleIndices;
	for ( i = 0; i < m_numMaterials; i++ )
		delete[] m_pMaterials[i].m_pTextureFilename;

	m_numMeshes = 0;
	if ( m_pMeshes != NULL )
	{
		delete[] m_pMeshes;
		m_pMeshes = NULL;
	}

	m_numMaterials = 0;
	if ( m_pMaterials != NULL )
	{
		delete[] m_pMaterials;
		m_pMaterials = NULL;
	}

	m_numTriangles = 0;
	if ( m_pTriangles != NULL )
	{
		delete[] m_pTriangles;
		m_pTriangles = NULL;
	}
	m_origNumTriangles = 0;
	if ( m_pOrigTriangles != NULL )
	{
		delete[] m_pOrigTriangles;
		m_pOrigTriangles = NULL;
	}

	m_numVertices = 0;
	if ( m_pVertices != NULL )
	{
		delete[] m_pVertices;
		m_pVertices = NULL;
	}
	m_origNumVertices = 0;
	if ( m_pOrigVertices != NULL )
	{
		delete[] m_pOrigVertices;
		m_pOrigVertices = NULL;
	}
	if ( m_pActVers != NULL )
	{
		delete[] m_pActVers;
		m_pActVers = NULL;
	}

	m_numJoints = 0;
	if ( m_pJoints != NULL )
	{
		delete[] m_pJoints;
		m_pJoints = NULL;
	}

	delete m_pWeapon;
}

void Model::setFinalMatrices(const float *root_pos)
{
 Matrix rpmat;
 rpmat.setTranslation(root_pos);
 // roota vyresim zvlast (pridava se k nemu root_pos a root_transform) (27.9.02)
 m_pJoints[0].m_final.set(root_transform.getMatrix());
 m_pJoints[0].m_final.postMultiply(rpmat);
 m_pJoints[0].m_final.postMultiply(m_pJoints[0].m_relative);
 m_pJoints[0].m_final.postMultiply(m_pJoints[0].m_transform);
 for ( int i = 1; i < m_numJoints; i++ )
	{	
		//Matrix transform;
		//transform.setRotationRadians(m_pJoints[i].rot);
		//transform.setTranslation(m_pJoints[i].trans);		

		Matrix relativeFinal( m_pJoints[i].m_relative );
		relativeFinal.postMultiply( m_pJoints[i].m_transform );

		if ( m_pJoints[i].m_parent == -1 )
			m_pJoints[i].m_final.set( relativeFinal.getMatrix());
		else
		{
			m_pJoints[i].m_final.set( m_pJoints[m_pJoints[i].m_parent].m_final.getMatrix());
			m_pJoints[i].m_final.postMultiply( relativeFinal );			
		}
		//if (i == 0) m_pJoints[i].m_final.addTranslation(root_pos);		
	}
}

void Model::compActVers(bool weapon, Matrix *swordholder) 
{	
//	GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );
//	static Matrix *swordholder;

	// Draw by group
	for ( int i = 0; i < m_numMeshes; i++ )
	{		
		{
			for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
			{
				int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
				const Triangle* pTri = &m_pTriangles[triangleIndex];

			   for ( int kc = 0; kc < 3; kc++ )
				{
				   int k;
				   if (m_leftHanded) k = 2 - kc; // zmenila se orientace
				   else k = kc;
					int index = pTri->m_vertexIndices[k];

					// { NEW }
					if ( m_pVertices[index].m_bone[0] == -1 && m_pVertices[index].m_bone[1] == -1 &&
						 m_pVertices[index].m_bone[2] == -1 && m_pVertices[index].m_bone[3] == -1 )
					{ // neni vazany k zadne kosti:
						//glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
						Vector newNormal( pTri -> m_vertexNormals[k] );
						if (weapon) newNormal.transform3( *swordholder );
						if (m_leftHanded) newNormal.mul(1, 1, -1); // reflexe v rovine xy
						//glNormal3fv( newNormal.getVector() );
						Vector::set3( m_pActVers[index].m_normal, newNormal.getVector());
						Vector newVertex( m_pVertices[index].m_location );
						if (weapon) newVertex.transform( *swordholder );
						if (m_leftHanded) newVertex.mul(1, 1, -1);
						//glVertex3fv( newVertex.getVector() );
						Vector::set3( m_pActVers[index].m_location, newVertex.getVector() );
					}
					else
					{ 
						if (m_renderingMethod == BONES_BLENDING || m_renderingMethod == BASIC_SA)
						{
							// bones deformation:
							const int bone = m_pVertices[index].m_boneID;
							const Matrix& final = m_pJoints[bone].m_final;
							const Matrix& mabs = m_pJoints[bone].m_absolute;
							const Matrix& transf = m_pJoints[bone].m_transform;
							const Matrix& rel = m_pJoints[bone].m_relative;
							const int parent = m_pJoints[bone].m_parent;
							
							Vector newNormal( pTri->m_vertexNormals[k] );
							float nn[3]; newNormal.get(nn);
							mabs.inverseRotateVect( nn ); 
							newNormal.set(nn);						
							
							Vector newVertex( m_pVertices[index].m_location );
							float nv[3]; newVertex.get(nv);
							mabs.inverseTranslateVect( nv ); // preneseni do pocatku
							mabs.inverseRotateVect( nv );
							newVertex.set(nv);
							
							if ((m_renderingMethod == BASIC_SA) || 
								(bone != m_ikj[SHOULDER] &&
								 bone != m_ikj[ELBOW] &&
								 bone != m_ikj[HAND]))
							{ // zadne bones blendovani:
								newNormal.transform3( final );
								newVertex.transform( final );
							}
							else
							{							
								float boneOr[3]; // smer kosti (v souradnicich parenta)
								int bone2=-1;
								if (bone == m_ikj[SHOULDER]) bone2 = m_ikj[ELBOW];
								else if (bone == m_ikj[ELBOW]) bone2 = m_ikj[HAND];
								else if (bone == m_ikj[HAND]) bone2 = m_midfinger;
								m_pJoints[bone2].m_relative.getTranslation(boneOr);
								unitize(boneOr);
								float P[3];
								newVertex.get(P);
								float t = DOT(boneOr, P);
								//if (bone == m_ikj[HAND]) TRACE("x = %f, y = %f, z = %f      t = %f\n", newVertex[0], newVertex[1], newVertex[2], t );
								
								Matrix id, tr2;
								// tyto hodnoty by se mohly editovat v animacnim programu
								if (bone == m_ikj[SHOULDER]) t = (t+5) / 15;
								else if (bone == m_ikj[ELBOW]) t = (t+5)/10;
								else if (bone == m_ikj[HAND]) t = (t+3)/8;
								if (t<0) t = 0;
								else if (t>=1) t = 1;
								id.SLERP(tr2, transf, t);
								newVertex.transform( tr2 );
								newVertex.transform( rel );
								newVertex.transform( m_pJoints[parent].m_final );
								
								newNormal.transform3( tr2 );
								newNormal.transform3( rel );
								newNormal.transform3( m_pJoints[parent].m_final );
							}
							
							if (m_leftHanded) newNormal.mul(1, 1, -1);
							newNormal.normalize();						
							Vector::set3( m_pActVers[index].m_normal, newNormal.getVector() );
							
							if (m_leftHanded) newVertex.mul(1, 1, -1);
							Vector::set3( m_pActVers[index].m_location, newVertex.getVector() );						
						}
						else if (m_renderingMethod == VERTEX_BLENDING)
						{
							Vector avNormal, avVertex;						
							for (int v=0;v<4;v++)
							{
								if (m_pVertices[index].m_bone[v] == -1) continue;
								const Matrix& final = m_pJoints[m_pVertices[index].m_bone[v]].m_final;
								const Matrix& mabs = m_pJoints[m_pVertices[index].m_bone[v]].m_absolute;													
								Vector newNormal( pTri->m_vertexNormals[k] );
								float nn[3]; newNormal.get(nn);
								mabs.inverseRotateVect( nn ); newNormal.set(nn);
								newNormal.transform3( final );
								if (m_leftHanded) newNormal.mul(1, 1, -1);						
								newNormal.normalize();							
								Vector newVertex( m_pVertices[index].m_location );
								float nv[3]; newVertex.get(nv);
								mabs.inverseTranslateVect( nv ); // preneseni do pocatku
								mabs.inverseRotateVect( nv ); newVertex.set(nv);
								newVertex.transform( final );
								if (m_leftHanded) newVertex.mul(1, 1, -1);
								
								newNormal.mul(m_pVertices[index].m_weight[v]);
								newVertex.mul(m_pVertices[index].m_weight[v]);
								avNormal.add(newNormal);
								avVertex.add(newVertex);							
							}													
							if (m_leftHanded) avNormal.mul(1, 1, -1);
							avNormal.normalize();						
							Vector::set3( m_pActVers[index].m_normal, avNormal.getVector() );
							
							if (m_leftHanded) avVertex.mul(1, 1, -1);
							Vector::set3( m_pActVers[index].m_location, avVertex.getVector() );						
						}
						else assert(false && "v m_renderingMethod je nesmysl");
					}
					// { end NEW }
					/* nee, root_transform se prida uz v setFinalMatrices
					// nakonec aplikuju root_transform (na hejbani s celym modelem)
					Vector newVer(m_pActVers[index].m_location);
					newVer.transform(root_transform);
					newVer.get(m_pActVers[index].m_location);

					Vector newNor(m_pActVers[index].m_normal);
					newNor.transform3(root_transform);
					newNor.get(m_pActVers[index].m_normal);
					*/
				}
			}
		}		
	}
	
	if (!weapon)  {
		if (m_swordholder<0) TRACE0("Warning: undefined m_swordholder\n");				
		if (m_pWeapon) 
		{
			// spravne nastaveni "fading" trojuhelniku:
			ActVertex oldA = m_pWeapon->m_pActVers[m_pWeapon->m_numVertices - vA];
			ActVertex oldD = m_pWeapon->m_pActVers[m_pWeapon->m_numVertices - vD];
			m_pWeapon -> compActVers(true, &m_pJoints[m_swordholder].m_final);
			m_pWeapon->m_pActVers[m_pWeapon->m_numVertices - vB] /*new B*/= oldA;
			m_pWeapon->m_pActVers[m_pWeapon->m_numVertices - vC] /*new C*/= oldD;
		}
	    return;
	}
}

// na vsechny vrcholy aplikuje matici root_transform
void Model::transformVers(bool weapon)
{
	assert(false); // obsoletni: nyni se dela uz v compActVers

	for (int n=0; n<m_numVertices; n++)
	{
		Vector newVer(m_pActVers[n].m_location);
		newVer.transform(root_transform);
		newVer.get(m_pActVers[n].m_location);

		Vector newNor(m_pActVers[n].m_normal);
		newNor.transform3(root_transform);
		newNor.get(m_pActVers[n].m_normal);
	}
	if (!weapon&&m_pWeapon) m_pWeapon->transformVers(true);
}

/*
// kresli s pomoci (drive spocitanych) m_pActVers
void Model::draw(bool weapon, bool tr[], bool fade) 
{	
	GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );
	bool txt = false;
	//static Matrix *swordholder;

	// Draw by group
	for ( int i = 0; i < m_numMeshes; i++ )
	{
		if (!fade && weapon && i == m_numMeshes-1) break; // posledni mesh zbrane je fade (konvence)

		int materialIndex = m_pMeshes[i].m_materialIndex;
		if ( materialIndex >= 0 )
		{
			glMaterialfv( GL_FRONT, GL_AMBIENT, m_pMaterials[materialIndex].m_ambient );
			glMaterialfv( GL_FRONT, GL_DIFFUSE, m_pMaterials[materialIndex].m_diffuse );
			glMaterialfv( GL_FRONT, GL_SPECULAR, m_pMaterials[materialIndex].m_specular );
			glMaterialfv( GL_FRONT, GL_EMISSION, m_pMaterials[materialIndex].m_emissive );
			glMaterialf( GL_FRONT, GL_SHININESS, m_pMaterials[materialIndex].m_shininess );

			if ( m_pMaterials[materialIndex].m_texture > 0 )
			{
				glBindTexture( GL_TEXTURE_2D, m_pMaterials[materialIndex].m_texture );
				glEnable( GL_TEXTURE_2D );
				txt = true;
			}
			else
				glDisable( GL_TEXTURE_2D );
		}
		else
		{
			// Material properties?
			glDisable( GL_TEXTURE_2D );
		}

		//float stdmat[4];
		//glGetMaterialfv( GL_FRONT, GL_DIFFUSE, stdmat);
		glBegin( GL_TRIANGLES );
		{
			for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
			{
				int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
				const Triangle* pTri = &m_pTriangles[triangleIndex];

				bool unset = false;
				if (tr && tr[triangleIndex])
				{ // zvyrazneny trojuhelnik
					unset = true;
					glEnd();
					glDisable(GL_LIGHTING);
					if (txt) glDisable(GL_TEXTURE_2D);
					glColor3f(0, 0, 1);
					glBegin( GL_TRIANGLES );
					//float dc[4] = {0, 0, 1, 1};						
					//glMaterialfv( GL_FRONT, GL_DIFFUSE, dc);
				}

				for ( int kc = 0; kc < 3; kc++ )
				{
					int k;
					if (m_leftHanded) k = 2 - kc; // zmenila se orientace
					else k = kc;
					int index = pTri->m_vertexIndices[k];
					
					glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
					glNormal3fv( m_pActVers[index].m_normal );
					glVertex3fv( m_pActVers[index].m_location );
				}

				if (unset)
				{
					glEnd();
					glEnable(GL_LIGHTING);
					if (txt) glEnable(GL_TEXTURE_2D);
					glBegin(GL_TRIANGLES);
				//glMaterialfv( GL_FRONT, GL_DIFFUSE, stdmat);
				}
			}
		}
		glEnd();
	}

	if ( texEnabled )
		glEnable( GL_TEXTURE_2D );
	else
		glDisable( GL_TEXTURE_2D );

	if (!weapon)  {		
		if (m_pWeapon) m_pWeapon -> draw(true, NULL, fade);
	    return;
	}
}
*/

// kresli s pomoci (drive spocitanych) m_pActVers
void Model::draw(int camId, bool weapon, bool fade, float transparency) 
{	
	GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );
	bool txt = false;
	//static Matrix *swordholder;

	// Draw by group (bez zvyraznenych)
	for ( int i = 0; i < m_numMeshes; i++ )
	{
		if (!fade && weapon && i == m_numMeshes-1) break; // posledni mesh zbrane je fade (konvence)

		int materialIndex = m_pMeshes[i].m_materialIndex;
		if ( materialIndex >= 0 )
		{
			glMaterialfv( GL_FRONT, GL_AMBIENT, m_pMaterials[materialIndex].m_ambient );
			glMaterialfv( GL_FRONT, GL_DIFFUSE, m_pMaterials[materialIndex].m_diffuse );
			glMaterialfv( GL_FRONT, GL_SPECULAR, m_pMaterials[materialIndex].m_specular );
			glMaterialfv( GL_FRONT, GL_EMISSION, m_pMaterials[materialIndex].m_emissive );
			glMaterialf( GL_FRONT, GL_SHININESS, m_pMaterials[materialIndex].m_shininess );

			if ( m_pMaterials[materialIndex].m_texture[camId] > 0 )
			{
				glBindTexture( GL_TEXTURE_2D, m_pMaterials[materialIndex].m_texture[camId] );
				glEnable( GL_TEXTURE_2D );
				txt = true;
			}
			else
				glDisable( GL_TEXTURE_2D );
		}
		else
		{
			// Material properties?
			glDisable( GL_TEXTURE_2D );
		}

		//if (transparent && !weapon) continue; // jsem pruhledny		

		//float stdmat[4];
		//glGetMaterialfv( GL_FRONT, GL_DIFFUSE, stdmat);
		
		if (transparency != -1)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			glDisable(GL_LIGHTING);			
			glDepthMask(GL_FALSE); // read-only Z-buffer
			glDisable(GL_TEXTURE_2D);
			//glDisable(GL_CULL_FACE);
			//glDisable(GL_DEPTH_TEST);
			
			glColor4f(0.5, 0.5, 0.5, transparency);
		}
		glBegin( GL_TRIANGLES );		
		{
			for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
			{
				int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
				const Triangle* pTri = &m_pTriangles[triangleIndex];

				if (m_pTr[triangleIndex])
				{ // je zasazeny a zaroven pruhledny -> smicham
					float wHit = (float)m_pTr[triangleIndex]/255;
					float en = wHit + transparency;
					float newtransp = en - wHit*transparency; //pst ze netrefim ani jednu barvu
					glColor4f((1.0*wHit + 0.5*transparency)/en, 0.5*transparency/en, 
						0.5*transparency/en, newtransp);
				}
				else glColor4f(0.5, 0.5, 0.5, transparency);					
	
				for ( int kc = 0; kc < 3; kc++ )
				{
					int k;
					if (m_leftHanded) k = 2 - kc; // zmenila se orientace
					else k = kc;
					int index = pTri->m_vertexIndices[k];
					
					//if (!transparent)
					//{
						glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
						glNormal3fv( m_pActVers[index].m_normal );
					//}
					glVertex3fv( m_pActVers[index].m_location );
				}
			}
		}
		glEnd();		
		if (transparency != -1)
		{
			glEnable(GL_LIGHTING);						
			//glEnable(GL_DEPTH_TEST);
			glDisable(GL_BLEND);			
			glDepthMask(GL_TRUE); // RW Z-buffer
		}		
	}

	// dokreslim ty zvyraznene:
	if (!weapon && transparency == -1)
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);		
		glDepthFunc(GL_EQUAL); // to protoze v z-bufferu ta sama hloubka uz nastavena je (ne pro pruhledny model)
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);		
		glBegin(GL_TRIANGLES);
		for ( i = 0; i < m_numMeshes; i++ )
		{
			if (!fade && weapon && i == m_numMeshes-1) break; // posledni mesh zbrane je fade (konvence)
			
			for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
			{
				int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
				const Triangle* pTri = &m_pTriangles[triangleIndex];
				
				if (m_pTr[triangleIndex] > 0)
				{
					float koef = (float)m_pTr[triangleIndex]/255;
					glColor4f(1, 0, 0, koef);

					for ( int kc = 0; kc < 3; kc++ )
					{
						int k;
						if (m_leftHanded) k = 2 - kc; // zmenila se orientace
						else k = kc;
						int index = pTri->m_vertexIndices[k];
						
						glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
						glNormal3fv( m_pActVers[index].m_normal );
						glVertex3fv( m_pActVers[index].m_location );
					}
				}				
			}
		}	
		glEnd();
		// vratim zpatky:
		glEnable(GL_LIGHTING);
		glDepthFunc(GL_LESS);
		glDisable(GL_BLEND);
	}

	if ( texEnabled )
		glEnable( GL_TEXTURE_2D );
	else
		glDisable( GL_TEXTURE_2D );

/* - to se ted musi volat explicitne (kvuli poradi vykreslovani pro alfa-blendingu):
	if (!weapon)  {		
		if (m_pWeapon) m_pWeapon -> draw(camId, true, fade, transpWeapon); //-1.0); //transparency);
	    return;
	}
*/
}

void Model::reloadTextures(int camId)
{
	for ( int i = 0; i < m_numMaterials; i++ )
		if ( strlen( m_pMaterials[i].m_pTextureFilename ) > 0 )
			m_pMaterials[i].m_texture[camId] = LoadGLTexture( m_pMaterials[i].m_pTextureFilename );
		else
			m_pMaterials[i].m_texture[camId] = 0;

	if (m_pWeapon) m_pWeapon -> reloadTextures(camId); // jeste zbran (je-li)
}

// nastavi v poli m_pTr zasazeny trojuhelnik cislo 'tri':
void Model::setTriangleHit(int tri, byte value)
{
	assert(0 <= tri && tri <= m_numTriangles);
	m_pTr[tri] = value;
}

// zjisti jestli je trojuhelnik 'tri' zasazeny:
byte Model::getTriangleHit(int tri)
{
	assert(0 <= tri && tri <= m_numTriangles);
	return m_pTr[tri];
}

bool Model::loadWeapon(const char *fname)  
{
   m_pWeapon = new MilkshapeModel();
   return m_pWeapon -> loadModelData(fname);
}

/*
void Model::setJointKeyframe( int jointIndex, int keyframeIndex, float time, float *parameter, bool isRotation )
{
	assert( m_numJoints > jointIndex );

	Keyframe& keyframe = isRotation ? m_pJoints[jointIndex].m_pRotationKeyframes[keyframeIndex] :
		m_pJoints[jointIndex].m_pTranslationKeyframes[keyframeIndex];

	keyframe.m_jointIndex = jointIndex;
	keyframe.m_time = time;
	memcpy( keyframe.m_parameter, parameter, sizeof( float )*3 );
}
*/

void Model::setupJoints()
{
	int i;
	for ( i = 0; i < m_numJoints; i++ )
	{
		Joint& joint = m_pJoints[i];

		joint.m_relative.setRotationRadians( joint.m_localRotation );
		joint.m_relative.setTranslation( joint.m_localTranslation );
		if ( joint.m_parent != -1 )
		{
			joint.m_absolute.set( m_pJoints[joint.m_parent].m_absolute.getMatrix());
			joint.m_absolute.postMultiply( joint.m_relative );
		}
		else
			joint.m_absolute.set( joint.m_relative.getMatrix());
		// aby (na zacatku) vsechny udy nebyly posunute do pocatku
		//joint.m_final.set(joint.m_absolute.getMatrix());
	}

/* to ted uz nemuzu, protoze jeden vrchol muze byt navazan az k 4 kostem
	for ( i = 0; i < m_numVertices; i++ )
	{
		Vertex& vertex = m_pVertices[i];

		if ( vertex.m_boneID != -1 )
		{
			const Matrix& matrix = m_pJoints[vertex.m_boneID].m_absolute;

			matrix.inverseTranslateVect( vertex.m_location );
			matrix.inverseRotateVect( vertex.m_location );
		}
	}

	for ( i = 0; i < m_numTriangles; i++ ) {
		Triangle& triangle = m_pTriangles[i];
		for ( int j = 0; j < 3; j++ ) {
			const Vertex& vertex = m_pVertices[triangle.m_vertexIndices[j]];
			if ( vertex.m_boneID != -1 ) {
				const Matrix& matrix = m_pJoints[vertex.m_boneID].m_absolute;
				matrix.inverseRotateVect( triangle.m_vertexNormals[j] );
			}
		}
	}
*/
}

/*
void Model::restart()
{
	for ( int i = 0; i < m_numJoints; i++ )
	{
		m_pJoints[i].m_currentRotationKeyframe = m_pJoints[i].m_currentTranslationKeyframe = 0;
		m_pJoints[i].m_final.set( m_pJoints[i].m_absolute.getMatrix());
	}

	m_pTimer->reset();
}

void Model::advanceAnimation()
{
	double time = m_pTimer->getTime();

	if ( time > m_totalTime )
	{
		if ( m_looping )
		{
			restart();
			time = 0;			
		}
		else
			time = m_totalTime;
	} 

	for ( int i = 0; i < m_numJoints; i++ )
	{
		float transVec[3];
		Matrix transform;
		int frame;
		Joint *pJoint = &m_pJoints[i];

		if ( pJoint->m_numRotationKeyframes == 0 && pJoint->m_numTranslationKeyframes == 0 )
		{
			pJoint->m_final.set( pJoint->m_absolute.getMatrix());
			continue;
		}

		frame = pJoint->m_currentTranslationKeyframe;
		while ( frame < pJoint->m_numTranslationKeyframes && pJoint->m_pTranslationKeyframes[frame].m_time < time )
		{
			frame++;
		}
		pJoint->m_currentTranslationKeyframe = frame;

		if ( frame == 0 )
			memcpy( transVec, pJoint->m_pTranslationKeyframes[0].m_parameter, sizeof ( float )*3 );
		else if ( frame == pJoint->m_numTranslationKeyframes )
			memcpy( transVec, pJoint->m_pTranslationKeyframes[frame-1].m_parameter, sizeof ( float )*3 );
		else
		{
			assert( frame > 0 && frame < pJoint->m_numTranslationKeyframes );

			const Model::Keyframe& curFrame = pJoint->m_pTranslationKeyframes[frame];
			const Model::Keyframe& prevFrame = pJoint->m_pTranslationKeyframes[frame-1];

			float timeDelta = curFrame.m_time-prevFrame.m_time;
			float interpValue = ( float )(( time-prevFrame.m_time )/timeDelta );

			transVec[0] = prevFrame.m_parameter[0]+( curFrame.m_parameter[0]-prevFrame.m_parameter[0] )*interpValue;
			transVec[1] = prevFrame.m_parameter[1]+( curFrame.m_parameter[1]-prevFrame.m_parameter[1] )*interpValue;
			transVec[2] = prevFrame.m_parameter[2]+( curFrame.m_parameter[2]-prevFrame.m_parameter[2] )*interpValue; 
		}

		frame = pJoint->m_currentRotationKeyframe;
		while ( frame < pJoint->m_numRotationKeyframes && pJoint->m_pRotationKeyframes[frame].m_time < time )
		{
			frame++;
		}
		pJoint->m_currentRotationKeyframe = frame;

		if ( frame == 0 )
			transform.setRotationRadians( pJoint->m_pRotationKeyframes[0].m_parameter );
		else if ( frame == pJoint->m_numRotationKeyframes )
			transform.setRotationRadians( pJoint->m_pRotationKeyframes[frame-1].m_parameter );
		else
		{
			assert( frame > 0 && frame < pJoint->m_numRotationKeyframes );

			const Model::Keyframe& curFrame = pJoint->m_pRotationKeyframes[frame];
			const Model::Keyframe& prevFrame = pJoint->m_pRotationKeyframes[frame-1];
			
			float timeDelta = curFrame.m_time-prevFrame.m_time;
			float interpValue = ( float )(( time-prevFrame.m_time )/timeDelta );

			assert( interpValue >= 0 && interpValue <= 1 );

#if 0
			Quaternion qPrev( prevFrame.m_parameter );
			Quaternion qCur( curFrame.m_parameter );
			Quaternion qFinal( qPrev, qCur, interpValue );
			transform.setRotationQuaternion( qFinal );
#else
			float rotVec[3];

			rotVec[0] = prevFrame.m_parameter[0]+( curFrame.m_parameter[0]-prevFrame.m_parameter[0] )*interpValue;
			rotVec[1] = prevFrame.m_parameter[1]+( curFrame.m_parameter[1]-prevFrame.m_parameter[1] )*interpValue;
			rotVec[2] = prevFrame.m_parameter[2]+( curFrame.m_parameter[2]-prevFrame.m_parameter[2] )*interpValue;

			transform.setRotationRadians( rotVec );
#endif
		}

		transform.setTranslation( transVec );
		Matrix relativeFinal( pJoint->m_relative );
		relativeFinal.postMultiply( transform );

		if ( pJoint->m_parent == -1 )
			pJoint->m_final.set( relativeFinal.getMatrix());
		else
		{
			pJoint->m_final.set( m_pJoints[pJoint->m_parent].m_final.getMatrix());
			pJoint->m_final.postMultiply( relativeFinal );			
		}
	}
}
*/


void Model::healTri(byte sub)
{
	for (int n=0; n<m_numTriangles; n++)
	{
		int nov = m_pTr[n];
		nov -= sub;
		if (nov>=0) m_pTr[n] = (byte)nov;
		else m_pTr[n] = 0;
	}
}
