/* 
    ModelSubDiv.cpp

	Implementation of mesh subdivision

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"			// Header File For MFC
#include <assert.h>

#include "Model.h"
#include "Vector.h"
#include "myvec.h"
#include "ik.h"

void Model::setNonhitTriangles()
{
	int first = m_ikj[ELBOW], last = m_swordholder-1;
	int ti[3];
	for (int n=0; n<m_numTriangles; n++)
	{	
		m_nonhitTr[n] = false;
		// projdu vsechny tri vrcholy:
		for (int m=0; m<3; m++)
		{
			ti[m] = m_pTriangles[n].m_vertexIndices[m];
			if (m_pVertices[ti[m]].m_boneID >= first && m_pVertices[ti[m]].m_boneID < last)
				m_nonhitTr[n] = true; // staci alespon jeden vrchol navazany k nezasahovym kloubum
		}
	}
}

// vraci vsechny 3 sousedni trojuhelniky (cisla viz. poznamky)
int Model::sousTroj(int t, int s[3])
{
	int vi[3], fnd=0;
	s[0] = s[1] = s[2] = -1;
	vi[0] = m_pTriangles[t].m_vertexIndices[0];
	vi[1] = m_pTriangles[t].m_vertexIndices[1];
	vi[2] = m_pTriangles[t].m_vertexIndices[2];
	for (int n=0; n<m_numTriangles; n++)
	{
		if (n==t) continue;
		int ti[3];
		ti[0] = m_pTriangles[n].m_vertexIndices[0];
		ti[1] = m_pTriangles[n].m_vertexIndices[1];
		ti[2] = m_pTriangles[n].m_vertexIndices[2];
		
		if ((eqVers(vi[0], ti[1]) && eqVers(vi[1], ti[0]))|| 
		    (eqVers(vi[0], ti[0]) && eqVers(vi[1], ti[2]))|| 
		    (eqVers(vi[0], ti[2]) && eqVers(vi[1], ti[1]))) { s[2]=n; fnd++; }
		else if ((eqVers(vi[0], ti[2]) && eqVers(vi[2], ti[0]))|| 
				 (eqVers(vi[0], ti[0]) && eqVers(vi[2], ti[1]))|| 
				 (eqVers(vi[0], ti[1]) && eqVers(vi[2], ti[2]))) { s[1]=n; fnd++; }
		else if ((eqVers(vi[1], ti[1]) && eqVers(vi[2], ti[0]))||
				 (eqVers(vi[1], ti[2]) && eqVers(vi[2], ti[1]))||
				 (eqVers(vi[1], ti[0]) && eqVers(vi[2], ti[2]))) { s[0]=n; fnd++; }
		if (fnd == 3) break;
	}	
	return fnd;
}

// vraci normalu 'N' trojuhelnika 't'
void Model::triangleNormal(int t, float N[3])
{
	ActVertex av[3];
	for (int n=0;n<3;n++)		
		memcpy( &av[n], &m_pActVers[m_pTriangles[t].m_vertexIndices[n]], sizeof(ActVertex) );
	// spocitam dve hrany a z nich normalu:
	float U[3], V[3];
	cpvector(U, av[1].m_location);
	Vector::sub3(U, av[0].m_location);
	unitize(U);
	cpvector(V, av[2].m_location);
	Vector::sub3(V, av[0].m_location);
	unitize(V);
	crossproduct(N, U, V);
	unitize(N);
}

/*
// vraci prumer krivosti vsech tri hran trojuhelnika 't'
float Model::triangleCurv(int t)
{
	int sous[3];
	float myN[3], sousN[3];
	if (sousTroj(t, sous) < 3) TRACE("Model::triangleCurv podezrely: mene nez tri sousedni trojuhelniky\n");
	triangleNormal(t, myN);
	float angleSum = 0;
	for (int n=0; n<3; n++) 
	{
		triangleNormal(sous[n], sousN);
		float dot = DOT(myN, sousN);
		float angle = acos(dot);
		if (dot >= 1) angle = 0;
		else if (dot <= -1) angle = M_PI;
		assert( 0<=angle && angle<=M_PI); // zajimavy, tenhle assert se objevi jen pri optimalizacich v rezimu FastDebug
		angleSum += angle;
	}
	return angleSum / 3;	
}

// compute the average curvature of triangles attached by all vertices to 'joint' (14.12.02):
float Model::getCurvature(int joint)
{
	Vertex v[3];

	int totTri = 0;
	float totAngle = 0;
	for (int t=0; t<m_numTriangles; t++)
	{
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		if (podminka(v, joint)) continue;
		totTri++;
		totAngle += triangleCurv(t);
	}
	float res = totAngle/totTri;
	return res;
}
/**/

// vraci maximum krivosti vsech tri hran trojuhelnika 't'
float Model::triangleCurv(int t)
{
	int sous[3];
	float myN[3], sousN[3];
	if (sousTroj(t, sous) < 3) TRACE("Model::triangleCurv podezrely: mene nez tri sousedni trojuhelniky\n");
	triangleNormal(t, myN);
	float angleMax = 0;
	for (int n=0; n<3; n++) 
	{
		triangleNormal(sous[n], sousN);
		float dot = DOT(myN, sousN);
		float angle = acos(dot);
		if (dot >= 1) angle = 0;
		else if (dot <= -1) angle = M_PI;
		assert( 0<=angle && angle<=M_PI); // zajimavy, tenhle assert se objevi jen pri optimalizacich v rezimu FastDebug
		if (angle > angleMax) angleMax = angle;
	}
	return angleMax;	
}

// compute the maximal curvature of triangles attached by all vertices to 'joint' (14.12.02):
float Model::getCurvature(int joint)
{
	Vertex v[3];
	
	float maxAngle = 0;
	for (int t=0; t<m_numTriangles; t++)
	{
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		if (podminka(v, joint)) continue;
		float curv = triangleCurv(t);
		if (curv > maxAngle) maxAngle = curv;
	}
	float res = maxAngle;
	return res;
}


/*
int Model::sousTroj(int t, int s[3])
{
	int vi[3], fnd=0;
	s[0] = s[1] = s[2] = -1;
	vi[0] = m_pTriangles[t].m_vertexIndices[0];
	vi[1] = m_pTriangles[t].m_vertexIndices[1];
	vi[2] = m_pTriangles[t].m_vertexIndices[2];
	for (int n=0; n<m_numTriangles; n++)
	{
		if (n==t) continue;
		int ti[3];
		ti[0] = m_pTriangles[n].m_vertexIndices[0];
		ti[1] = m_pTriangles[n].m_vertexIndices[1];
		ti[2] = m_pTriangles[n].m_vertexIndices[2];

		if (eqVers(ti[0], 102) || eqVers(ti[1], 102) || eqVers(ti[2], 102))
		{
			int hm=1;
			if (eqVers(ti[0], 103) || eqVers(ti[1], 103) || eqVers(ti[2], 103))
			{
				int hm=2;
			}
		}

		if ((eqVers(vi[0], ti[1]) && eqVers(vi[1], ti[0]))|| 
		    (eqVers(vi[0], ti[0]) && eqVers(vi[1], ti[2]))|| 
		    (eqVers(vi[0], ti[2]) && eqVers(vi[1], ti[1]))) { s[2]=n; fnd++; }
		else if ((eqVers(vi[0], ti[2]) && eqVers(vi[2], ti[0]))|| 
				 (eqVers(vi[0], ti[0]) && eqVers(vi[2], ti[1]))|| 
				 (eqVers(vi[0], ti[1]) && eqVers(vi[2], ti[2]))) { s[1]=n; fnd++; }
		else if ((eqVers(vi[1], ti[1]) && eqVers(vi[2], ti[0]))||
				 (eqVers(vi[1], ti[2]) && eqVers(vi[2], ti[1]))||
				 (eqVers(vi[1], ti[0]) && eqVers(vi[2], ti[2]))) { s[0]=n; fnd++; }
		if (fnd == 3) break;
	}	
	return fnd;
}
*/

bool Model::checkDup(void)
{
	bool ret = false;
	for (int i=0;i<m_numVertices;i++)
		for (int j=0;j<m_numVertices;j++)
		{			
			if (i==j) continue;
			if (eqPoints(m_pVertices[i].m_location, m_pVertices[j].m_location))
			{
				TRACE("Vrcholy %i %i (kosti %i %i) jsou duplicitni\n", i, j, m_pVertices[i].m_boneID, m_pVertices[j].m_boneID);
				ret = true;
			}
		}
	return ret;
}

//#define PODMINKA (v[0].m_boneID != joint || v[1].m_boneID != joint || v[2].m_boneID != joint)
//#define PODMINKA (v[0].m_boneID != joint && v[1].m_boneID != joint && v[2].m_boneID != joint)

// 1 na 3:
void Model::subdivMesh1k3(int joint)
{
	Vertex v[3];
	Vertex vn;
	int vi[3], vni; // index stareho a noveho vrcholu
	int new_ver=0, new_tri=0;
	//float zero[3] = {0,0,0};
	// spocitam kolik pribyde vrcholu a trojuhelnicku:
	for (int t=0; t<m_numTriangles; t++)
	{
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		if (podminka(v, joint)) continue;
		//if (v[0].m_boneID != joint && v[1].m_boneID != joint && v[2].m_boneID != joint) continue;
		new_ver ++;
		new_tri += 2;
	}
	Vertex *newVer;
	Triangle *newTri;
	// naalokuju nova, vetsi pole:
	newVer = new Vertex[m_numVertices + new_ver];
	newTri = new Triangle[m_numTriangles + new_tri];
	for (int n=0; n<m_numVertices; n++)
		memcpy(&newVer[n], &m_pVertices[n], sizeof( Vertex ));	
	int nt=0;
	for (t=0; t<m_numTriangles; t++) 
	{
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		// nedotyka-li se 'jointu', tak ho ponechame:
		if (podminka(v, joint))
		//if (v[0].m_boneID != joint && v[1].m_boneID != joint && v[2].m_boneID != joint)
			memcpy(&newTri[nt++], &m_pTriangles[t], sizeof( Triangle ));		
	}

	int nv = m_numVertices;
	//int nt = m_numTriangles;
	for (t=0; t<m_numTriangles; t++)
	{
		for (int n=0;n<3;n++)
		{
			vi[n] = m_pTriangles[t].m_vertexIndices[n];
			memcpy( &v[n], &m_pVertices[vi[n]], sizeof(Vertex) );
		}
		if (podminka(v, joint)) continue;
		//if (v[0].m_boneID != joint && v[1].m_boneID != joint && v[2].m_boneID != joint) continue;
		// vyroba novych vrcholu:
		Vector::set3(vn.m_location, v[0].m_location);
		Vector::add3(vn.m_location, v[1].m_location);
		Vector::add3(vn.m_location, v[2].m_location);
		Vector::mul3(vn.m_location, (float)1/3);
		float nn[3];
		Vector::set3(nn, m_pTriangles[t].m_vertexNormals[0]);
		Vector::add3(nn, m_pTriangles[t].m_vertexNormals[1]);
		Vector::add3(nn, m_pTriangles[t].m_vertexNormals[2]);
		Vector::mul3(nn, (float)1/3);
		float news = (m_pTriangles[t].m_s[0]+m_pTriangles[t].m_s[1]+m_pTriangles[t].m_s[2])/3;
		float newt = (m_pTriangles[t].m_t[0]+m_pTriangles[t].m_t[1]+m_pTriangles[t].m_t[2])/3;

		int j0 = v[0].m_boneID, j1 = v[1].m_boneID, j2 = v[2].m_boneID;
		if (j0 == j1) vn.m_boneID = j0;
		else if (j0 == j2) vn.m_boneID = j0;
		else if (j1 == j2) vn.m_boneID = j1;
		else vn.m_boneID = joint;
		vni = nv;
		memcpy( &newVer[nv++], &vn, sizeof(Vertex) );
		
		// vyroba novych trojuhelnicku:
		newTri[nt].m_vertexIndices[0] = vi[0];
		newTri[nt].m_vertexIndices[1] = vi[1];
		newTri[nt].m_vertexIndices[2] = vni;
		Vector::set3(newTri[nt].m_vertexNormals[0], m_pTriangles[t].m_vertexNormals[0]);
		Vector::set3(newTri[nt].m_vertexNormals[1], m_pTriangles[t].m_vertexNormals[1]);
		Vector::set3(newTri[nt].m_vertexNormals[2], nn);
		newTri[nt].m_s[0] = m_pTriangles[t].m_s[0];
		newTri[nt].m_s[1] = m_pTriangles[t].m_s[1];
		newTri[nt].m_s[2] = news;
		newTri[nt].m_t[0] = m_pTriangles[t].m_t[0];
		newTri[nt].m_t[1] = m_pTriangles[t].m_t[1];
		newTri[nt].m_t[2] = newt;
		nt++;
		newTri[nt].m_vertexIndices[0] = vni;
		newTri[nt].m_vertexIndices[1] = vi[1];
		newTri[nt].m_vertexIndices[2] = vi[2];
		Vector::set3(newTri[nt].m_vertexNormals[0], nn);
		Vector::set3(newTri[nt].m_vertexNormals[1], m_pTriangles[t].m_vertexNormals[1]);
		Vector::set3(newTri[nt].m_vertexNormals[2], m_pTriangles[t].m_vertexNormals[2]);
		newTri[nt].m_s[0] = news;
		newTri[nt].m_s[1] = m_pTriangles[t].m_s[1];
		newTri[nt].m_s[2] = m_pTriangles[t].m_s[2];
		newTri[nt].m_t[0] = newt;
		newTri[nt].m_t[1] = m_pTriangles[t].m_t[1];
		newTri[nt].m_t[2] = m_pTriangles[t].m_t[2];
		nt++;
		newTri[nt].m_vertexIndices[0] = vni;
		newTri[nt].m_vertexIndices[1] = vi[2];
		newTri[nt].m_vertexIndices[2] = vi[0];
		Vector::set3(newTri[nt].m_vertexNormals[0], nn);
		Vector::set3(newTri[nt].m_vertexNormals[1], m_pTriangles[t].m_vertexNormals[2]);
		Vector::set3(newTri[nt].m_vertexNormals[2], m_pTriangles[t].m_vertexNormals[0]);
		newTri[nt].m_s[0] = news;
		newTri[nt].m_s[1] = m_pTriangles[t].m_s[2];
		newTri[nt].m_s[2] = m_pTriangles[t].m_s[0];
		newTri[nt].m_t[0] = newt;
		newTri[nt].m_t[1] = m_pTriangles[t].m_t[2];
		newTri[nt].m_t[2] = m_pTriangles[t].m_t[0];
		nt++;		
	}
	TRACE("nv = %i, newVer+m_numVertices = %i, nt = %i, newTri+m_numTriangles = %i\n",
		nv, new_ver+m_numVertices, nt, new_tri+m_numTriangles);

	delete[] m_pVertices;
	delete[] m_pTriangles;
	m_numVertices += new_ver;
	m_numTriangles += new_tri;
	m_pVertices = newVer;
	m_pTriangles = newTri;

	// predpokladam jedinou mesh:
	delete[] m_pMeshes[0].m_pTriangleIndices;
	m_pMeshes[0].m_numTriangles = nt;
	m_pMeshes[0].m_pTriangleIndices = new int[nt];
	for (n=0; n<nt; n++)
		m_pMeshes[0].m_pTriangleIndices[n] = n;

	delete[] m_pActVers;
	m_pActVers = new ActVertex[m_numVertices];
	for (n=0; n<m_numVertices; n++) m_pActVers[n].m_location[0]=m_pActVers[n].m_location[1]=m_pActVers[n].m_location[2]=0.0;

	delete[] m_pTr;
	m_pTr = new byte[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_pTr[n] = 0;

	delete[] m_nonhitTr;
	m_nonhitTr = new bool[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_nonhitTr[n] = false;
}

int Model::newVertex(Vertex newVer[], int &nv, Vertex &v1, Vertex &v2, int joint)
{
	Vertex vn;
	Vector::set3(vn.m_location, v1.m_location);
	Vector::add3(vn.m_location, v2.m_location);
	Vector::mul3(vn.m_location, 0.5);
	
	if (v1.m_boneID == v2.m_boneID) vn.m_boneID = v1.m_boneID;
	else vn.m_boneID = joint;
	
	memcpy( &newVer[nv++], &vn, sizeof(Vertex) );
	return nv-1;
}

void Model::newTriangle(Triangle newTri[], int &nt, int v0, int v1, int v2, int t,
						int va0, int vb0, int va1, int vb1, int va2, int vb2)
{
	int va[3] = {va0, va1, va2};
	int vb[3] = {vb0, vb1, vb2};
	for (int n=0;n<3;n++)
	{
		newTri[nt].m_s[n] = (m_pTriangles[t].m_s[va[n]] + m_pTriangles[t].m_s[vb[n]]) / 2;
		newTri[nt].m_t[n] = (m_pTriangles[t].m_t[va[n]] + m_pTriangles[t].m_t[vb[n]]) / 2;
		
		Vector::set3(newTri[nt].m_vertexNormals[n], m_pTriangles[t].m_vertexNormals[va[n]]);
		Vector::add3(newTri[nt].m_vertexNormals[n], m_pTriangles[t].m_vertexNormals[vb[n]]);
		Vector::mul3(newTri[nt].m_vertexNormals[n], 0.5);
	}
	newTri[nt].m_vertexIndices[0] = v0;
	newTri[nt].m_vertexIndices[1] = v1;
	newTri[nt].m_vertexIndices[2] = v2;
	nt++;
}

int Model::protiVrchol(int v1, int v2, int vi[3])
{
	if ((!eqVers(vi[0], v1)) && (!eqVers(vi[0], v2))) return vi[0];
	else if ((!eqVers(vi[1], v1)) && (!eqVers(vi[1], v2))) return vi[1];
	else if ((!eqVers(vi[2], v1)) && (!eqVers(vi[2], v2))) return vi[2];
	else {
		TRACE0("Model::protiVrchol nenalezen\n");
		return -1;
	}
}

void Model::subdivMesh1k4(int joint, int joint2)
{
	Vertex v[3];
//	Vertex vn[3];
	int vi[3], vni[3]; // index stareho a noveho vrcholu
	int new_ver=0, new_tri=0;
//	float zero[3] = {0,0,0};	

/*
	checkDup();
	int sous[3];
	for (int tt=0;tt<m_numTriangles;tt++)
	{
		if (sousTroj(tt, sous)<3) 
		{
			m_pTriangles[tt].m_vertexIndices[0] = 0;
			m_pTriangles[tt].m_vertexIndices[1] = 0;
			m_pTriangles[tt].m_vertexIndices[2] = 0;			
			break;
		}
	}	
	return;
*/

	// trojuhelniky co nebudou deleny a ani zadny soused nebude deleny (kvuli rychlosti):
	bool *imunni = new bool[m_numTriangles]; 
	for (int t=0; t<m_numTriangles; t++) imunni[t] = true;
	for (t=0; t<m_numTriangles; t++)
	{
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		if (podminka(v, joint, joint2)) continue;
		imunni[t] = false;
		// poznacim si cisla sousedu deleneho trojuhelniku:
		int sous[3];
		sousTroj(t, sous);
		for (int k=0;k<3;k++) if (sous[k]!=-1) imunni[sous[k]] = false;
	}

	// spocitam kolik pribyde vrcholu a trojuhelnicku:
	for (t=0; t<m_numTriangles; t++)
	{
		if (imunni[t]) continue;
		for (int n=0;n<3;n++)		
		  memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );			
		if (podminka(v, joint, joint2))
		{ // jen indukovane rozdeleni
			int sous[3];			
			sousTroj(t, sous);
			new_tri += 3;
			new_ver += 3;
			for (int m=0;m<3;m++)
			{
				if (sous[m]==-1) { new_tri--; new_ver--; continue; }
				for (int n=0;n<3;n++)		
					memcpy( &v[n], &m_pVertices[m_pTriangles[sous[m]].m_vertexIndices[n]], sizeof(Vertex) );
				if (podminka(v, joint, joint2)) { new_tri--; new_ver--; }
			}			
		}
		else
		{ // rozdeleni
			new_ver += 3;
			new_tri += 3;
		}
	}
	Vertex *newVer;
	Triangle *newTri;
	// naalokuju nova, vetsi pole a prekopiruju obsah:
	newVer = new Vertex[m_numVertices + new_ver];
	newTri = new Triangle[m_numTriangles + new_tri];
	for (int n=0; n<m_numVertices; n++)
		memcpy(&newVer[n], &m_pVertices[n], sizeof( Vertex ));	
	int nt=0;
	for (t=0; t<m_numTriangles; t++) 
	{
		// ponecham pouze imunni trojuhelnicky (ostatni budou zjemneny)
		if (imunni[t]) memcpy(&newTri[nt++], &m_pTriangles[t], sizeof( Triangle ));
	}

	int nv = m_numVertices;
	//int nt = m_numTriangles;	

	// vlastni deleni trojuhelniku:
	for (t=0; t<m_numTriangles; t++)
	{
		if (imunni[t]) continue;
		for (int n=0;n<3;n++)
		{
			vi[n] = m_pTriangles[t].m_vertexIndices[n];
			memcpy( &v[n], &m_pVertices[vi[n]], sizeof(Vertex) );			
		}
		if (podminka(v, joint, joint2)) continue; // indukovane deleni se udela v dalsim cyklu		
		// vyroba novych vrcholu:
		for (n=0; n<3; n++)
		{
			int v1 = (n+1)%3, v2 = (n+2)%3; // ostatni dva vrcholy
			vni[n] = newVertex(newVer, nv, v[v1], v[v2], joint);
		}
		// vyroba novych trojuhelnicku:
		newTriangle(newTri, nt, vi[1], vni[0], vni[2], t, 1,1, 1,2, 0,1);
		newTriangle(newTri, nt, vni[0], vni[1], vni[2], t, 1,2, 0,2, 0,1);
		newTriangle(newTri, nt, vni[0], vi[2], vni[1], t, 1,2, 2,2, 0,2);
		newTriangle(newTri, nt, vni[2], vni[1], vi[0], t, 0,1, 0,2, 0,0);		
	}

	// osetreni pripadneho indukovaneho rozdeleni:
	for (t=0; t<m_numTriangles; t++)
	{ 
		if (imunni[t]) continue;
		for (int n=0;n<3;n++)		
		{
			vi[n] = m_pTriangles[t].m_vertexIndices[n];
			memcpy( &v[n], &m_pVertices[m_pTriangles[t].m_vertexIndices[n]], sizeof(Vertex) );
		}
		if (!podminka(v, joint, joint2)) continue;
		int sous[3], num=3;
		BOOL sd[3] = {true, true, true};
		sousTroj(t, sous);
		for (int m=0;m<3;m++)
		{
			if (sous[m]==-1) { num--; sd[m]=false; continue; }
			for (int n=0;n<3;n++)		
				memcpy( &v[n], &m_pVertices[m_pTriangles[sous[m]].m_vertexIndices[n]], sizeof(Vertex) );
			if (podminka(v, joint, joint2)) { sd[m]=false; num--; }
		}
		switch (num)
		{
			case 0: // zadne deleni
				TRACE0("subdiv1k4 - neni imunni?\n");
				break;
			case 1:
				{					
					int subdiv;
					if (sd[0]) subdiv = 0;
					else if (sd[1]) subdiv = 1;
					else subdiv = 2;
					int o1 = (subdiv+1)%3, o2 = (subdiv+2)%3;					
					int nov = newVertex(newVer, nv, m_pVertices[vi[o1]], m_pVertices[vi[o2]], joint);
					newTriangle(newTri, nt, vi[subdiv], vi[o1], nov, t, subdiv,subdiv, o1,o1 ,o1,o2);
					newTriangle(newTri, nt, vi[subdiv], nov, vi[o2], t, subdiv,subdiv, o1,o2 ,o2,o2);
					break;
				}
			case 2:
				{
					TRACE0("Model::subdiv1k4 zajimavy typ rozdeleni (2)\n");
					int nonsd;
					if (!sd[0]) nonsd = 0;
					else if (!sd[1]) nonsd = 1;
					else nonsd=2;
					int vpo = (nonsd+1)%3, vpred = (nonsd+2)%3;
					int nvpo = newVertex(newVer, nv, m_pVertices[vi[vpo]], m_pVertices[vi[nonsd]], joint);
					int nvpred = newVertex(newVer, nv, m_pVertices[vi[vpred]], m_pVertices[vi[nonsd]], joint);
					newTriangle(newTri, nt, vi[vpo], vi[vpred], nvpred, t, vpo,vpo, vpred,vpred ,vpred,nonsd);
					newTriangle(newTri, nt, vi[vpo], nvpred, nvpo, t, vpo,vpo, vpred,nonsd ,nonsd,vpo);
					newTriangle(newTri, nt, nvpo, nvpred, vi[nonsd], t, nonsd,vpo, vpred,nonsd ,nonsd,nonsd);
					break;
				}
			case 3:
				{ // v praxi asi nenastane, stejne jako normalni deleni trojuhelniku
					TRACE0("Model::subdiv1k4 zajimavy typ rozdeleni (3)\n");
					for (n=0; n<3; n++)
					{
						int v1 = (n+1)%3, v2 = (n+2)%3; // ostatni dva vrcholy
						vni[n] = newVertex(newVer, nv, v[v1], v[v2], joint);
					}
					// vyroba novych trojuhelnicku:
					newTriangle(newTri, nt, vi[1], vni[0], vni[2], t, 1,1, 1,2, 0,1);
					newTriangle(newTri, nt, vni[0], vni[1], vni[2], t, 1,2, 0,2, 0,1);
					newTriangle(newTri, nt, vni[0], vi[2], vni[1], t, 1,2, 2,2, 0,2);
					newTriangle(newTri, nt, vni[2], vni[1], vi[0], t, 0,1, 0,2, 0,0);		
					break;
				}
		}	
	}


	TRACE("nv = %i, newVer+m_numVertices = %i, nt = %i, newTri+m_numTriangles = %i\n",
		nv, new_ver+m_numVertices, nt, new_tri+m_numTriangles);

	delete[] imunni;
	delete[] m_pVertices;
	delete[] m_pTriangles;
	m_numVertices += new_ver;
	m_numTriangles += new_tri;
	m_pVertices = newVer;
	m_pTriangles = newTri;

	// predpokladam jedinou mesh:
	delete[] m_pMeshes[0].m_pTriangleIndices;
	m_pMeshes[0].m_numTriangles = nt;
	m_pMeshes[0].m_pTriangleIndices = new int[nt];
	for (n=0; n<nt; n++)
		m_pMeshes[0].m_pTriangleIndices[n] = n;

	delete[] m_pActVers;
	m_pActVers = new ActVertex[m_numVertices];
	for (n=0; n<m_numVertices; n++) m_pActVers[n].m_location[0]=m_pActVers[n].m_location[1]=m_pActVers[n].m_location[2]=0.0;

	delete[] m_pTr;
	m_pTr = new byte[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_pTr[n] = 0;

	delete[] m_nonhitTr;
	m_nonhitTr = new bool[m_numTriangles];
	for (n=0; n<m_numTriangles; n++) m_nonhitTr[n] = false;
}