/* 
    Physics.cpp

	Routines for center of mass and inertia tensor approximation of homogenous rigid body

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "stdafx.h"
#include "Physics.h"
#include "Matrix.h"
#include "Model.h"
#include "MilkshapeModel.h"
#include "myvec.h"

RBVolume::RBVolume(char *fweapon, char *fstena, bool log)
{ 
	m_log=log;
	//m_weapon=m; m_cdw=c; 
	// musim naloadovat weapon (a nebrat ji jako weapon, ale jako cely model)
	m_weapon = new MilkshapeModel();
	if (!m_weapon ->loadModelData(fweapon))
	{
		::MessageBox(NULL, "Can't load the weapon file.\n", "Error", MB_OK);
		return;
	}
	Matrix id;
	// nastavime actVers v normalni poloze 
	m_weapon->compActVers(true, &id);
	// zrusime stin:	
	m_weapon->getActVers()[m_weapon->getNumVertices() - 3]/*B*/= 
		m_weapon->getActVers()[m_weapon->getNumVertices() - 4]; /*A*/
	m_weapon->getActVers()[m_weapon->getNumVertices() - 2]/*C*/= 
		m_weapon->getActVers()[m_weapon->getNumVertices() - 1]; /*D*/

	m_cdw = new CDModel(m_weapon, 26, false);
	m_cdw->build();

	m_usecka = new MilkshapeModel();
	if (!m_usecka ->loadModelData(fstena))
	{
		::MessageBox(NULL, "Can't load the facet file.\n", "Error", MB_OK);
		return;
	}
	m_cdu = new CDModel(m_usecka, 26, false); // musi byt stejne k
	m_cdu->build();
}

RBVolume::~RBVolume()
{
	delete m_cdu;
	delete m_cdw;
	delete m_usecka;
	delete m_weapon;
}

// aproximace integralu
// 'fci' je pocet funkci, 'f' je pole tech funkci
// 'res' je pole na ulozeni vysledku
void RBVolume::integrate(float step, int fci, intfun f[], float res[])
{		
	FILE *logf=0;
	if (m_log) logf = fopen("logf","wt");
	float minproc = 0;

	float min[3], max[3]; // obalovy kvadr
	float p[3], pmin[3];
	bool uvnitr;

	for (int n=0; n<fci; n++) res[n] = 0;

	m_cdw->getBBox(min, max);	
	for (p[0] = min[0]+step/2; p[0] <= max[0]+step/2; p[0]+=step)
		for (p[1] = min[1]+step/2; p[1] <= max[1]+step/2; p[1]+=step)
		{	
			if (m_log)
			{
				float proc = (p[0]-min[0]-step/2)*(p[1]-min[1]-step/2)/((max[0]-min[0])*(max[1]-min[1]));
				if (proc*1000>ceil(minproc*1000)) {
					fprintf(logf,"\rHotovo %f%%",proc*100);fflush(logf);
					minproc=proc;
				}
			}

			cpvector(pmin, p); pmin[2] = min[2] - step/2;
			// optimalizace: (vyrazeni cele linky)			
			float ppom[3] = {p[0], p[1], max[2]+step/2};
			cpvector(m_usecka->getActVers()[0].m_location, pmin);
			cpvector(m_usecka->getActVers()[1].m_location, ppom);
			cpvector(m_usecka->getActVers()[2].m_location, ppom);
			m_cdu->refit();
			if (!m_cdu->collide(m_cdw)) continue;
			/**/

			uvnitr = false;
			for (p[2] = min[2]+step/2; p[2] <= max[2]+step/2; p[2]+=step)
			{
				cpvector(m_usecka->getActVers()[0].m_location, pmin);
				cpvector(m_usecka->getActVers()[1].m_location, p);
				cpvector(m_usecka->getActVers()[2].m_location, p);
				m_cdu->refit();
				if (m_cdu->collide(m_cdw)) uvnitr = !uvnitr;				
				if (uvnitr)
				{
					// prevedu vzhledem k tezisti (neni-li nulove)
					float pc[3] = {p[0]-m_cm[0], p[1]-m_cm[1], p[2]-m_cm[2]};
					for (int m=0; m<fci; m++)
					{
						res[m] += f[m](pc);
					}
				}
				cpvector(pmin,p);
			}
		}
	if (m_log) fclose(logf);
}

float volfun(const float p[3]) { int a = p[0];a++; return 1.0; }
float x_com(const float p[3]) { return p[0]; }
float y_com(const float p[3]) { return p[1]; }
float z_com(const float p[3]) { return p[2]; }
float fIx(const float p[3]) { return p[1]*p[1] + p[2]*p[2]; }
float fIy(const float p[3]) { return p[0]*p[0] + p[2]*p[2]; }
float fIz(const float p[3]) { return p[0]*p[0] + p[1]*p[1]; }
float fIyz(const float p[3]) { return p[1]*p[2]; }
float fIxz(const float p[3]) { return p[0]*p[2]; }
float fIxy(const float p[3]) { return p[0]*p[1]; }

// pocita objem a stred hmotnosti
void RBVolume::volcom(float step, float &vol, float com[3])
{
	intfun fce[4] = {volfun, x_com, y_com, z_com};
	float res[4];
	m_cm[0] = m_cm[1] = m_cm[2] = 0;
	integrate(step, 4, fce, res);
	vol = res[0]*step*step*step;
	com[0] = res[1]*step*step*step/vol;
	com[1] = res[2]*step*step*step/vol;
	com[2] = res[3]*step*step*step/vol;
}

// pocita tenzor setrvacnosti ze stredu hmotnosti, vysledky jsou v poradi Ix, Iy, Iz, Iyz, Ixz, Ixy
void RBVolume::inertia(float step, const float com[3], float Iax[3], float Idev[3])
{
	cpvector(m_cm, com);
	intfun fce[6] = {fIx, fIy, fIz, fIyz, fIxz, fIxy};
	float res[6];
	integrate(step, 6, fce, res);
	Iax[0] = res[0]*step*step*step;
	Iax[1] = res[1]*step*step*step;
	Iax[2] = res[2]*step*step*step;
	Idev[0] = res[3]*step*step*step;
	Idev[1] = res[4]*step*step*step;
	Idev[2] = res[5]*step*step*step;
}