/* 
    Physics.h

	Routines for center of mass and inertia tensor approximation of homogenous rigid body

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#include "Model.h"
#include "MilkShapeModel.h"
#include "ColDet.h"
#include "Matrix.h"
#include "ikan.h"
#include "myvec.h"

#ifndef PHYSICS_H
#define PHYSICS_H

class RBVolume  { // Rigid Body Volume (integration)
protected:
	Model	*m_weapon; 
	CDModel *m_cdw;
	float	m_cm[3]; // stred hmotnosti (nebo nuly)
	Model	*m_usecka;
	CDModel *m_cdu;
	bool	m_log; // psat log soubor?

public:
	// integrovana funkce 3Dbod -> float
	typedef float (*intfun)(const float p[3]);
	RBVolume(char *fweapon, char *fstena, bool log = false);
	~RBVolume();
	// aproximace integralu; pocita najednou vice funkci, vysledek predava v res:
	void integrate(float step, int fci, intfun f[], float res[]);
	// pocita objem a stred hmotnosti
	void volcom(float step, float &vol, float com[3]);
	// pocita tenzor setrvacnosti ze stredu hmotnosti, vysledky jsou v poradi Ix, Iy, Iz, Iyz, Ixz, Ixy
	void inertia(float step, const float com[3], float Iax[3], float Idev[3]);

};

#endif PHYSICS_H

