/* 
    Ikan.h

	Interface for IKAN system

    Author: Ladislav Kavan, 2002
	E-mail: lkav8604@barbora.ms.mff.cuni.cz
*/

#ifndef IKAN_H
#define IKAN_H

#include "Model.h"
#include "Vector.h"
#include "Matrix.h"
#include "myvec.h"
#include "srs.h"

class IKAN  {
protected:
	Model *m_model;
	SRS srs;
	float amin[2], amax[2], bmin[2], bmax[2]; // limity "swivel" ramene a zapesti (a je podle y, b podle x)
	float tmax[2]; // max "twist" ramene ([0]) a zapesti ([1])
	int m_shoulder, m_elbow, m_hand; // indexy kloubu v modelu	
	ikMatrix T, S; // transformace rameno-loket a loket-zapesti
	float m_dosah; // maximalni dosah manipulatoru

	// promitne cil tak, aby byl dosazitelny manipulatorem max. delky 'm_dosah'
	void scaleGoal( Vector &goal );	
	// test, jestli swing lezi ve sfericke elipse definovane amin[i], amax[i], bmin[i], bmax[i]
	// pokud ano, vraci zapornou hodnotu
	float inEllipse(int i, float a, float b);
	// pocita "pohodlnost" polohy
	// (prvni a, b, t se tykaji ramene, dalsi zapesti)
	float comfort(float a, float b, float t, float za, float zb, float zt);
	// spolecna rutina pro setGoal a compElbowPos
	// mod 1: spocita elPos pomoci optimalizace pohodlnosti; nenastavuje rotace kloubu
	// mod 2: spocita IK podle daneho elPos
	int pomSetGoal( Vector goalPos, Matrix orient, float elPos[3], int mode);


public:
	int hit;
// constructor accepting model for animation:
	IKAN(Model *m);

// destructor:
	virtual ~IKAN();

	// nastavi cilovou pozici a orientaci EE (odpovida swordholder.final);
	// vstupem je pozice a orientace swordholderu ve svetovych souradnicich a pozice loktu (v souradnicich rootu)
	int setGoal( Vector goalPos, Matrix orient, float elPos[3]);
	int compElbowPos( Vector goalPos, Matrix orient, float elPos[3]);

// prevede cilovou pozici a orientaci swordholderu do pozice a orientace ruky
	void SHGoal2HandGoal( Vector &goalPos, Matrix &orient);

// pomocne:
	int swivdeg;
	float elbowPos[3]; 
};

// prevody mezi konvencemi zapisu matic
void Matrix2ikMatrix(const Matrix &m, ikMatrix im);
void ikMatrix2Matrix(const ikMatrix im, Matrix &m);

#endif